package crmdna.api.endpoint;

import com.google.api.server.spi.config.Api;
import com.google.api.server.spi.config.ApiMethod;
import com.google.api.server.spi.config.ApiMethod.HttpMethod;
import com.google.appengine.api.datastore.Cursor;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.appengine.api.datastore.QueryResultIterator;
import com.google.appengine.api.users.User;
import com.google.gdata.data.spreadsheet.CustomElementCollection;
import com.google.gdata.data.spreadsheet.ListEntry;
import com.google.gdata.data.spreadsheet.ListFeed;
import com.googlecode.objectify.Key;
import com.googlecode.objectify.cmd.Query;
import com.googlecode.objectify.cmd.QueryKeys;
import crmdna.api.endpoint.ClientApi.ClientEnum;
import crmdna.calling.Campaign;
import crmdna.calling.CampaignEntity;
import crmdna.calling.CampaignProp;
import crmdna.calling.CampaignQueryCondition;
import crmdna.client.Client;
import crmdna.client.isha.IshaAPAC;
import crmdna.client.isha.ieo.IshaIEO;
import crmdna.client.isha.ieo.IshaIEOProfileEntity;
import crmdna.client.isha.ieo.IshaIEOProfileProp;
import crmdna.common.AssertUtils;
import crmdna.common.DateUtils;
import crmdna.common.StopWatch;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.common.config.ConfigCRMDNA;
import crmdna.common.contact.Contact;
import crmdna.common.contact.ContactProp;
import crmdna.datamigration.DataMigration;
import crmdna.email.EmailProp;
import crmdna.email.GAEEmail;
import crmdna.gspreadsheet.GSpreadSheet;
import crmdna.interaction.*;
import crmdna.list.*;
import crmdna.member.Member;
import crmdna.member.MemberEntity;
import crmdna.member.MemberLoader;
import crmdna.member.MemberProp;
import crmdna.member.MemberQueryCondition;
import crmdna.participant.ParticipantEntity;
import crmdna.practice.Practice;
import crmdna.program.Program;
import crmdna.programtype.ProgramType;
import crmdna.registration.Registration;
import crmdna.registration.RegistrationEntity;
import crmdna.registration.RegistrationProp;
import crmdna.registration.RegistrationQueryCondition;
import crmdna.user.User.ClientLevelPrivilege;
import crmdna.user.UserEntity;

import javax.annotation.Nullable;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import java.util.*;
import java.util.List;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.ensure;
import static crmdna.common.OfyService.ofy;
import static crmdna.member.MemberBulkSaver.memberBulkSaver;

@Api(name = "developersOnly")
public class DevelopersOnly {

    private final static Logger LOGGER = Logger.getLogger(DevelopersOnly.class.getName());

    @ApiMethod(name = "purgeAllMemberData", httpMethod = HttpMethod.POST)
    public APIResponse purgeAllMemberData(@Named("client") String client,
                                          @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (client == null)
            client = "isha";

        String login = null;

        try {
            login = Utils.getLoginEmail(user);

            crmdna.user.User.ensureClientLevelPrivilege(client, login,
                    ClientLevelPrivilege.PURGE_MEMBER_DATA);

            // ensure dev mode is set
            boolean devMode = ConfigCRMDNA.get().toProp().devMode;
            if (!devMode)
                throw new APIException().status(Status.ERROR_OPERATION_NOT_ALLOWED).message(
                        "This operation is allowed only in dev mode");

            QueryKeys<MemberEntity> memberQueryKeys = ofy(client).load().type(MemberEntity.class).keys();

            QueryKeys<ParticipantEntity> participantQueryKeys =
                    ofy(client).load().type(ParticipantEntity.class).keys();

            QueryKeys<RegistrationEntity> registrationQueryKeys =
                    ofy(client).load().type(RegistrationEntity.class).keys();

            List<Key<MemberEntity>> memberKeys = memberQueryKeys.list();
            List<Key<ParticipantEntity>> participantKeys = participantQueryKeys.list();
            List<Key<RegistrationEntity>> registrationKeys = registrationQueryKeys.list();


            ofy(client).delete().keys(memberKeys);
            ofy(client).delete().keys(participantKeys);
            ofy(client).delete().keys(registrationKeys);

            String message = "";
            message += "Deleted [" + memberQueryKeys.list().size() + "] members. ";
            message += "Deleted [" + participantQueryKeys.list().size() + "] participants. ";
            message += "Deleted [" + registrationQueryKeys.list().size() + "] registrants. ";

            return new APIResponse().status(Status.SUCCESS).message(message);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(name = "reconstituteKey", httpMethod = HttpMethod.GET)
    public APIResponse reconstituteKey(@Named("webSafeKeyString") String webSafeKey,
                                       @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req) {
        try {

            StopWatch stopWatch = StopWatch.createStarted();
            com.google.appengine.api.datastore.Key key = KeyFactory.stringToKey(webSafeKey);

            Map<String, String> map = new HashMap<>();
            map.put("Kind", key.getKind());
            map.put("AppId", key.getAppId());
            map.put("Name", key.getName());
            map.put("Namespace", key.getNamespace());
            map.put("Parent", key.getNamespace());
            map.put("Id", key.getId() + "");

            return new APIResponse()
                    .status(Status.SUCCESS)
                    .message(
                            "Data store key successfully reconstituted in [" + stopWatch.nsElapsed() + "] ns. ")
                    .object(map).processingTimeInMS(stopWatch.msElapsed());


        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req));
        }
    }

    @ApiMethod(name = "testKeyToIDConversion", httpMethod = HttpMethod.GET)
    public APIResponse testKeyToIDConversion(@Named("client") String client,
                                             @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {

            login = Utils.getLoginEmail(user);
            crmdna.user.User.ensureValidUser(client, login);

            String message = "";

            StopWatch stopWatch = StopWatch.createStarted();
            List<Key<MemberEntity>> keys =
                    ofy(client).load().type(MemberEntity.class).limit(1000).keys().list();

            message += "[" + keys.size() + "] member keys loaded in [" + stopWatch.msElapsed() + "] ms. ";

            List<Long> ids = new ArrayList<>();

            stopWatch = StopWatch.createStarted();
            for (Key<MemberEntity> key : keys) {
                ids.add(key.getId());
            }

            message += "[" + ids.size() + "] ids obtained in [" + stopWatch.nsElapsed() + "] ns. ";

            List<String> keyStrings = new ArrayList<>(keys.size());
            for (Key<MemberEntity> key : keys) {
                keyStrings.add(KeyFactory.keyToString(key.getRaw()));
            }

            return new APIResponse().status(Status.SUCCESS).message(message)
                    .object(keyStrings);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(name = "testMaxDatastoreAsyncQueryLimit", httpMethod = HttpMethod.GET)
    public APIResponse testMaxDatastoreAsyncQueryLimit(@Named("client") String client,
                                                       @Nullable @Named("numQueriesDefault10000") Integer numQueries,
                                                       @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (client == null)
            client = "isha";

        String login = null;

        try {

            if (numQueries == null)
                numQueries = 10000;

            ensure(numQueries > 0, "Invalid numQueries [" + numQueries + "]. Should be positive.");

            login = Utils.getLoginEmail(user);

            crmdna.user.User.ensureValidUser(client, login);

            Set<String> emails = new HashSet<>(numQueries);
            for (int i = 0; i < numQueries; i++) {
                if (i == 0)
                    emails.add("sathya.t@ishafoundation.org");
                else if (i == 1)
                    emails.add("thulasidhar@gmail.com");
                else
                    emails.add("email" + i + "@invalid.com");
            }

            StopWatch sw = StopWatch.createStarted();
            Map<String, Long> emailVsMemberId = Member.getMemberIdFromEmail(client, emails);

            return new APIResponse()
                    .status(Status.SUCCESS)
                    .object(emailVsMemberId)
                    .message(
                        "Completed [" + numQueries + "] async datastore queries to get memberId from email for [" + numQueries
                            + "] email(s)").processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(name = "copyAllEntitiesToAnotherClient", httpMethod = HttpMethod.POST)
    public APIResponse copyAllEntitiesToAnotherClient(@Named("sourceClient") String sourceClient,
                                                      @Named("targetClient") String targetClient,
                                                      @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;

        try {

            login = Utils.getLoginEmail(user);

            DataMigration.copyAllEntitiesToAnotherClient(sourceClient, targetClient, login);

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                    new RequestInfo().client("multiple").req(req).login(login));
        }
    }

    // To be removed after bhairavi data migration
    @ApiMethod(name = "resaveAllPrograms", httpMethod = HttpMethod.POST)
    public APIResponse resaveAllPrograms(@Named("clientEnum") ClientEnum clientEnum,
                                         @Nullable @Named("clientIfOther") String clientOther,
                                         @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = EndpointUtils.getClient(clientEnum, clientOther);

        try {

            login = Utils.getLoginEmail(user);

            Program.resaveAll(client, login);

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                    new RequestInfo().client("multiple").req(req).login(login));
        }
    }

    // To be removed after bhairavi data migration
    @ApiMethod(name = "resaveAllProgramTypes", httpMethod = HttpMethod.POST)
    public APIResponse resaveAllProgramTypes(@Named("clientEnum") ClientEnum clientEnum,
                                             @Nullable @Named("clientIfOther") String clientOther,
                                             @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = EndpointUtils.getClient(clientEnum, clientOther);

        try {

            login = Utils.getLoginEmail(user);

            ProgramType.resaveAll(client, login);

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                    new RequestInfo().client("multiple").req(req).login(login));
        }
    }

    @ApiMethod(name = "resaveAllMembers", httpMethod = HttpMethod.POST)
    public APIResponse resaveAllMembers(@Named("clientEnum") ClientEnum clientEnum,
        @Nullable @Named("clientIfOther") String clientOther,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = EndpointUtils.getClient(clientEnum, clientOther);

        try {

            login = Utils.getLoginEmail(user);

            Client.ensureValid(client);

            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<MemberEntity> all = ofy(client).load().type(MemberEntity.class).list();

            ofy(client).save().entities(all).now();

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("multiple").req(req).login(login));
        }
    }

    @ApiMethod(name = "resaveAllRegistrations", httpMethod = HttpMethod.POST)
    public APIResponse resaveAllRegistrations(@Named("clientEnum") ClientEnum clientEnum,
        @Nullable @Named("clientIfOther") String clientOther,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = EndpointUtils.getClient(clientEnum, clientOther);

        try {

            login = Utils.getLoginEmail(user);
            Client.ensureValid(client);

            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<RegistrationEntity> all = ofy(client).load().type(RegistrationEntity.class).list();

            ofy(client).save().entities(all).now();

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("multiple").req(req).login(login));
        }
    }

    @ApiMethod(name = "populateSubUnsubGroupIds", httpMethod = HttpMethod.POST)
    public APIResponse populateSubUnsubGroupIds(@Named("clientEnum") ClientEnum clientEnum,
                                        @Nullable @Named("clientOther") String clientOther,
                                        @Named("firstChar") Utils.SingleChar firstChar,
                                        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = EndpointUtils.getClient(clientEnum, clientOther);

        try {

            login = Utils.getLoginEmail(user);

            Client.ensureValid(client);
            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            MemberQueryCondition mqc = new MemberQueryCondition(client, 20000);
            mqc.nameFirstChar = firstChar.toString();
            List<MemberEntity> memberEntities = MemberLoader.queryEntities(mqc, login);

            Set<Long> memberIds = new HashSet<>();
            for (MemberEntity memberEntity : memberEntities) {
                memberIds.add(memberEntity.getId());
            }

            Set<Long> changedMemberIds = Member.populateSubUnsubGroupIds(client, memberIds, login);
            return new APIResponse().status(Status.SUCCESS)
                    .message("[" + changedMemberIds.size() + "] member entities changed").object(changedMemberIds);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                    new RequestInfo().client(client).req(req).login(login));
        }
    }

    @ApiMethod(name = "fixIEOUserGroupMapping", httpMethod = HttpMethod.POST)
    public APIResponse fixIEOUserGroupMapping(
        @Nullable @Named("cursor") String cursor,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            Client.ensureValid(client);
            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<MemberEntity> updatedMemberEntities = new ArrayList<>();

            Query<MemberEntity> q = ofy("isha").load().type(MemberEntity.class).filter("ieoProfileId >", 0).limit(1000);
            if (cursor != null) {
                q = q.startAt(Cursor.fromWebSafeString(cursor));
            }

            boolean proceed = false;
            QueryResultIterator<MemberEntity> iterator = q.iterator();
            while (iterator.hasNext()) {
                MemberEntity memberEntity = iterator.next();
                IshaIEOProfileProp profileProp = IshaIEO.getProp(memberEntity.ieoProfileId);
                if (profileProp != null) {
                    ContactProp contactProp = new ContactProp();
                    contactProp.homeAddress.city = profileProp.city;
                    contactProp.homeAddress.country = profileProp.country;
                    contactProp.homeAddress.pincode = profileProp.postalCode;
                    contactProp.asOfyyyymmdd = DateUtils.toYYYYMMDD(new Date());

                    long newGroupId = IshaAPAC.getGroupId(profileProp.country);
                    long oldGroupId = (long) memberEntity.getGroupIds().toArray()[0];

                    boolean upd1 = memberEntity.addRemoveGroupId(newGroupId, oldGroupId);
                    boolean upd2 = Member.populateContactDetails(memberEntity, contactProp);

                    if (upd1 || upd2) {
                        updatedMemberEntities.add(memberEntity);
                    }
                }
                proceed = true;
            }

            cursor = proceed ? iterator.getCursor().toWebSafeString() : "NA";

            /*
            List<MemberEntity> memberEntities = ofy("isha").load().type(MemberEntity.class).filter("ieoProfileId >", 0).limit(2500).list();
            for (MemberEntity memberEntity : memberEntities) {
            }*/

            ofy(client).save().entities(updatedMemberEntities);

            return new APIResponse().status(Status.SUCCESS)
                .message("[" + updatedMemberEntities.size() + "] member entities have been updated. Cursor " + cursor);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                    new RequestInfo().client(client).req(req).login(login));
        }
    }

    @ApiMethod(name = "resyncIEOProfileIdsInMemberEntities", httpMethod = HttpMethod.POST)
    public APIResponse resyncIEOProfileIdsInMemberEntities(
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String login = null;

        try {

            login = Utils.getLoginEmail(user);

            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<IshaIEOProfileEntity> entities = ofy("isha").load().type(IshaIEOProfileEntity.class).list();
            Map<Long, Long> memberIdvsProfileIdMap = new HashMap<>();
            for (IshaIEOProfileEntity entity : entities) {
                memberIdvsProfileIdMap.put(entity.getMemberId(), entity.getId());
            }

            Map<Long, MemberEntity> memberEntitiesMap = ofy("isha").load().type(MemberEntity.class).ids(memberIdvsProfileIdMap.keySet());

            for (MemberEntity memberEntity : memberEntitiesMap.values()) {
                memberEntity.ieoProfileId = memberIdvsProfileIdMap.get(memberEntity.getId());
            }

            ofy("isha").save().entities(memberEntitiesMap.values()).now();

            return new APIResponse().status(Status.SUCCESS).message("[" + memberEntitiesMap.size() + "] member entities changed");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().req(req).login(login));
        }
    }

    @ApiMethod(path = "updateCampaignProgramIdInInteractions", httpMethod = HttpMethod.POST)
    public APIResponse updateCampaignProgramIdInInteractions(
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String login = null;

        try {

            login = Utils.getLoginEmail(user);

            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<CampaignEntity> campaignEntities = ofy("isha").load().type(CampaignEntity.class).list();
            Map<Long,Long> campIdVsProgIdMap = new HashMap<>();
            for (CampaignEntity campaignEntity : campaignEntities) {
                CampaignProp campaignProp = campaignEntity.toProp();
                campIdVsProgIdMap.put(campaignProp.campaignId, campaignProp.programId);
            }

            List<InteractionEntity> entities = ofy("isha").load().type(InteractionEntity.class).list();
            List<InteractionEntity> updEntities = new ArrayList<>();
            for (InteractionEntity entity : entities) {
                if (entity.getCampaignId() != null) {
                    Interaction.setProgramId(entity, campIdVsProgIdMap.get(entity.getCampaignId()));
                    updEntities.add(entity);
                }
            }

            ofy("isha").save().entities(updEntities).now();

            return new APIResponse().status(Status.SUCCESS).message("[" + updEntities.size() + "] interactions changed");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().req(req).login(login));
        }
    }

    @ApiMethod(name = "fixSubscribedGroupIds", httpMethod = HttpMethod.POST)
    public APIResponse fixSubscribedGroupIds(
        @Nullable @Named("cursor") String cursor,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            Client.ensureValid(client);
            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            List<MemberEntity> updatedMemberEntities = new ArrayList<>();

            Query<MemberEntity> q = ofy("isha").load().type(MemberEntity.class).limit(3000);
            if (cursor != null) {
                q = q.startAt(Cursor.fromWebSafeString(cursor));
            }

            boolean proceed = false;
            QueryResultIterator<MemberEntity> iterator = q.iterator();
            while (iterator.hasNext()) {
                MemberEntity memberEntity = iterator.next();
                if (memberEntity.updateSubscribeGroupIds()) {
                    updatedMemberEntities.add(memberEntity);
                }
                proceed = true;
            }

            cursor = proceed ? iterator.getCursor().toWebSafeString() : "NA";

            ofy(client).save().entities(updatedMemberEntities);

            return new APIResponse().status(Status.SUCCESS)
                .message("[" + updatedMemberEntities.size() + "] member entities have been updated. Cursor " + cursor);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client(client).req(req).login(login));
        }
    }

    @ApiMethod(path = "processSpreadsheetForProspects", httpMethod = HttpMethod.POST)
    public APIResponse processSpreadsheetForProspects(@Named("spreadsheetKey") String spreadsheetKey,
        @Named("startIndex") int startIndex, @Named("numRecords") int numRecords,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        //https://developers.google.com/google-apps/spreadsheets/

        String authString = req.getHeader("Authorization");
        if (authString == null) {
            return new APIResponse().status(Status.ERROR_LOGIN_REQUIRED);
        }

        Logger logger = Logger.getLogger(DevelopersOnly.class.getName());

        String accessToken = authString.substring("Bearer ".length());
        String login = Utils.getLoginEmail(user);

        try {
            ListFeed listFeed = GSpreadSheet.getSpreadSheetListFeed(accessToken, spreadsheetKey);

            Map<String,Integer> map = new HashMap<>();
            for (int i = startIndex; i < listFeed.getEntries().size(); i++) {

                if (i == (startIndex + numRecords)) {
                    break;
                }

                ListEntry row = listFeed.getEntries().get(i);
                CustomElementCollection rowCEC = row.getCustomElements();

                String email = rowCEC.getValue("emailaddress");
                if (email != null) {
                    email = email.trim().toLowerCase();
                    if (Utils.isValidEmailAddress(email)) {
                        map.put(email, i);
                    } else {
                        rowCEC.setValueLocal("Remarks", "Invalid Email");
                        row.update();
                    }
                } else {
                    rowCEC.setValueLocal("Remarks", "Invalid Email");
                    row.update();
                }
            }

            MemberQueryCondition qc = new MemberQueryCondition();
            qc.client = "isha";
            qc.emails = map.keySet();
            List<MemberProp> memberProps = MemberLoader.queryProps(qc, crmdna.user.User.SUPER_USER);
            MemberLoader.populateDependents("isha", memberProps, crmdna.user.User.SUPER_USER);

            for (MemberProp memberProp : memberProps) {
                int rowNumber = map.get(memberProp.contact.email);
                ListEntry row = listFeed.getEntries().get(rowNumber);

                String contents = row.getCustomElements().getValue("lists");
                StringBuilder lists = new StringBuilder(contents == null ? "" : contents);
                for (ListProp listProp : memberProp.listProps) {
                    lists.append(listProp.displayName).append(",");
                }
                row.getCustomElements().setValueLocal("lists", lists.toString());
                row.getCustomElements().setValueLocal("memberid", memberProp.memberId + "");
                if (memberProp.ieoClassNumber != null) {
                    row.getCustomElements().setValueLocal("ieoclassnumber", memberProp.ieoClassNumber.toString() + "");
                }
                logger.info("row " + rowNumber + ", email " + memberProp.contact.email);
                row.setEtag("*");
                row.update();
            }

            ListEntry row = listFeed.getEntries().get(startIndex + numRecords);
            row.getCustomElements().setValueLocal("status", "updated-upto-here");
            row.setEtag("*");
            row.update();

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login));
        }

    }

    @ApiMethod(path = "importInteractionsForCampaign", httpMethod = HttpMethod.GET)
    public APIResponse importInteractionsForCampaign(
        @Named("campaignId") long campaignId,
        @Named("spreadsheetKey") String spreadsheetKey,
        @Named("startIndex") int startIndex,
        @Named("numRecords") int numRecords,
        @Nullable @Named("dryRun") Boolean dryRun,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;
        if (dryRun == null) {
            dryRun = false;
        }

        try {
            login = Utils.getLoginEmail(user);
            StopWatch sw = StopWatch.createStarted();
            int created = 0, skipped = 0;

            List<Map<String, String>> listOfMap =
                GSpreadSheet.getPublishedSpreadsheetAsListOfMap(spreadsheetKey, numRecords, startIndex);

            CampaignProp campaignProp = Campaign.safeGet("isha", campaignId).toProp();
            CampaignQueryCondition cqc = new CampaignQueryCondition();
            cqc.programIds.add(campaignProp.programId);

            List<CampaignEntity> campaignEntities = Campaign.query("isha", cqc, login);
            List<Long> campaignIds = new ArrayList<>();
            for (CampaignEntity entity : campaignEntities) {
                campaignIds.add(entity.toProp().campaignId);
            }

            List<Long> memberIds = new ArrayList<>();
            int lineNumber = 2;
            for (Map<String, String> map : listOfMap) {
                String memberIdStr = map.get("memberid");
                AssertUtils.ensureNotNullNotEmpty(memberIdStr, lineNumber + ": member id missing");
                AssertUtils.ensure(Utils.canParseAsLong(memberIdStr));
                memberIds.add(Long.parseLong(memberIdStr));
                lineNumber++;
            }

            Map<Long, MemberEntity> mapIdVsEntity = MemberLoader.get("isha", memberIds, login);
            List<ContactProp> updContactProps = new ArrayList<>();
            List<MemberEntity> updMemberEntities = new ArrayList<>();

            lineNumber = 2;
            for (Map<String, String> map : listOfMap) {

                String userEmail = map.get("user");
                AssertUtils.ensureNotNullNotEmpty(userEmail, lineNumber + ": user missing");
                userEmail = userEmail.trim().toLowerCase();
                UserEntity userEntity = crmdna.user.User.get("isha", userEmail);
                AssertUtils.ensureNotNull(userEntity, lineNumber + ": user " + userEmail + " not found");

                long memberId = Long.parseLong(map.get("memberid"));

                String firstName = map.get("firstname");
                AssertUtils.ensureNotNullNotEmpty(firstName, lineNumber + ": firstname missing");

                String lastName = map.get("lastname");

                Contact.Gender gender = Contact.getGender(map.get("gender"));

                String dateYYYYMMDDStr = map.get("date");
                AssertUtils.ensureNotNullNotEmpty(dateYYYYMMDDStr, lineNumber + ": date missing");
                AssertUtils.ensure(Utils.canParseAsLong(dateYYYYMMDDStr));
                int dateYYYYMMDD = Integer.parseInt(dateYYYYMMDDStr);
                DateUtils.ensureFormatYYYYMMDD(dateYYYYMMDD);

                String progressStr = map.get("progress");
                AssertUtils.ensureNotNullNotEmpty(progressStr, lineNumber + ": progress missing");
                Interaction.Progress progress = Interaction.Progress.valueOf(progressStr.toUpperCase().trim());

                String responseStr = map.get("response");
                Interaction.Response response = Utils.isNotNullNotEmpty(responseStr) ? Interaction.Response.valueOf(responseStr.toUpperCase().trim()) : null;
                AssertUtils.ensure((progress != Interaction.Progress.COMPLETE) || (response != null),
                    lineNumber + ": response missing for COMPLETE interactions");

                String interaction = map.get("interaction");
                if (interaction == null) {
                    interaction = "";
                }

                InteractionQueryCondition qc = new InteractionQueryCondition();
                qc.campaignIds.addAll(campaignIds);
                qc.memberIds.add(memberId);

                List<InteractionProp> interactionProps = Interaction.queryProps("isha", qc, crmdna.user.User.SUPER_USER);
                if (interactionProps.size() == 0) {
                    if (!dryRun) {
                        InteractionProp interactionProp =
                            Interaction.createInteraction("isha", memberId, interaction, Interaction.InteractionType.PHONE,
                                DateUtils.toDate(dateYYYYMMDD), campaignId, true, userEmail);
                        Interaction.updateInteraction("isha", interactionProp.interactionId, null, null, null, progress, response, login);
                    }

                    LOGGER.info(lineNumber + ": created interaction");
                    created++;

                } else {
                    for (InteractionProp interactionProp : interactionProps) {
                        ensure(interactionProp.campaignId == campaignId,
                            "Existing assignment for Campaign " + interactionProp.campaignId + " found for member " + memberId);
                    }
                    LOGGER.info(lineNumber + ": already created interaction");
                    skipped++;
                }

                MemberEntity memberEntity = mapIdVsEntity.get(memberId);
                MemberProp memberProp = memberEntity.toProp();
                boolean changed = false;
                if (!firstName.equals(memberProp.contact.firstName)) {
                    changed = true;
                    memberProp.contact.firstName = firstName;
                }
                if ((lastName != null) && !lastName.equals(memberProp.contact.lastName)) {
                    changed = true;
                    memberProp.contact.lastName = lastName;
                }
                if ((gender != null) && !gender.equals(memberProp.contact.gender)) {
                    changed = true;
                    memberProp.contact.gender = gender;
                }
                if (changed) {
                    updMemberEntities.add(memberEntity);
                    updContactProps.add(memberProp.contact);
                }

                lineNumber++;
                if (sw.msElapsed() > 40000) {
                    break;
                }
            }

            if (!dryRun) {
                memberBulkSaver("isha", updMemberEntities).setContactsSameSizeList(updContactProps).populateDependantsAndSave();
            }

            return new APIResponse().status(Status.SUCCESS).message("Imported " + created + " interactions. " + "Skipped " + skipped);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(name = "getShambhaviListViaEmail", httpMethod = HttpMethod.POST)
    public APIResponse getShambhaviListViaEmail(
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {


        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            Client.ensureValid(client);
            ensure(crmdna.user.User.isSuperUser(login), "Allowed only for super user");

            MemberQueryCondition mqc = new MemberQueryCondition(client, 3000);
            mqc.practiceIds.add(Practice.safeGetByIdOrName(client, "Shambhavi").toProp().practiceId);
            mqc.subscribedGroupIds.add(1L);
            List<MemberProp> memberProps = MemberLoader.queryProps(mqc,login);

            Map<Long, MemberProp> memberPropMap = new HashMap<>();
            for (MemberProp memberProp : memberProps) {
                if (memberProp.contact.getPhoneNos().length() > 0) {
                    memberPropMap.put(memberProp.memberId, memberProp);
                }
            }

            CampaignQueryCondition cqc = new CampaignQueryCondition();
            cqc.programIds.add(163L);
            List<CampaignEntity> campaignEntities = Campaign.query(client, cqc, login);

            List<Long> campaignIds = new ArrayList<>();
            for (CampaignEntity campaignEntity : campaignEntities) {
                campaignIds.add(campaignEntity.toProp().campaignId);
            }

            InteractionQueryCondition iqc = new InteractionQueryCondition();
            iqc.campaignIds.addAll(campaignIds);
            iqc.memberIds.addAll(memberPropMap.keySet());

            List<InteractionProp> interactionProps = Interaction.queryProps(client, iqc, login);

            LOGGER.info("Already assigned " + interactionProps.size() + " members in earlier campaigns");

            for (InteractionProp interactionProp : interactionProps) {
                memberPropMap.remove(interactionProp.memberId);
            }

            RegistrationQueryCondition rqc = new RegistrationQueryCondition();
            rqc.memberIds.addAll(memberPropMap.keySet());
            rqc.programId = 163L;

            List<RegistrationProp> registrationProps = Registration.query(client, rqc, login);

            LOGGER.info(registrationProps.size() + " members already registered");

            for (RegistrationProp registrationProp : registrationProps) {
                memberPropMap.remove(registrationProp.memberId);
            }

            rqc = new RegistrationQueryCondition();
            rqc.memberIds.addAll(memberPropMap.keySet());
            rqc.programId = 164L;

            registrationProps = Registration.query(client, rqc, login);

            for (RegistrationProp registrationProp : registrationProps) {
                memberPropMap.remove(registrationProp.memberId);
            }

            EmailProp emailProp = new EmailProp();
            emailProp.toEmailAddresses.add(login);
            emailProp.bodyHtml = "<br>Query results attached";
            emailProp.csvAttachmentData = Member.getCSV(client, memberPropMap.values());
            emailProp.attachmentName = "Shambhavi-Members.csv";
            emailProp.subject = "Member Query Results [Shambhavi]";
            GAEEmail.send(emailProp);

            return new APIResponse().status(Status.SUCCESS).message("[" + memberPropMap.size() + "] member props emailed ");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client(client).req(req).login(login));
        }
    }

    @ApiMethod(path = "assignCallsForUser", httpMethod = HttpMethod.GET)
    public APIResponse assignCallsForUser(
        @Named("campaignId") long campaignId,
        @Named("userEmail") String userEmail,
        @Named("memberIdsCSV") String memberIdsCSV,
        @Nullable @Named("previewDefaultFalse") Boolean preview,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            crmdna.user.User.ensureValidUser(client, userEmail);
            long userId = crmdna.user.User.safeGet(client, userEmail).toProp(client).userId;
            Set<Long> userIds = Utils.getSet(userId);

            List<String> memberIdsStr = Arrays.asList(memberIdsCSV.split("[,\\s]+"));
            Set<Long> memberIds = new HashSet<>();
            for (String memberIdStr : memberIdsStr) {
                memberIds.add(Long.parseLong(memberIdStr));
            }

            CampaignQueryCondition cqc = new CampaignQueryCondition();
            cqc.programIds.add(163L);
            List<CampaignEntity> campaignEntities = Campaign.query(client, cqc, login);

            List<Long> campaignIds = new ArrayList<>();
            for (CampaignEntity campaignEntity : campaignEntities) {
                campaignIds.add(campaignEntity.toProp().campaignId);
            }

            InteractionQueryCondition iqc = new InteractionQueryCondition();
            iqc.campaignIds.addAll(campaignIds);
            iqc.memberIds.addAll(memberIds);

            List<InteractionProp> interactionProps = Interaction.queryProps(client, iqc, login);
            StringBuilder sb = new StringBuilder();
            for (InteractionProp interactionProp : interactionProps) {
                sb.append(interactionProp.memberId).append(",");
            }
            ensure(sb.length() == 0, "Already assigned interaction for the same program to " + sb.toString());

            CallAssignmentResult result;
            if (preview == null) preview = false;

            Map<Long, Integer> userIdVsMaxMembers = CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());

            if (preview) {
                result = Calling.preview(client, campaignId, memberIds, userIds, userIdVsMaxMembers, true, login);
            } else {
                result = Calling.assign(client, campaignId, memberIds, userIds, userIdVsMaxMembers, true, login);
            }

            return new APIResponse().status(Status.SUCCESS).object(result);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(path = "transferAssignedCalls", httpMethod = HttpMethod.GET)
    public APIResponse transferAssignedCalls(
        @Named("campaignId") long campaignId,
        @Named("newUserEmail") String newUserEmail,
        @Named("memberIdsCSV") String memberIdsCSV,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            crmdna.user.User.ensureValidUser(client, newUserEmail);

            List<String> memberIdsStr = Arrays.asList(memberIdsCSV.split("[,\\s]+"));
            Set<Long> memberIds = new HashSet<>();
            for (String memberIdStr : memberIdsStr) {
                memberIds.add(Long.parseLong(memberIdStr));
            }

            Campaign.safeGet(client, campaignId).toProp();

            InteractionQueryCondition iqc = new InteractionQueryCondition();
            iqc.campaignIds.add(campaignId);
            iqc.memberIds.addAll(memberIds);

            List<InteractionProp> interactionProps = Interaction.queryProps(client, iqc, login);
            ensure(interactionProps.size() == memberIds.size(),
                "interactionProps.size(" + interactionProps.size() + ") != " +
                "memberIds.size(" + memberIds.size() + ")");

            for (InteractionProp interactionProp : interactionProps) {
                Interaction.updateInteraction(client, interactionProp.interactionId, null, null, newUserEmail, null, null, login);
            }

            return new APIResponse().status(Status.SUCCESS).message(interactionProps.size() + " updated");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }
}
