package crmdna.api.endpoint;

import com.google.api.server.spi.config.Api;
import com.google.api.server.spi.config.ApiMethod;
import com.google.api.server.spi.config.ApiMethod.HttpMethod;
import com.google.appengine.api.users.User;
import com.google.gson.Gson;
import crmdna.calling.Campaign;
import crmdna.common.DateUtils;
import crmdna.common.DateUtils.DateRange;
import crmdna.common.StopWatch;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.interaction.*;
import crmdna.interaction.Interaction.InteractionType;
import crmdna.member.MemberLoader;
import crmdna.member.MemberQueryCondition;
import crmdna.program.Program;
import crmdna.program.ProgramProp;

import javax.annotation.Nullable;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.ensure;

@Api(name = "interaction")
public class InteractionApi {

    private static final Logger LOG = Logger.getLogger(InteractionApi.class.getName());

    @ApiMethod(path = "createInteraction", httpMethod = HttpMethod.POST)
    public APIResponse createInteraction(@Named("client") String client,
                                         @Named("memberId") long memberId, @Named("interactionType") InteractionType interactionType,
                                         @Named("content") String content,
                                         @Nullable @Named("campaignId") Long campaignId,
                                         @Nullable @Named("ensureWithinCampaignDatesDefaultTrue") Boolean ensureWithinCampaignDates,
                                         @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                         HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();

            login = Utils.getLoginEmail(user);

            ensureWithinCampaignDates = (ensureWithinCampaignDates == null) ? true : ensureWithinCampaignDates;
            InteractionProp prop =
                    Interaction.createInteraction(client, memberId, content, interactionType, new Date(),
                            campaignId, ensureWithinCampaignDates, login);

            return new APIResponse().status(Status.SUCCESS).object(prop)
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "updateInteraction", httpMethod = HttpMethod.POST)
    public APIResponse updateInteraction(@Named("client") String client,
                                         @Named("interactionId") long interactionId, @Nullable @Named("newMemberId") Long newMemberId,
                                         @Nullable @Named("newInteractionType") InteractionType newInteractionType,
                                         @Nullable @Named("newUserEmail") String newUserEmail,
                                         @Nullable @Named("newProgress") Interaction.Progress newProgress,
                                         @Nullable @Named("newResponse") Interaction.Response newResponse,
                                         @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();

            login = Utils.getLoginEmail(user);

            InteractionProp prop =
                    Interaction.updateInteraction(client, interactionId, newMemberId, newInteractionType,
                            newUserEmail, newProgress, newResponse, login);

            return new APIResponse().status(Status.SUCCESS).object(prop)
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "deleteInteraction", httpMethod = HttpMethod.DELETE)
    public APIResponse deleteInteraction(@Named("client") String client,
                                         @Named("interactionId") long interactionId,
                                         @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();
            login = Utils.getLoginEmail(user);

            Interaction.deleteInteraction(client, interactionId, login);

            return new APIResponse().status(Status.SUCCESS)
                    .object("interaction Id [" + interactionId + "] deleted")
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "query", httpMethod = HttpMethod.GET)
    public APIResponse query(@Named("client") String client,
                             @Nullable @Named("memberId") Long memberId, @Nullable @Named("userId") Long userId,
                             @Nullable @Named("campaignId") Long campaignId,
                             @Nullable @Named("response") Interaction.Response response,
                             @Nullable @Named("progress") Interaction.Progress progress,
                             @Nullable @Named("interactionType") InteractionType interactionType,
                             @Nullable @Named("dateRange") DateRange dateRange,
                             @Nullable @Named("startIndex") Integer startIndex,
                             @Nullable @Named("numResults") Integer numResults,
                             @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();
            login = Utils.getLoginEmail(user);

            InteractionQueryCondition qc = new InteractionQueryCondition();
            if (memberId != null) {
                qc.memberIds.add(memberId);
            }
            if (userId != null) {
                qc.userIds.add(userId);
            }
            if (null != interactionType) {
                qc.interactionTypes.add(interactionType.toString());
            }
            if (null != campaignId) {
                qc.campaignIds.add(campaignId);
            }
            qc.progress = progress;
            qc.response = response;
            qc.end = new Date();
            if (dateRange != null) {
                qc.start = new Date(qc.end.getTime() - DateUtils.getMilliSecondsFromDateRange(dateRange));
            }

            qc.numResults = numResults;

            InteractionQueryResult result =
                    Interaction.query(client, qc, login);

            return new APIResponse().status(Status.SUCCESS).object(result)
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "createSubInteraction", httpMethod = HttpMethod.POST)
    public APIResponse createSubInteraction(@Named("client") String client,
                                            @Named("interactionId") long interactionId, @Named("content") String content,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();
            login = Utils.getLoginEmail(user);

            Interaction.createSubInteraction(client, interactionId, content, new Date(), login);

            return new APIResponse().status(Status.SUCCESS).message("Added sub interaction")
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "deleteSubInteraction", httpMethod = HttpMethod.DELETE)
    public APIResponse deleteSubInteraction(@Named("client") String client,
                                            @Named("interactionId") long interactionId, @Named("subInteractionId") long subInteractionId,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();

            login = Utils.getLoginEmail(user);

            Interaction.deleteSubInteraction(client, interactionId, subInteractionId, login);

            return new APIResponse().status(Status.SUCCESS)
                    .message("Sub interaction [" + subInteractionId + "] deleted")
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "updateSubInteraction", httpMethod = HttpMethod.POST)
    public APIResponse updateSubInteraction(@Named("client") String client,
                                            @Named("interactionId") long interactionId, @Named("subInteractionId") long subInteractionId,
                                            @Named("content") String content, @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            StopWatch sw = StopWatch.createStarted();
            login = Utils.getLoginEmail(user);

            Interaction.updateSubInteraction(client, interactionId, subInteractionId, content, null, login);

            return new APIResponse().status(Status.SUCCESS)
                    .message("Sub interaction [" + subInteractionId + "] updated")
                    .processingTimeInMS(sw.msElapsed());

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "getInteractionScore", httpMethod = HttpMethod.GET)
    public APIResponse getInteractionScore(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("memberId") long memberId, @Named("userId") long userId,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req) {

        String client = null;
        try {

            client = EndpointUtils.getClient(clientEnum, clientOther);

            UserMemberProp userMember = new UserMemberProp();
            userMember.memberId = memberId;
            userMember.userId = userId;

            List<InteractionScoreProp> props = InteractionScore.get(client, Utils.getList(userMember));

            return new APIResponse().status(Status.SUCCESS)
                    .object(props);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req));
        }
    }

    @ApiMethod(path = "assignCalls", httpMethod = HttpMethod.POST)
    public APIResponse assignCalls(@Named("client") String client,
                                            @Named("campaignId") long campaignId,
                                            @Named("memberIds") Set<Long> memberIds,
                                            @Named("userIds") Set<Long> userIds,
                                            @Nullable @Named("ensureWithinCampaignDatesDefaultTrue") Boolean ensureWithinCampaignDates,
                                            @Nullable @Named("previewDefaultFalse") Boolean preview,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            login = Utils.getLoginEmail(user);

            if (ensureWithinCampaignDates == null) {
                ensureWithinCampaignDates = true;
            }
            if (preview == null) {
                preview = false;
            }

            CallAssignmentResult result;

            Map<Long, Integer> userIdVsMaxMembers =
                    CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());
            LOG.info("userIdVsMaxMembers: " + new Gson().toJson(userIdVsMaxMembers));
            if (preview) {
                result = Calling.preview(client, campaignId, memberIds, userIds,
                        userIdVsMaxMembers, ensureWithinCampaignDates, login);
            } else {
                result = Calling.assign(client, campaignId, memberIds, userIds,
                        userIdVsMaxMembers, ensureWithinCampaignDates, login);
            }

            return new APIResponse().status(Status.SUCCESS)
                    .object(result);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "assignCallsToParticipants", httpMethod = HttpMethod.POST)
    public APIResponse assignCallsToParticipants(@Named("client") String client,
                                   @Named("campaignId") long campaignId,
                                   @Named("programIds") Set<Long> programIds,
                                   @Named("userIds") Set<Long> userIds,
                                   @Nullable @Named("ensureWithinCampaignDatesDefaultTrue") Boolean ensureWithinCampaignDates,
                                   @Nullable @Named("ensureProgramInSameGroupAsCampaignDefaultTrue") Boolean ensureProgramInSameGroupAsCampaign,
                                   @Nullable @Named("previewDefaultFalse") Boolean preview,
                                   @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                   HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            login = Utils.getLoginEmail(user);

            if (ensureWithinCampaignDates == null) {
                ensureWithinCampaignDates = true;
            }
            if (preview == null) {
                preview = false;
            }
            if (ensureProgramInSameGroupAsCampaign == null) {
                ensureProgramInSameGroupAsCampaign = true;
            }

            if (programIds.size() > 20) {
                throw new APIException().message("Cannot assign calls to more than 20 program participants in one shot. Please split it up into multiple batches.")
                        .status(Status.ERROR_INVALID_INPUT);
            }

            if (ensureProgramInSameGroupAsCampaign) {
                long groupId = Campaign.safeGet(client, campaignId).toProp().groupId;

                Map<Long, ProgramProp> programIdVsProp = Program.getProps(client, programIds);
                for (Map.Entry<Long, ProgramProp> e : programIdVsProp.entrySet()) {
                    ProgramProp programProp = e.getValue();
                    ensure(e.getValue().groupProp.groupId == groupId,
                            "Program [" + programProp.programId + "] has group [" +
                            programProp.groupProp.groupId + "] which is different from the campaign's group [" +
                            groupId + "]. To remove this restriction you can set the flag [ensureProgramInSameGroupAsCampaign] to false");
                }
            }

            MemberQueryCondition qc = new MemberQueryCondition(client, 10000);
            qc.programIds.addAll(programIds);
            List<Long> memberIds = MemberLoader.queryIds(qc, login);

            return assignCalls(client, campaignId, new HashSet<>(memberIds), userIds, ensureWithinCampaignDates,
                    preview, showStackTrace, req, user);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "assignCallsForUser", httpMethod = HttpMethod.POST)
    public APIResponse assignCallsForUser(@Named("client") String client,
        @Named("campaignId") long campaignId,
        @Named("memberIdsCSV") String memberIdsCSV,
        @Named("userEmail") String userEmail,
        @Nullable @Named("ensureWithinCampaignDatesDefaultTrue") Boolean ensureWithinCampaignDates,
        @Nullable @Named("previewDefaultFalse") Boolean preview,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        if (null == client)
            client = "isha";

        String login = null;

        try {
            login = Utils.getLoginEmail(user);

            if (ensureWithinCampaignDates == null) {
                ensureWithinCampaignDates = true;
            }
            if (preview == null) {
                preview = false;
            }

            CallAssignmentResult result;

            crmdna.user.User.ensureValidUser(client, userEmail);
            long userId = crmdna.user.User.safeGet(client, userEmail).toProp(client).userId;
            Set<Long> userIds = Utils.getSet(userId);

            String[] memberIdsArr = memberIdsCSV.split(",");
            Set<Long> memberIds = new HashSet<>();
            for (String memberIdStr : memberIdsArr) {
                memberIds.add(Long.parseLong(memberIdStr));
            }

            Map<Long, Integer> userIdVsMaxMembers =
                CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());

            LOG.info("userIdVsMaxMembers: " + new Gson().toJson(userIdVsMaxMembers));

            if (preview) {
                result = Calling.preview(client, campaignId, memberIds, userIds, userIdVsMaxMembers, ensureWithinCampaignDates, login);
            } else {
                result = Calling.assign(client, campaignId, memberIds, userIds, userIdVsMaxMembers, ensureWithinCampaignDates, login);
            }

            return new APIResponse().status(Status.SUCCESS).object(result);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req).login(login));
        }
    }

}
