package crmdna.api.endpoint;

import com.google.api.server.spi.config.Api;
import com.google.api.server.spi.config.ApiMethod;
import com.google.appengine.api.users.User;
import crmdna.common.DateUtils;
import crmdna.common.Utils;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.mail2.*;
import crmdna.program.Program;
import crmdna.program.ProgramEntity;
import crmdna.program.ProgramProp;

import javax.annotation.Nullable;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import java.util.Date;
import java.util.List;

import static crmdna.common.AssertUtils.ensure;

/**
 * Created by sathya on 15/10/15.
 */
@Api(name = "mail")
public class MailScheduleApi {

    @ApiMethod(path = "createScheduledEmail", httpMethod = ApiMethod.HttpMethod.POST)
    public APIResponse createScheduledEmail(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("mailContentId") long mailContentId,
                                            @Named("listId") long listId,
                                            @Nullable @Named("programId") Long programId,
                                            @Nullable @Named("firstNameIfMissingDefaultIsha") String defaultFirstName,
                                            @Nullable @Named("lastNameIfMissingDefaultIsha") String defaultLastName,
                                            @Named("timeToSendHoursFromNow") double hoursFromNow,
                                            @Named("senderEmail") String senderEmail,
                                            @Nullable @Named("maxTimeHoursFromTimeToSendDefault48") Double maxTimeHoursFromTimeToSend,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailScheduleInput msi = new MailScheduleInput();
            msi.mailContentId = mailContentId;
            msi.listId = listId;
            msi.programId = programId;
            msi.defaultFirstName = defaultFirstName == null ? "Isha" : defaultFirstName;
            msi.defaultLastName = defaultLastName == null ? "Isha" : defaultLastName;

            ensure(hoursFromNow > 0 && hoursFromNow < 120,
                    "hoursFromNow should be greater than 1 and less than 120");
            long msFromNow = (long) (hoursFromNow * 3600 * 1000);
            msi.scheduledTime = new Date(new Date().getTime() + msFromNow);
            if (maxTimeHoursFromTimeToSend == null) {
                maxTimeHoursFromTimeToSend = 48.0;
            }
            ensure(maxTimeHoursFromTimeToSend > 0, "maxTime should be after scheduledTime");
            long maxTimeMsFromTimeToSend = (long) (maxTimeHoursFromTimeToSend * 3600 * 1000);
            msi.maxTime = new Date(msi.scheduledTime.getTime() + maxTimeMsFromTimeToSend);
            msi.senderEmail = senderEmail;

            MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).object(mailScheduleProp);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "getScheduledEmail", httpMethod = ApiMethod.HttpMethod.GET)
    public APIResponse getScheduledEmail(@Named("client") ClientApi.ClientEnum clientEnum,
                                         @Nullable @Named("clientIfOther") String clientOther,
                                         @Named("mailScheduleId") long mailScheduleId,
                                         @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                         HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailScheduleProp mailScheduleProp = MailSchedule.safeGet(client, mailScheduleId).toProp();

            return new APIResponse().status(APIResponse.Status.SUCCESS).object(mailScheduleProp);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "updateScheduledEmail", httpMethod = ApiMethod.HttpMethod.POST)
    public APIResponse updateScheduledEmail(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("mailScheduleId") long mailScheduleId,
                                            @Nullable @Named("newDefaultFirstName") String newDefaultFirstName,
                                            @Nullable @Named("newDefaultLastName") String newDefaultLastName,
                                            @Nullable @Named("delayInHoursFromCurrentScheduledTime") Integer hours,
                                            @Nullable @Named("newEnabled") Boolean newEnabled,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailScheduleProp mailScheduleProp = MailSchedule.safeGet(client, mailScheduleId).toProp();

            Date newScheduledTime = null;
            if (hours != null) {
                newScheduledTime = new Date(mailScheduleProp.scheduledTime.getTime() + hours * 3600 * 1000);
            }

            mailScheduleProp = MailSchedule.update(client, mailScheduleId, newDefaultFirstName,
                    newDefaultLastName, newEnabled, newScheduledTime, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).object(mailScheduleProp);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "deleteScheduledEmail", httpMethod = ApiMethod.HttpMethod.DELETE)
    public APIResponse deleteScheduledEmail(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("mailScheduleId") long mailScheduleId,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailSchedule.delete(client, mailScheduleId, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS)
                    .object("Deleted mailScheduleId [" + mailScheduleId + "]");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "queryScheduledEmail", httpMethod = ApiMethod.HttpMethod.GET)
    public APIResponse queryScheduledEmail(@Named("client") ClientApi.ClientEnum clientEnum,
                                           @Nullable @Named("clientIfOther") String clientOther,
                                           @Nullable @Named("listId") Long listId,
                                           @Nullable @Named("programId") Long programId,
                                           @Nullable @Named("mailContentId") Long mailContentId,
                                           @Nullable @Named("groupId") Long groupId,
                                           @Nullable @Named("userEmail") String userEmail,
                                           @Nullable @Named("senderEmail") String senderEmail,
                                           @Nullable @Named("processed") Boolean processed,
                                           @Nullable @Named("sendSuccess") Boolean sendSuccess,
                                           @Nullable @Named("scheduleTimeStartYYYYMMDD") Integer scheduleStartYYYYMMDD,
                                           @Nullable @Named("scheduleTimeEndYYYYMMDD") Integer scheduleEndYYYYMMDD,
                                           @Nullable @Named("enabled") Boolean enabled,
                                           @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                           HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailScheduleQueryCondition qc = new MailScheduleQueryCondition();
            qc.userEmail = userEmail;
            qc.groupId = groupId;
            qc.listId = listId;

            if (mailContentId != null) {
                qc.mailContentIds.add(mailContentId);
            }
            qc.processed = processed;
            qc.sendSuccess = sendSuccess;
            qc.programId = programId;
            qc.senderEmail = senderEmail;
            qc.enabled = enabled;

            if (scheduleStartYYYYMMDD != null) {
                qc.scheduledTimeStart = DateUtils.toDate(scheduleStartYYYYMMDD);
            }
            if (scheduleEndYYYYMMDD != null) {
                qc.scheduledTimeEnd = DateUtils.toDate(scheduleEndYYYYMMDD);
            }

            List<MailScheduleProp> props = MailSchedule.query(client, qc, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).object(props);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "createMailSequence", httpMethod = ApiMethod.HttpMethod.POST)
    public APIResponse createMailSequence(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("mailContentIds") List<Long> mailContentIds,
                                            @Named("dayOffsets") List<Integer> dayOffsets,
                                            @Named("displayName") String displayName,
                                            @Named("groupId") long groupId,
                                            @Named("programTypeId") long programTypeId,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailSequenceProp input = new MailSequenceProp();
            input.groupId = groupId;
            input.mailContentIds = mailContentIds;
            input.dayOffsets = dayOffsets;
            input.programTypeId = programTypeId;
            input.displayName = displayName;

            MailSequenceProp prop = MailSequence.create(client, input, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).object(prop);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "deleteMailSequence", httpMethod = ApiMethod.HttpMethod.DELETE)
    public APIResponse deleteMailSequence(@Named("client") ClientApi.ClientEnum clientEnum,
                                          @Nullable @Named("clientIfOther") String clientOther,
                                          @Named("mailSequenceId") long mailSequenceId,
                                          @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                          HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            MailSequence.delete(client, mailSequenceId, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).message("Mail sequence [" + mailSequenceId + "] deleted");

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }

    @ApiMethod(path = "triggerNurtureEMails", httpMethod = ApiMethod.HttpMethod.POST)
    public APIResponse triggerNurtureEMails(@Named("client") ClientApi.ClientEnum clientEnum,
                                            @Nullable @Named("clientIfOther") String clientOther,
                                            @Named("programId") long programId,
                                            @Named("listId") long listId,
                                            @Named("senderEmail") String senderEmail,
                                            @Nullable @Named("firstNameIfMissingDefaultIsha") String defaultFirstName,
                                            @Nullable @Named("lastNameIfMissingDefaultIsha") String defaultLastName,
                                            @Nullable @Named("showStackTrace") Boolean showStackTrace,
                                            HttpServletRequest req, User user) {

        String login = null;
        String client = null;

        try {
            client = EndpointUtils.getClient(clientEnum, clientOther);
            login = Utils.getLoginEmail(user);

            if (defaultFirstName == null) {
                defaultFirstName = "Isha";
            }
            if (defaultLastName == null) {
                defaultLastName = "Isha";
            }

            ProgramProp programProp = Program.safeGet(client, programId).toProp(client);

            MailSequenceQueryCondition qc = new MailSequenceQueryCondition();
            qc.groupId = programProp.groupProp.groupId;
            qc.programTypeId = programProp.programTypeProp.programTypeId;
            List<MailSequenceProp> mailSequenceProps = MailSequence.query(client, qc, login);
            ensure(mailSequenceProps.size() == 1,
                    "This api method expects exactly 1 MailSequence for the specified groupId and programTypeId. Found ["
                    + mailSequenceProps.size() + "]");

            long mailSequenceId = mailSequenceProps.get(0).mailSequenceId;
            Date nowPlus2Min = new Date(new Date().getTime() + 2 * 60 * 1000);
            List<MailScheduleProp> mailScheduleProps = MailSequence.trigger(client, mailSequenceId, programId, listId, defaultFirstName,
                    defaultLastName, nowPlus2Min, senderEmail, login);

            return new APIResponse().status(APIResponse.Status.SUCCESS).message("Nurture campaign triggered").object(mailScheduleProps);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().req(req).login(login)
                    .client(client));
        }
    }
}
