package crmdna.api.endpoint;

import com.google.api.server.spi.config.Api;
import com.google.api.server.spi.config.ApiMethod;
import com.google.api.server.spi.config.ApiMethod.HttpMethod;
import com.google.appengine.api.users.User;
import com.google.gdata.data.spreadsheet.ListEntry;
import com.google.gdata.data.spreadsheet.ListFeed;
import crmdna.api.endpoint.ClientApi.ClientEnum;
import crmdna.api.endpoint.ProgramIshaApi.GroupEnum;
import crmdna.client.Client;
import crmdna.common.DateUtils;
import crmdna.common.DateUtils.DateRange;
import crmdna.common.Utils;
import crmdna.common.Utils.Currency;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.gspreadsheet.GSpreadSheet;
import crmdna.payment.Receipt;
import crmdna.payment.ReceiptProp;
import crmdna.payment2.Payment;
import crmdna.payment2.Payment.PaymentType;
import crmdna.payment2.PaymentEntity;
import crmdna.payment2.PaymentProp;
import crmdna.payment2.PaymentQueryCondition;

import javax.annotation.Nullable;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;

@Api(name = "payment")
public class PaymentApi {
    @ApiMethod(path = "recordCashOrChequePayment", httpMethod = HttpMethod.POST)
    public APIResponse recordCashOrChequePayment(@Named("client") ClientEnum clientEnum,
                                                 @Named("amount") double amount, @Nullable @Named("currencyDefaultSGD") Currency currency,
                                                 @Nullable @Named("paymentTypeDefaultCash") PaymentType paymentType,
                                                 @Nullable @Named("collectedByDefaultLoggedInUser") String collectedBy,
                                                 @Nullable @Named("chequeNo") String chequeNo, @Nullable @Named("bank") String bank,
                                                 @Nullable @Named("tags") Set<String> tags,
                                                 @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String client =
                Utils.removeSpaceUnderscoreBracketAndHyphen(clientEnum.toString().toLowerCase());

        if (currency == null)
            currency = Currency.SGD;

        if (paymentType == null)
            paymentType = PaymentType.CASH;

        String login = null;

        try {
            Client.ensureValid(client);

            login = Utils.getLoginEmail(user);

            if (collectedBy == null)
                collectedBy = login;

            PaymentProp paymentProp = new PaymentProp();
            paymentProp.amount = amount;
            paymentProp.currency = currency;
            paymentProp.bank = bank;
            paymentProp.paymentType = paymentType;
            paymentProp.collectedBy = collectedBy;
            paymentProp.chequeNo = chequeNo;
            paymentProp.bank = bank;
            paymentProp.date = new Date();
            paymentProp.transactionId = "MANUAL";
            paymentProp.tags = tags;

            PaymentProp prop = Payment.recordPayment(client, paymentProp, null, login);

            return new APIResponse().status(Status.SUCCESS).object(prop);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "query", httpMethod = HttpMethod.POST)
    public APIResponse query(@Named("client") ClientEnum clientEnum,
                             @Nullable @Named("currency") Currency currency,
                             @Nullable @Named("paymentType") PaymentType paymentType,
                             @Nullable @Named("collectedByEmail") String collectedBy,
                             @Nullable @Named("chequeNo") String chequeNo,
                             @Nullable @Named("dateRange") DateRange dateRange, @Nullable @Named("tags") Set<String> tags,
                             @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String client =
                Utils.removeSpaceUnderscoreBracketAndHyphen(clientEnum.toString().toLowerCase());

        String login = null;

        try {
            Client.ensureValid(client);

            login = Utils.getLoginEmail(user);

            PaymentQueryCondition qc = new PaymentQueryCondition();
            qc.chequeNo = chequeNo;
            qc.collectedBy = collectedBy;
            qc.currency = currency;
            qc.paymentType = paymentType;

            if (dateRange != null) {
                qc.endDate = new Date();
                long startDateMS = qc.endDate.getTime() - DateUtils.getMilliSecondsFromDateRange(dateRange);
                qc.startDate = new Date(startDateMS);
            }

            qc.tags = tags;

            List<PaymentEntity> entities = Payment.query(client, qc, login);

            List<PaymentProp> props = new ArrayList<>();

            for (PaymentEntity entity : entities) {
                props.add(entity.toProp());
            }

            return new APIResponse().status(Status.SUCCESS).object(props);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                    .login(login));
        }
    }

    @ApiMethod(path = "generateReceiptForRegId", httpMethod = HttpMethod.POST)
    public APIResponse generateReceiptForRegId(@Named("client") ClientEnum clientEnum,
        @Named("group") GroupEnum groupEnum,
        @Named("registrationId") long registrationId,
        @Named("sendEmail") boolean sendEmail,
        @Nullable @Named("clientOther") String clientOther,
        @Nullable @Named("groupOtherIdOrName") String groupOtherIdOrName,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String client = EndpointUtils.getClient(clientEnum, clientOther);
        long groupId;
        String login = null;

        try {

            groupId = EndpointUtils.getGroupId(client, groupEnum, groupOtherIdOrName);
            login = Utils.getLoginEmail(user);

            ReceiptProp receiptProp = Receipt.generateForRegistration(client, groupId,
                registrationId, sendEmail);

            return new APIResponse().status(Status.SUCCESS).object(receiptProp);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                .login(login));
        }
    }

    @ApiMethod(path = "generateAdhocReceiptsFromSpreadsheet", httpMethod = HttpMethod.POST)
    public APIResponse generateAdhocReceiptsFromSpreadsheet(@Named("client") ClientEnum clientEnum,
        @Named("group") GroupEnum groupEnum,
        @Named("spreadsheetKey") String spreadsheetKey,
        @Named("startIndex") int startIndex,
        @Named("numRecords") int numRecords,
        @Named("sendEmail") boolean sendEmail,
        @Nullable @Named("clientOther") String clientOther,
        @Nullable @Named("groupOtherIdOrName") String groupOtherIdOrName,
        @Nullable @Named("showStackTrace") Boolean showStackTrace, HttpServletRequest req, User user) {

        String authString = req.getHeader("Authorization");
        if (authString == null) {
            return new APIResponse().status(Status.ERROR_LOGIN_REQUIRED);
        }

        String client = EndpointUtils.getClient(clientEnum, clientOther);
        String accessToken = authString.substring("Bearer ".length());
        String login = Utils.getLoginEmail(user);

        try {
            ListFeed listFeed = GSpreadSheet.getSpreadSheetListFeed(accessToken, spreadsheetKey);
            long groupId = EndpointUtils.getGroupId(client, groupEnum, groupOtherIdOrName);

            List<ListEntry> listEntries = listFeed.getEntries();
            for (int i = startIndex; i < listEntries.size(); i++) {

                if (i == (startIndex + numRecords)) {
                    break;
                }

                ListEntry row = listEntries.get(i);
                Receipt.generateForAdhocDonation(client, groupId,
                    row.getCustomElements().getValue("firstName"),
                    row.getCustomElements().getValue("lastName"),
                    row.getCustomElements().getValue("email"),
                    row.getCustomElements().getValue("adhocReference"),
                    Currency.valueOf(row.getCustomElements().getValue("ccy")),
                    Double.parseDouble(row.getCustomElements().getValue("amount")),
                    Integer.parseInt(row.getCustomElements().getValue("date")
                        .replaceAll("(\\d\\d)/(\\d\\d)/(\\d\\d\\d\\d)", "$3$2$1")),
                    sendEmail);
            }

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client(client).req(req)
                .login(login));
        }

    }
}
