package crmdna.client.isha.ieo;

import com.google.gson.Gson;

import crmdna.client.Client;
import crmdna.client.isha.IshaAPAC;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.contact.ContactProp;
import crmdna.email.EmailProp;
import crmdna.email.GAEEmail;
import crmdna.member.*;
import crmdna.user.User;
import org.apache.http.HttpEntity;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.methods.RequestBuilder;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;

import javax.mail.MessagingException;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.OfyService.ofy;
import static crmdna.member.MemberBulkSaver.memberBulkSaver;

public class IshaIEO {

    private static final Logger logger = Logger.getLogger("IshaIEO");

    public static List<IshaIEOProfileProp> readIEOProfiles(String countries, boolean all)
        throws IOException, URISyntaxException {

        List<IshaIEOProfileProp> profiles = new ArrayList<>();

        try (CloseableHttpClient httpclient = HttpClients.custom().setSSLHostnameVerifier(NoopHostnameVerifier.INSTANCE).build()) {

            HttpUriRequest login =
                RequestBuilder.post()
                    .setUri(new URI("https://api1.ishayoga.sg/getIEOProfiles"))
                    .addParameter("countries", countries.replace(" ", "%20"))
                    .addParameter("all", Boolean.toString(all))
                    .addParameter("apiKey", Client.getToolsApiKey("isha"))
                    .build();

            try (CloseableHttpResponse response = httpclient.execute(login)) {
                HttpEntity entity = response.getEntity();
                StatusLine statusLine = response.getStatusLine();

                if (statusLine.getStatusCode() == 200) {
                    String body = EntityUtils.toString(entity);
                    EntityUtils.consume(entity);

                    IshaIEOProfileResponse res = new Gson().fromJson(body, IshaIEOProfileResponse.class);
                    if (res == null || res.statusCode != APIResponse.Status.SUCCESS) {
                        logger.warning("readIEOProfiles: " + (res == null ? "null" : res.statusCode));
                    } else {
                        profiles = res.object;
                    }
                }
            }
        }

        return profiles;
    }

    public static IshaIEOProfileEntity getIEOProfileEntity(long profileId) {
        return ofy("isha").load().type(IshaIEOProfileEntity.class).id(profileId).now();
    }

    public static Map<Long, IshaIEOProfileEntity> getIEOProfileEntities(Set<Long> profileIds) {
        return ofy("isha").load().type(IshaIEOProfileEntity.class).ids(profileIds);
    }

    public static Map<String, IshaIEOProfileEntity> getIEOProfileEntitiesByEmails(List<String> emails) {
        List<IshaIEOProfileEntity> entities = ofy("isha").load().type(IshaIEOProfileEntity.class).filter("email in", emails).list();
        Map<String, IshaIEOProfileEntity> map = new HashMap<>();
        for (IshaIEOProfileEntity ieoProfileEntity : entities) {
            map.put(ieoProfileEntity.email, ieoProfileEntity);
        }
        return map;
    }

    public static IshaIEOProfileProp getProp(Long profileId) {
        IshaIEOProfileEntity entity = ofy("isha").load().type(IshaIEOProfileEntity.class).id(profileId).now();
        if (entity == null) {
            return null;
        }
        return entity.toProp();
    }

    public static IshaIEOProfileEntity getIEOProfileEntityByRegId(long registrationId) {
        List<IshaIEOProfileEntity> entities = ofy("isha").load().type(IshaIEOProfileEntity.class).filter("registrationId", registrationId).list();
        if (entities.size() == 0) {
            return null;
        }
        if (entities.size() > 1) {
            throw new APIException("Error getting IEO Profile for Reg ID " + registrationId + " (size " + entities.size() + ")")
                .status(Status.ERROR_OVERFLOW);
        }
        return entities.get(0);
    }

    public static List<IshaIEOProfileEntity> getIEOProfileEntityByRegIds(Collection<Long> registrationIds) {
        return ofy("isha").load().type(IshaIEOProfileEntity.class).filter("registrationId in", registrationIds).list();
    }

    static boolean update(IshaIEOProfileProp profile, IshaIEOProfileEntity entity, List<String> changes) {

        if (entity.classNumber != profile.classNumber) {
            changes.add("classNumber: " + entity.classNumber + " <- " + profile.classNumber);
            entity.classNumber = profile.classNumber;
        }

        if ((profile.startMs > 0) && (entity.startMs != profile.startMs)) {
            changes.add("startMs: " + entity.startMs + " <- " + profile.startMs);
            entity.startMs = profile.startMs;
        }

        if ((profile.endYYYYMMDD > 0) && (entity.endYYYYMMDD != profile.endYYYYMMDD)) {
            changes.add("endYYYYMMDD: " + entity.endYYYYMMDD + " <- " + profile.endYYYYMMDD);
            entity.endYYYYMMDD = profile.endYYYYMMDD;
        }

        if ((profile.completedYYYYMMDD > 0) && (entity.completedYYYYMMDD != profile.completedYYYYMMDD)) {
            changes.add("completedYYYYMMDD: " + entity.completedYYYYMMDD + " <- " + profile.completedYYYYMMDD);
            entity.completedYYYYMMDD = profile.completedYYYYMMDD;
        }

        if ((profile.paymentYYYYMMDD > 0) && (entity.paymentYYYYMMDD != profile.paymentYYYYMMDD)) {
            changes.add("paymentYYYYMMDD: " + entity.paymentYYYYMMDD + " <- " + profile.paymentYYYYMMDD);
            entity.paymentYYYYMMDD = profile.paymentYYYYMMDD;
        }

        if (Utils.isDifferent(entity.coupon, profile.coupon)) {
            changes.add("coupon: " + entity.coupon + " <- " + profile.coupon);
            entity.coupon = profile.coupon;
        }

        profile.email = Utils.sanitizeEmail(profile.email);
        if (Utils.isDifferent(entity.email, profile.email)) {
            changes.add("email: " + entity.email + " <- " + profile.email);
            entity.email = profile.email;
        }

        if (Utils.isDifferent(entity.firstName, profile.firstName)) {
            changes.add("firstName: " + entity.firstName + " <- " + profile.firstName);
            entity.firstName = profile.firstName;
        }

        if (Utils.isDifferent(entity.lastName, profile.lastName)) {
            changes.add("lastName: " + entity.lastName + " <- " + profile.lastName);
            entity.lastName = profile.lastName;
        }

        profile.homePhone = Utils.sanitizePhoneNo(profile.homePhone, null);
        if (Utils.isDifferent(entity.homePhone, profile.homePhone)) {
            changes.add("homePhone: " + entity.homePhone + " <- " + profile.homePhone);
            entity.homePhone = profile.homePhone;
        }

        profile.mobilePhone = Utils.sanitizePhoneNo(profile.mobilePhone, null);
        if (Utils.isDifferent(entity.mobilePhone, profile.mobilePhone)) {
            changes.add("mobilePhone: " + entity.mobilePhone + " <- " + profile.mobilePhone);
            entity.mobilePhone = profile.mobilePhone;
        }

        if (Utils.isDifferent(entity.gender, profile.gender)) {
            changes.add("gender: " + entity.gender + " <- " + profile.gender);
            entity.gender = profile.gender;
        }

        if (Utils.isDifferent(entity.country, profile.country)) {
            changes.add("country: " + entity.country + " <- " + profile.country);
            entity.country = profile.country;
        }

        if (Utils.isDifferent(entity.city, profile.city)) {
            changes.add("city: " + entity.city + " <- " + profile.city);
            entity.city = profile.city;
        }

        if (Utils.isDifferent(entity.postalCode, profile.postalCode)) {
            changes.add("postalCode: " + entity.postalCode + " <- " + profile.postalCode);
            entity.postalCode = profile.postalCode;
        }

        return (changes.size() > 0);
    }

    public static Map<Long, List<String>> createOrUpdateIEOProfiles(List<IshaIEOProfileProp> profiles,
        long groupId) {

        Map<Long, IshaIEOProfileProp> profilesMap = new HashMap<>();
        for (IshaIEOProfileProp profile : profiles) {
            profilesMap.put(profile.id, profile);
        }

        // Get the entities from datastore based on the profiles received from IEO system
        // The profiles from IEO system are the superset, so we might not have all in our system
        Map<Long, IshaIEOProfileEntity> entitiesMap = getIEOProfileEntities(profilesMap.keySet());
        Map<Long, List<String>> changesMap = new HashMap<>();
        List<Long> updatedProfileIds = new ArrayList<>();
        Map<String, Long> contactsMap = new HashMap<>();
        List<ContactProp> contactPropsList = new ArrayList<>();
        Map<Long, Long> updMemberIdvsIEOProfileIds = new HashMap<>();

        // Iterate of the profiles received
        for (IshaIEOProfileProp profile : profiles) {

            List<String> changes = new ArrayList<>();
            IshaIEOProfileEntity entity = entitiesMap.get(profile.id);

            // If there no such entity in our system, create it
            if (entity == null) {
                entity = new IshaIEOProfileEntity();
                entity.id = profile.id;
                entitiesMap.put(profile.id, entity);
            }

            // Capture the differences
            if (update(profile, entity, changes)) {
                updatedProfileIds.add(profile.id);
                changesMap.put(profile.id, changes);
            }

            // If corresponding memberId is not present, collate the contactProps
            // Later the member will be created
            if (entity.memberId == 0) {
                contactsMap.put(profile.email, profile.id);
                contactPropsList.add(profile.getContactProp());
            }
        }

        // Get the memberIds for the contacts list we have collated
        List<Long> memberIds = Member.getMatchingMemberIds("isha", contactPropsList);
        List<ContactProp> newContactPropsList = new ArrayList<>();
        for (int i = 0; i < memberIds.size(); i++) {
            ContactProp contactProp = contactPropsList.get(i);
            Long memberId = memberIds.get(i);
            // If the corresponding memberId doesnt exist, create it
            if (memberId == null) {
                newContactPropsList.add(contactProp);
            } else {
                entitiesMap.get(contactsMap.get(contactProp.email)).memberId = memberId;
                updMemberIdvsIEOProfileIds.put(memberId, entitiesMap.get(contactsMap.get(contactProp.email)).id);
            }
        }

        // Create the member entities in bulk and update the ids
        List<MemberEntity> newMemberEntities = MemberEntity.MemberFactory.create("isha", newContactPropsList.size());
        for (int i = 0; i < newMemberEntities.size(); i++) {
            Long profileId = contactsMap.get(newContactPropsList.get(i).email);
            newMemberEntities.get(i).ieoProfileId = profileId;
            entitiesMap.get(profileId).memberId = newMemberEntities.get(i).getId();
        }

        // Persist them in bulk
        try {
            memberBulkSaver("isha", newMemberEntities)
                .setContactsSameSizeList(newContactPropsList)
                .addGroupToAll(groupId)
                .populateDependantsAndSave();
        } catch (APIException ex) {
            System.out.println(ex.toString());
        }

        List<IshaIEOProfileEntity> updatedEntities = new ArrayList<>();
        for (Long profileId : updatedProfileIds) {
            updatedEntities.add(entitiesMap.get(profileId));
        }

        ofy("isha").save().entities(updatedEntities);

        Map<Long, MemberEntity> memberEntityMap = MemberLoader.get("isha", updMemberIdvsIEOProfileIds.keySet(), User.SUPER_USER);
        for (Map.Entry<Long, MemberEntity> entry : memberEntityMap.entrySet()) {
            entry.getValue().ieoProfileId = updMemberIdvsIEOProfileIds.get(entry.getKey());
        }
        memberBulkSaver("isha", new ArrayList<>(memberEntityMap.values())).populateDependantsAndSave();

        return changesMap;
    }

    public static void syncApacProfiles () throws IOException, APIException, MessagingException, URISyntaxException {
        Date startMS = new Date();
        Map<Long, List<String>> changesMap = new HashMap<>();
        for (IshaAPAC.CountryCfg cty : IshaAPAC.countryCfgs) {
            List<IshaIEOProfileProp> profiles = readIEOProfiles(cty.name + "," + cty.shortName, true);
            changesMap.putAll(createOrUpdateIEOProfiles(profiles, cty.groupId));
        }
        Date endMS = new Date();

        EmailProp emailProp = new EmailProp();
        emailProp.subject = "Sync Report - IEO APAC";
        StringBuilder sb = new StringBuilder();
        sb.append("Number of Changes: ")
            .append(changesMap.size())
            .append(" (")
            .append(endMS.getTime() - startMS.getTime())
            .append("ms)<br><br>");

        if (changesMap.size() <= 100) {
            for (Map.Entry<Long, List<String>> entry : changesMap.entrySet()) {
                sb.append(entry.getKey()).append(":<br>");
                for (String change : entry.getValue()) {
                    sb.append("&nbsp;&nbsp;&nbsp;&nbsp;").append(change).append("<br>");
                }
                sb.append("<br>");
            }
        }

        emailProp.bodyHtml = sb.toString();
        emailProp.toEmailAddresses.add("thulasidhar@gmail.com");
        emailProp.toEmailAddresses.add("sathya.t@ishafoundation.org");
        GAEEmail.send(emailProp);
    }

    public static void update(List<MemberProp> memberProps) {
        Map<Long, MemberProp> memberPropMap = new HashMap<>();
        for (MemberProp memberProp : memberProps) {
            if (memberProp.ieoProfileId != null) {
                memberPropMap.put(memberProp.ieoProfileId, memberProp);
            }
        }

        Map<Long, IshaIEOProfileEntity> ieoProfileEntityMap =
            getIEOProfileEntities(memberPropMap.keySet());

        for (Map.Entry<Long, IshaIEOProfileEntity> entry : ieoProfileEntityMap.entrySet()) {
            MemberProp memberProp = memberPropMap.get(entry.getKey());
            IshaIEOProfileEntity entity = entry.getValue();

            memberProp.ieoClassNumber = entity.classNumber;
            memberProp.ieoCompletedYYYYMMDD = entity.completedYYYYMMDD;
        }
    }

}
