package crmdna.client.isha.ieo;

import com.google.api.server.spi.config.Api;
import com.google.api.server.spi.config.ApiMethod;
import com.google.api.server.spi.config.ApiMethod.HttpMethod;
import com.google.appengine.api.users.User;
import com.googlecode.objectify.Key;
import com.googlecode.objectify.cmd.QueryKeys;
import crmdna.api.endpoint.EndpointUtils;
import crmdna.api.endpoint.ProgramIshaApi.GroupEnum;
import crmdna.calling.Campaign;
import crmdna.common.AssertUtils;
import crmdna.common.DateUtils;
import crmdna.common.StopWatch;
import crmdna.common.Utils;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.gspreadsheet.GSpreadSheet;
import crmdna.interaction.CallAssignmentHelper;
import crmdna.interaction.CallAssignmentResult;
import crmdna.interaction.Calling;
import crmdna.interaction.Interaction;
import crmdna.interaction.InteractionProp;
import crmdna.interaction.InteractionQueryCondition;
import crmdna.registration.RegistrationEntity;
import crmdna.user.UserEntity;
import crmdna.user.UserProp;

import javax.annotation.Nullable;
import javax.inject.Named;
import javax.servlet.http.HttpServletRequest;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import static crmdna.common.OfyService.ofy;

@Api(name = "isha")
public class IshaIEOApi {

    private final static Logger LOGGER = Logger.getLogger(IshaIEOApi.class.getName());

    @ApiMethod(path = "createSMMProgramMap", httpMethod = HttpMethod.GET)
    public APIResponse createSMMProgramMap(
        @Named("SMMProgramId") long smmProgramId,
        @Named("programId") long programId,
        @Named("comments") String comments,
        @Named("groupEnum") GroupEnum groupEnum,
        @Nullable @Named("groupOtherIdOrName") String groupOtherIdOrName,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;
        try {

            login = Utils.getLoginEmail(user);
            IshaSMM.createProgramMap(smmProgramId, programId,
                EndpointUtils.getGroupId("isha", groupEnum, groupOtherIdOrName), comments, login);

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client("isha").req(req)
                .login(login));
        }
    }

    @ApiMethod(path = "syncSMMRegistrations", httpMethod = HttpMethod.GET)
    public APIResponse syncSMMRegistrations(@Named("programId") long programId,
        @Named("groupEnum") GroupEnum groupEnum,
        @Nullable @Named("groupOtherIdOrName") String groupOtherIdOrName,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;

        try {
            login = Utils.getLoginEmail(user);
            IshaSMM.getRegistrations(programId, login);
            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(path = "syncApacIEOProfiles", httpMethod = HttpMethod.GET)
    public APIResponse syncApacIEOProfiles(
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;

        try {
            login = Utils.getLoginEmail(user);
            IshaIEO.syncApacProfiles();
            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(path = "assignCallsForUser", httpMethod = HttpMethod.GET)
    public APIResponse assignCallsForUser(
        @Named("campaignId") long campaignId,
        @Named("userEmail") String userEmail,
        @Named("memberEmailsCSV") String memberEmailsCSV,
        @Nullable @Named("previewDefaultFalse") Boolean preview,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;
        String client = "isha";

        try {
            login = Utils.getLoginEmail(user);

            crmdna.user.User.ensureValidUser(client, userEmail);
            long userId = crmdna.user.User.safeGet(client, userEmail).toProp(client).userId;
            Set<Long> userIds = Utils.getSet(userId);

            List<String> memberEmails = Arrays.asList(memberEmailsCSV.split("[,\\s]+"));
            Map<String, IshaIEOProfileEntity> mapEmailVsProfileEntity = IshaIEO.getIEOProfileEntitiesByEmails(memberEmails);

            AssertUtils.ensure(memberEmails.size() == mapEmailVsProfileEntity.size(),
                "Emails(" + memberEmails.size() + ") vs Entities(" + mapEmailVsProfileEntity.size() + ") do not match!");

            Set<Long> memberIds = new HashSet<>();
            for (IshaIEOProfileEntity profileEntity : mapEmailVsProfileEntity.values()) {
                AssertUtils.ensureNotNull(profileEntity.memberId, "Member ID not found for " + profileEntity.email);
                memberIds.add(profileEntity.memberId);
            }

            CallAssignmentResult result;
            if (preview == null) preview = false;

            Map<Long, Integer> userIdVsMaxMembers = CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());

            if (preview) {
                result = Calling.preview(client, campaignId, memberIds, userIds, userIdVsMaxMembers, true, login);
            } else {
                result = Calling.assign(client, campaignId, memberIds, userIds, userIdVsMaxMembers, true, login);
            }

            return new APIResponse().status(Status.SUCCESS).object(result);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(path = "purgeIEOEntities", httpMethod = HttpMethod.GET)
    public APIResponse purgeIEOEntities(
        @Nullable @Named("smmProgramId") Long smmProgramId,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;

        try {
            login = Utils.getLoginEmail(user);

            QueryKeys<IshaIEOProfileEntity> ieoQueryKeys = ofy("isha").load().type(IshaIEOProfileEntity.class).keys();
            List<Key<IshaIEOProfileEntity>> ieoKeys = ieoQueryKeys.list();
            ofy("isha").delete().keys(ieoKeys);

            if (smmProgramId != null) {
                QueryKeys<RegistrationEntity> smmQueryKeys =
                    ofy("isha").load().type(RegistrationEntity.class)
                        .filter("programId", smmProgramId).keys();
                List<Key<RegistrationEntity>> smmKeys = smmQueryKeys.list();
                ofy("isha").delete().keys(smmKeys);

            }

            return new APIResponse().status(Status.SUCCESS);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace,
                new RequestInfo().client("isha").login(login).req(req));
        }
    }

    @ApiMethod(path = "importInteractionsForCampaign", httpMethod = HttpMethod.GET)
    public APIResponse importInteractionsForCampaign(
        @Named("campaignId") long campaignId,
        @Named("spreadsheetKey") String spreadsheetKey,
        @Nullable @Named("showStackTrace") Boolean showStackTrace,
        HttpServletRequest req, User user) {

        String login = null;

        try {
            login = Utils.getLoginEmail(user);
            StopWatch sw = StopWatch.createStarted();
            int created = 0, skipped = 0;

            List<Map<String, String>> listOfMap = GSpreadSheet.getPublishedSpreadsheetAsListOfMap(spreadsheetKey, 2500);

            Campaign.safeGet("isha", campaignId);

            List<String> memberEmails = new ArrayList<>();
            int lineNumber = 2;
            for (Map<String, String> map : listOfMap) {
                String memberEmail = map.get("member");
                AssertUtils.ensureNotNullNotEmpty(memberEmail, lineNumber + ": member email missing");
                memberEmails.add(memberEmail.toLowerCase().trim());
                lineNumber++;
            }
            Map<String, IshaIEOProfileEntity> mapEmailVsProfileEntity = IshaIEO.getIEOProfileEntitiesByEmails(memberEmails);

            lineNumber = 2;
            for (Map<String, String> map : listOfMap) {

                String userEmail = map.get("user");
                AssertUtils.ensureNotNull(userEmail, lineNumber + ": user column missing");
                userEmail = userEmail.trim().toLowerCase();

                String memberEmail = map.get("member").toLowerCase().trim();
                AssertUtils.ensureNotNullNotEmpty(memberEmail, lineNumber + ": member email missing");
                memberEmail = memberEmail.toLowerCase().trim();

                String interaction = map.get("interaction");
                AssertUtils.ensureNotNull(interaction, lineNumber + ": interaction column missing");

                String dateYYYYMMDDStr = map.get("date");
                AssertUtils.ensureNotNull(dateYYYYMMDDStr, lineNumber + ": date column missing");
                AssertUtils.ensure(Utils.canParseAsLong(dateYYYYMMDDStr));
                int dateYYYYMMDD = Integer.parseInt(dateYYYYMMDDStr);
                DateUtils.ensureFormatYYYYMMDD(dateYYYYMMDD);

                String responseStr = map.get("response");
                AssertUtils.ensureNotNull(responseStr, lineNumber + ": response column missing");
                Interaction.Response response = Interaction.Response.valueOf(responseStr.toUpperCase().trim());

                String progressStr = map.get("progress");
                AssertUtils.ensureNotNull(progressStr, lineNumber + ": progress column missing");
                Interaction.Progress progress = Interaction.Progress.valueOf(progressStr.toUpperCase().trim());

                InteractionQueryCondition qc = new InteractionQueryCondition();
                qc.campaignIds.add(campaignId);

                UserEntity userEntity = crmdna.user.User.get("isha", userEmail);
                AssertUtils.ensureNotNull(userEntity, lineNumber + ": user " + userEmail + " not found");
                qc.userIds.add(userEntity.toProp("isha").userId);
                UserProp userProp = userEntity.toProp("isha");

                IshaIEOProfileEntity ieoProfileEntity = mapEmailVsProfileEntity.get(memberEmail);
                AssertUtils.ensureNotNull(ieoProfileEntity, lineNumber + ": member " + memberEmail + "'s profile missing");

                IshaIEOProfileProp ieoProfileProp = ieoProfileEntity.toProp();
                qc.memberIds.add(ieoProfileProp.memberId);
                qc.interactionTypes.add(Interaction.InteractionType.PHONE.toString());

                List<InteractionProp> interactionProps = Interaction.query("isha", qc, crmdna.user.User.SUPER_USER).interactionProps;
                if (interactionProps.size() == 0) {
                    InteractionProp interactionProp = Interaction.createInteraction(
                        "isha",  ieoProfileProp.memberId, interaction, Interaction.InteractionType.PHONE,
                        DateUtils.toDate(dateYYYYMMDD), campaignId, true, userProp.email);
                    Interaction.updateInteraction("isha", interactionProp.interactionId, null, null, null, progress, response, login);
                    LOGGER.info(lineNumber + ": created interaction");
                    created++;
                } else {
                    LOGGER.info(lineNumber + ": already created interaction");
                    skipped++;
                }

                lineNumber++;
                if (sw.msElapsed() > 56000) {
                    break;
                }
            }

            return new APIResponse().status(Status.SUCCESS).message("Imported " + created + " interactions. " + "Skipped " + skipped);

        } catch (Exception ex) {
            return APIUtils.toAPIResponse(ex, showStackTrace, new RequestInfo().client("isha").login(login).req(req));
        }
    }
}
