package crmdna.client.isha.ieo;

import com.google.gson.Gson;
import com.googlecode.objectify.ObjectifyFilter;
import crmdna.client.Client;
import crmdna.client.isha.IshaAPAC;
import crmdna.common.DateUtils;
import crmdna.common.StopWatch;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.api.APIUtils;
import crmdna.common.api.RequestInfo;
import crmdna.common.contact.Contact;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.registration.Registration;
import crmdna.registration.RegistrationEntity;
import crmdna.registration.RegistrationInput;
import crmdna.registration.RegistrationProp;
import crmdna.registration.RegistrationQueryCondition;
import crmdna.user.User;
import org.apache.http.HttpEntity;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.methods.RequestBuilder;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.ensure;
import static crmdna.common.AssertUtils.ensureNotNull;
import static crmdna.common.OfyService.ofy;

public class IshaSMM {

    private static final Logger logger = Logger.getLogger("IshaSMM");

    public static IshaSMMProgramEntity getEntity(long smmProgramId) {
        return ofy("isha").load().type(IshaSMMProgramEntity.class).id(smmProgramId).now();
    }

    public static Map<Long,IshaSMMProgramEntity> getEntityMap(long programId) {
        List<IshaSMMProgramEntity> entities = ofy("isha").load().type(IshaSMMProgramEntity.class).filter("programId", programId).list();
        Map<Long,IshaSMMProgramEntity> programEntityMap = new HashMap<>();
        for (IshaSMMProgramEntity entity : entities) {
            programEntityMap.put(entity.id, entity);
        }
        return programEntityMap;
    }

    public static List<IshaIEOProfileProp> readRegistrations(long smmProgramId)
        throws IOException, URISyntaxException {

        List<IshaIEOProfileProp> profiles = new ArrayList<>();
        StopWatch sw = StopWatch.createStarted();

        try (CloseableHttpClient httpclient = HttpClients.custom().setSSLHostnameVerifier(NoopHostnameVerifier.INSTANCE).build()) {

            HttpUriRequest login =
                RequestBuilder.post()
                    .setUri(new URI("https://api1.ishayoga.sg/getSMMRegistrations"))
                    .addParameter("programId", Long.toString(smmProgramId))
                    .addParameter("apiKey", Client.getToolsApiKey("isha"))
                    .build();

            try (CloseableHttpResponse response = httpclient.execute(login)) {
                HttpEntity entity = response.getEntity();
                StatusLine statusLine = response.getStatusLine();

                if (statusLine.getStatusCode() == 200) {
                    String body = EntityUtils.toString(entity);
                    EntityUtils.consume(entity);

                    IshaIEOProfileResponse res = new Gson().fromJson(body, IshaIEOProfileResponse.class);
                    if (res == null || res.statusCode != APIResponse.Status.SUCCESS) {
                        logger.warning("readRegistrations: " + (res == null ? "null" : res.statusCode));
                    } else {
                        profiles = res.object;
                    }
                }
            }
        }

        logger.info("getSMMRegistrations for PID :" + smmProgramId + " (" + profiles.size() + " profiles) " + sw.msElapsed() + "ms");

        return profiles;
    }

    public static void createProgramMap(long smmProgramId, long programId, long groupId, String comments,
        String login) {

        User.ensureGroupLevelPrivilege("isha", groupId, login,
            User.GroupLevelPrivilege.UPDATE_PROGRAM);

        IshaSMMProgramEntity entity = getEntity(smmProgramId);
        ensure(entity == null, "IshaSMMProgramEntity:" + smmProgramId + " already exists");

        entity = new IshaSMMProgramEntity();
        entity.id = smmProgramId;
        entity.programId = programId;
        entity.comments = comments;

        ofy("isha").save().entity(entity).now();
    }

    private static List<RegistrationProp> getSMMRegistrations(long smmProgramId, long programId, long groupId) {

        Map<Long, IshaIEOProfileProp> registeredProfilesMap = new HashMap<>();

        try {
            List<IshaIEOProfileProp> registrations = readRegistrations(smmProgramId);
            for (IshaIEOProfileProp profile : registrations) {
                registeredProfilesMap.put(profile.id, profile);
            }
        } catch (Exception ex) {
            APIUtils.toAPIResponse(ex, true, new RequestInfo().client(null));
            return new ArrayList<>();
        }

        Map<Long, IshaIEOProfileEntity> registeredEntitiesMap = IshaIEO.getIEOProfileEntities(registeredProfilesMap.keySet());
        Set<Long> updatedProfileIds = new HashSet<>();

        /*
        * Sometimes the SMM registrations are from countries outside APAC, so their entities are not
        * captured in the system. In this case, we need to create the corresponding entities here.
        */
        if (registeredEntitiesMap.size() < registeredProfilesMap.size()) {
            Set<Long> profileIds = new HashSet<>(registeredProfilesMap.keySet());
            Set<Long> entityIds = new HashSet<>(registeredEntitiesMap.keySet());
            profileIds.removeAll(entityIds);

            List<IshaIEOProfileProp> missingProfiles = new ArrayList<>();
            for (long missingProfileId : profileIds) {
                IshaIEOProfileProp prop = registeredProfilesMap.get(missingProfileId);
                missingProfiles.add(prop);
            }

            IshaIEO.createOrUpdateIEOProfiles(missingProfiles, groupId);
            ObjectifyFilter.complete();
            registeredEntitiesMap.putAll(IshaIEO.getIEOProfileEntities(profileIds));
        }

        List<Long> registrationIds = new ArrayList<>();
        for (Map.Entry<Long, IshaIEOProfileEntity> entry : registeredEntitiesMap.entrySet()) {
            IshaIEOProfileEntity entity = entry.getValue();
            boolean addToUpdated = false;
            if (entity.registrationId == null) {
                addToUpdated = true;
            } else {
                registrationIds.add(entity.registrationId);
            }

            IshaIEOProfileProp registeredProfile = registeredProfilesMap.get(entity.id);
            if (!IshaAPAC.isCountryPresent(entity.country)) {
                List<String> changes = new ArrayList<>();
                if (IshaIEO.update(registeredProfile, entity, changes)) {
                    addToUpdated = true;
                }
            }

            if ((entity.smmProgramId == null) || (entity.smmProgramId != smmProgramId)) {
                entity.smmProgramId = smmProgramId;
                addToUpdated = true;
            }

            if ((entity.packageReg == null) || (entity.packageReg != registeredProfile.packageReg)) {
                entity.packageReg = registeredProfile.packageReg;
                addToUpdated = true;
            }

            if (addToUpdated) {
                updatedProfileIds.add(entity.id);
            }
        }

        Map<Long, RegistrationEntity> registrationMap = Registration.safeGet("isha", registrationIds);

        List<RegistrationProp> updRegistrationProps = new ArrayList<>();
        List<RegistrationProp> newRegistrationProps = new ArrayList<>();
        List<RegistrationEntity> updatedRegistrationEntities = new ArrayList<>();
        for (Map.Entry<Long, IshaIEOProfileEntity> entry : registeredEntitiesMap.entrySet()) {

            IshaIEOProfileEntity entity = entry.getValue();
            RegistrationProp registrationProp;

            if (entity.registrationId == null) {
                IshaIEOProfileProp profile = registeredProfilesMap.get(entry.getKey());

                try {
                    registrationProp = Registration.register("isha",
                        new RegistrationInput(entity.getContactProp(), programId, profile.amount,
                            (profile.paymentReference == null) ? "IEO-Payment" : profile.paymentReference));
                    entity.registrationId = registrationProp.registrationId;
                    newRegistrationProps.add(registrationProp);
                } catch (Exception ex) {
                    APIUtils.toAPIResponse(ex, true, new RequestInfo().client(null));
                }
            } else {
                RegistrationEntity registrationEntity = registrationMap.get(entry.getValue().registrationId);
                registrationProp = registrationEntity.toProp();
                if (update(registrationProp, entity.toProp())) {
                    updatedRegistrationEntities.add(registrationEntity.update(registrationProp));
                }
                updRegistrationProps.add(registrationProp);
            }
        }

        List<IshaIEOProfileEntity> updatedEntities = new ArrayList<>();
        for (Long profileId : updatedProfileIds) {
            updatedEntities.add(registeredEntitiesMap.get(profileId));
        }
        ofy("isha").save().entities(updatedEntities).now();
        ofy("isha").save().entities(updatedRegistrationEntities).now();

        try {
            if (newRegistrationProps.size() > 0) {
                Registration.sendConfirmationEmail("isha", newRegistrationProps, true);
            }
        } catch (Exception ex) {
            APIUtils.toAPIResponse(ex, true, new RequestInfo().client(null));
        }

        updRegistrationProps.addAll(newRegistrationProps);
        return updRegistrationProps;
    }

    public static List<RegistrationProp> getRegistrations(long programId, String login)
        throws IOException, APIException {

        User.ensureValidUser("isha", login);

        ProgramProp programProp = Program.safeGet("isha", programId).toProp("isha");

        Map<Long,IshaSMMProgramEntity> smmProgramEntities = getEntityMap(programId);

        ensureNotNull(smmProgramEntities, "SMMProgramEntity not created for ProgramId " + programId);
        ensure(smmProgramEntities.size() > 0,
            "SMMProgramEntity not created for ProgramId " + programId);

        List<RegistrationProp> registrationProps = new ArrayList<>();
        for (Long smmProgramId : smmProgramEntities.keySet()) {
            registrationProps.addAll(
                getSMMRegistrations(smmProgramId, programId, programProp.groupProp.groupId));
        }

        return registrationProps;
    }

    public static Object update(ProgramProp programProp, String login) {
        IshaSMMRegistrationStats registrationStatus = new IshaSMMRegistrationStats();
        RegistrationQueryCondition qc = new RegistrationQueryCondition();
        Map<Long,IshaSMMRegistrationStats.Category> categoryMap = new HashMap<>();

        if (!programProp.programTypeProp.displayName.contains("Shambhavi")) {
            return null;
        }

        Map<Long,IshaSMMProgramEntity> programEntityMap = getEntityMap(programProp.programId);
        for (IshaSMMProgramEntity programEntity : programEntityMap.values()) {
            IshaSMMRegistrationStats.Category category = new IshaSMMRegistrationStats.Category(programEntity.comments);
            categoryMap.put(programEntity.id, category);
            registrationStatus.categories.add(category);
        }

        qc.programId = programProp.programId;
        List<RegistrationProp> props = Registration.query("isha", qc, User.SUPER_USER);
        for (RegistrationProp prop : props) {

            if (prop.status != Registration.RegistrationStatus.REGISTRATION_COMPLETE) {
                continue;
            }

            IshaIEOProfileEntity entity = IshaIEO.getIEOProfileEntityByRegId(prop.registrationId);
            ensureNotNull(entity, "Unable to get entity for RID " + prop.registrationId);
            IshaSMMRegistrationStats.Category category = categoryMap.get(entity.smmProgramId);
            Integer count;

            //Build Class-wise Data
            String className = (entity.completedYYYYMMDD == 0) ? ("Class " + entity.classNumber) : "Completed";
            count = category.classWiseMap.get(className);
            if (count == null) count = 0;
            category.classWiseMap.put(className, count + 1);

            //Build Country-wise Data
            String country = IshaAPAC.getCountryCode(entity.country);
            count = category.countryWiseMap.get(country);
            if (count == null) count = 0;
            category.countryWiseMap.put(country, count + 1);

            //Build Gender-wise Data
            String gender = entity.gender.toLowerCase();
            count = category.genderWiseMap.get(gender);
            if (count == null) count = 0;
            category.genderWiseMap.put(gender, count + 1);

            String date = DateUtils.toISOString(prop.getRegisteredDate(), true);
            count = category.dateWiseMap.get(date);
            if (count == null) count = 0;
            category.dateWiseMap.put(date, count + 1);

            //Build Package-wise Data
            boolean pkg = entity.packageReg;
            if (pkg) {
                count = registrationStatus.dateWisePkgMap.get(date);
                if (count == null) count = 0;
                registrationStatus.dateWisePkgMap.put(date, count + 1);
            } else {
                count = registrationStatus.dateWiseSmmMap.get(date);
                if (count == null) count = 0;
                registrationStatus.dateWiseSmmMap.put(date, count + 1);
            }

            category.total++;

            if (User.hasPrivilege("isha", login, User.ClientLevelPrivilege.VIEW_FINANCIALS) && (prop.amount != null)) {
                registrationStatus.amtCollected += Double.parseDouble(prop.amount);
            }
        }

        return registrationStatus;
    }

    public static boolean update(RegistrationProp registrationProp, IshaIEOProfileProp profileProp) {

        boolean updated = false;

        if ((profileProp.firstName != null) && !profileProp.firstName.equals(registrationProp.firstName)) {
            updated = true;
            registrationProp.firstName = profileProp.firstName;
        }

        if ((profileProp.lastName != null) && !profileProp.lastName.equals(registrationProp.lastName)) {
            updated = true;
            registrationProp.lastName = profileProp.lastName;
        }

        if ((profileProp.email != null) && !profileProp.email.equals(registrationProp.email)) {
            updated = true;
            registrationProp.email = profileProp.email;
        }

        String gender = (registrationProp.gender != null) ? registrationProp.gender.toString() : null;
        if ((profileProp.gender != null) && !profileProp.gender.equals(gender)) {
            updated = true;
            registrationProp.gender = Contact.getGender(gender);
        }

        if ((profileProp.mobilePhone != null) && !profileProp.mobilePhone.equals(registrationProp.mobilePhone)) {
            updated = true;
            registrationProp.mobilePhone = profileProp.mobilePhone;
        }

        if ((profileProp.homePhone != null) && !profileProp.homePhone.equals(registrationProp.homePhone)) {
            updated = true;
            registrationProp.homePhone = profileProp.homePhone;
        }

        if ((profileProp.country != null) && !profileProp.country.equals(registrationProp.country)) {
            updated = true;
            registrationProp.country = profileProp.country;
        }

        if ((profileProp.city != null) && !profileProp.city.equals(registrationProp.city)) {
            updated = true;
            registrationProp.city = profileProp.city;
        }

        if ((profileProp.postalCode != null) && !profileProp.postalCode.equals(registrationProp.postalCode)) {
            updated = true;
            registrationProp.postalCode = profileProp.postalCode;
        }

        return updated;
    }
}
