package crmdna.gspreadsheet;

import com.google.api.client.googleapis.auth.oauth2.GoogleCredential;
import com.google.gson.*;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse.Status;

import com.google.gdata.client.spreadsheet.*;
import com.google.gdata.data.spreadsheet.*;
import com.google.gdata.util.*;

import java.io.IOException;
import java.net.*;
import java.util.*;

import java.util.Map.Entry;

public class GSpreadSheet {

    final static int MAX_LINES = 25000;

    public static List<Map<String, String>> getPublishedSpreadsheetAsListOfMap(
        String gsKey, int numLinesExclHeader) throws IOException {

        return getPublishedSpreadsheetAsListOfMap(gsKey, numLinesExclHeader, 0);
    }

    public static List<Map<String, String>> getPublishedSpreadsheetAsListOfMap(
        String gsKey, int numLinesExclHeader, int startIndex) throws IOException {

        if (numLinesExclHeader > MAX_LINES)
            throw new APIException().status(Status.ERROR_RESOURCE_INCORRECT).message(
                "Num lines to read (excl. header) [" + numLinesExclHeader + "] cannot be greater than ["
                    + MAX_LINES + "]");

        String jsonFeedURL = getJSONFeedURL(gsKey);
        String json = Utils.readDataFromURL(jsonFeedURL);

        return getListOfMap(json, numLinesExclHeader, startIndex);
    }

    static List<Map<String, String>> getListOfMap(String json, int maxLines, int startIndex) {

        Set<String> keysToIgnore = new HashSet<>();
        keysToIgnore.add("id");
        keysToIgnore.add("updated");
        keysToIgnore.add("category");
        keysToIgnore.add("title");
        keysToIgnore.add("content");
        keysToIgnore.add("link");

        JsonParser parser = new JsonParser();

        JsonObject jsonObject;

        try {
            jsonObject = parser.parse(json).getAsJsonObject().getAsJsonObject("feed");
        } catch (JsonSyntaxException jse) {
            throw new APIException().status(Status.ERROR_RESOURCE_INCORRECT).message(
                    "Unable to read spreadsheet data. "
                            + "Please check if spreadhsheet key is valid and the spreadsheet is published.");
        }

        if (null == jsonObject) {
            throw new APIException().status(Status.ERROR_RESOURCE_INCORRECT).message(
                    "Invalid json. Cannot find element [feed]");
        }

        JsonArray rows = jsonObject.getAsJsonArray("entry");

        List<Map<String, String>> list = new ArrayList<>();

        for (int lineNo = startIndex; lineNo < (startIndex + maxLines); lineNo++) {

            if (lineNo >= rows.size()) {
                break;
            }

            jsonObject = (JsonObject) rows.get(lineNo);

            Map<String, String> map = new HashMap<>();
            list.add(map);

            for (Entry<String, JsonElement> element : jsonObject.entrySet()) {

                String key = element.getKey();
                if (keysToIgnore.contains(key))
                    continue;

                if (!key.startsWith("gsx$"))
                    throw new APIException().status(Status.ERROR_RESOURCE_INCORRECT).message(
                            "Invalid column header key [" + key
                                    + "]. Key for column header should start with gsx$");

                key = key.substring(4);
                String value = ((JsonObject) element.getValue()).get("$t").getAsString();
                map.put(key, value);
            }
        }

        return list;
    }

    static String getJSONFeedURL(String gsKey) {
        return "https://spreadsheets.google.com/feeds/list/" + gsKey + "/od6/public/values?alt=json";
    }

    public static ListFeed getSpreadSheetListFeed(String accessToken, String spreadSheetKey) throws
        IOException, ServiceException, APIException {

        SpreadsheetService service = new SpreadsheetService("MySpreadsheetIntegration-v1");
        GoogleCredential credential = new GoogleCredential().setAccessToken(accessToken);
        service.setOAuth2Credentials(credential);
        //service.setHeader("Authorization", "Bearer " + accessToken);

        URL SPREADSHEET_FEED_URL = new URL("https://spreadsheets.google.com/feeds/spreadsheets/private/full");

        SpreadsheetFeed feed = null;
        try {
            feed = service.getFeed(SPREADSHEET_FEED_URL, SpreadsheetFeed.class);
        } catch (ServiceException ex) {
            throw new APIException()
                .status(Status.ERROR_INSUFFICIENT_PERMISSION)
                .message("Please check scopes: https://www.googleapis.com/auth/drive https://spreadsheets.google.com/feeds");
        }

        List<SpreadsheetEntry> spreadsheets = feed.getEntries();
        if (spreadsheets.size() == 0) {
            throw new APIException()
                .status(Status.ERROR_INVALID_INPUT)
                .message("No Spreadsheets found.");
        }

        SpreadsheetEntry spreadsheet = null;
        for (SpreadsheetEntry sheet : spreadsheets) {
            if (sheet.getKey().equals(spreadSheetKey)) {
                spreadsheet = sheet;
                break;
            }
        }

        if (spreadsheet == null) {
            throw new APIException()
                .status(Status.ERROR_RESOURCE_NOT_FOUND)
                .message("Spreadsheets " + spreadSheetKey + " not accessible.");
        }

        WorksheetFeed worksheetFeed;
        try {
            worksheetFeed = service.getFeed(spreadsheet.getWorksheetFeedUrl(), WorksheetFeed.class);
        } catch (ServiceException ex) {
            throw new APIException()
                .status(Status.ERROR_INSUFFICIENT_PERMISSION)
                .message("Insufficient Perms."
                    + " Pls check if correct scopes are set to access this feature");
        }

        List<WorksheetEntry> worksheets = worksheetFeed.getEntries();

        WorksheetEntry worksheet = worksheets.get(0);
        // Fetch the list feed of the worksheet.
        URL listFeedUrl = worksheet.getListFeedUrl();
        return service.getFeed(listFeedUrl, ListFeed.class);
    }
}
