package crmdna.interaction;

import crmdna.calling.Campaign;
import crmdna.calling.CampaignProp;
import crmdna.client.Client;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.member.Member;
import crmdna.member.MemberLoader;
import crmdna.user.User;

import java.util.*;

import static crmdna.common.OfyService.ofy;

public class Calling {

    public static CallAssignmentResult preview(String client, long campaignId, Set<Long> memberIds, Set<Long> userIds,
                                                        Map<Long, Integer> userIdVsMaxMembers,
                                                        boolean ensureWithinCampaignDates,
                                                        String login) {
        return previewOrAssign(client, campaignId, memberIds, userIds, userIdVsMaxMembers, ensureWithinCampaignDates, true, login);
    }

    public static CallAssignmentResult assign(String client, long campaignId, Set<Long> memberIds, Set<Long> userIds,
                                        Map<Long, Integer> userIdVsMaxMembers,
                                        boolean ensureWithinCampaignDates,
                                        String login) {
        return previewOrAssign(client, campaignId, memberIds, userIds, userIdVsMaxMembers, ensureWithinCampaignDates, false, login);
    }

    private static CallAssignmentResult previewOrAssign(String client, long campaignId, Set<Long> memberIds, Set<Long> userIds,
                                       Map<Long, Integer> userIdVsMaxMembers,
                                       boolean ensureWithinCampaignDates, boolean preview,
                                       String login) {

        Client.ensureValid(client);

        long groupId = Campaign.safeGet(client, campaignId).toProp().groupId;

        User.ensureGroupLevelPrivilege(client, groupId, login, User.GroupLevelPrivilege.ASSIGN_CALLER);
        Set<Long> alreadyAssignedMemberIds = getAlreadyAssignedMemberIds(client, campaignId);
        memberIds.removeAll(alreadyAssignedMemberIds);

        //include only subscribed members
        Set<Long> subscribedMemberIds = MemberLoader.getSubscribedMemberIds(client, groupId, login);
        memberIds.retainAll(subscribedMemberIds);

        List<UserMemberProp> userMemberProps = getUserMemberProps(memberIds, userIds);

        List<InteractionScoreProp> interactionScoreProps = InteractionScore.get(client, userMemberProps);

        CallAssignmentResult result = CallAssignmentHelper.assignMembersBasedOnScore(interactionScoreProps, userIdVsMaxMembers);

        if (preview) {
            return result;
        }

        if (! result.allMembersAssigned()) {
            throw new APIException("Could not assign [" + result.unassignedMembers.size() +
                    "] due to max limit constraints")
                    .object(result).status(APIResponse.Status.ERROR_INVALID_INPUT);
        }

        Interaction.createPlaceHolderForCalling(client,
                getUserMemberProps(result.userIdVsMemberIds), campaignId,
                ensureWithinCampaignDates, new Date(), login);

        return result;
    }

    static List<UserMemberProp> getUserMemberProps(Map<Long, Set<Long>> userIdVsMemberIds) {
        List<UserMemberProp> userMemberProps = new ArrayList<>();
        for (long userId : userIdVsMemberIds.keySet()) {
            for (long memberId : userIdVsMemberIds.get(userId)) {
                UserMemberProp prop = new UserMemberProp();
                prop.userId = userId;
                prop.memberId = memberId;
                userMemberProps.add(prop);
            }
        }

        return userMemberProps;
    }

    static List<UserMemberProp> getUserMemberProps(Iterable<Long> memberIds, Iterable<Long> userIds) {
        List<UserMemberProp> userMemberProps = new ArrayList<>();
        for (long memberId : memberIds) {
            for (long userId : userIds) {
                UserMemberProp prop = new UserMemberProp();
                prop.userId = userId;
                prop.memberId = memberId;
                userMemberProps.add(prop);
            }
        }

        return userMemberProps;
    }

    public static Set<Long> getAlreadyAssignedMemberIds(String client, long campaignId) {
        Client.ensureValid(client);

        CampaignProp campaignProp = Campaign.safeGet(client, campaignId).toProp();

        List<InteractionEntity> entities = ofy(client).load().
                type(InteractionEntity.class).filter("campaignProgramId", campaignProp.programId)
                .project("memberId").distinct(true).list();

        Set<Long> memberIds = new HashSet<>();
        for (InteractionEntity e : entities) {
            memberIds.add(e.memberId);
        }

        return memberIds;
    }

    public static Set<Long> getAssignedCampaignIds(String client, long userId) {

        Client.ensureValid(client);

        List<InteractionEntity> entities = ofy(client).load().type(InteractionEntity.class)
            .filter("userId", userId).project("campaignId").distinct(true).list();

        Set<Long> campaignIds = new HashSet<>();
        for (InteractionEntity e : entities) {
            campaignIds.add(e.campaignId);
        }

        return campaignIds;
    }
}
