package crmdna.interaction;

import crmdna.calling.Campaign;
import crmdna.calling.CampaignEntity;
import crmdna.calling.CampaignProp;
import crmdna.client.Client;
import crmdna.common.DateUtils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.member.MemberEntity;
import crmdna.member.MemberLoader;
import crmdna.member.MemberProp;
import crmdna.sequence.Sequence;
import crmdna.user.User;
import crmdna.user.User.ClientLevelPrivilege;
import crmdna.user.UserHelper;
import crmdna.user.UserProp;

import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.ensure;
import static crmdna.common.AssertUtils.ensureNotNull;
import static crmdna.common.OfyService.ofy;

public class Interaction {

    public enum Progress {
        NOT_STARTED, IN_PROGRESS, COMPLETE, CLOSED
    }


    public enum Response {
        NA, COLD, NEUTRAL, WARM, HOT
    }

    public static InteractionProp createInteraction(String client, long memberId, String content, InteractionType interactionType,
        Date timestamp, Long campaignId, boolean ensureWithinCampaignDates, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        ensureNotNull(interactionType, "interactionType is null");

        MemberLoader.safeGet(client, memberId, login);
        long userId = User.safeGet(client, login).toProp(client).userId;

        if (timestamp == null)
            timestamp = new Date(); // current time

        if (campaignId != null) {
            CampaignProp campaignProp = Campaign.safeGet(client, campaignId).toProp();
            if (ensureWithinCampaignDates) {
                long yyyymmdd = DateUtils.toYYYYMMDD(timestamp);
                if (yyyymmdd < campaignProp.startYYYYMMDD || yyyymmdd > campaignProp.endYYYYMMDD) {
                    throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_INCORRECT)
                        .object("Timestamp does not fall with campaign start [" + campaignProp.startYYYYMMDD +
                            "] and end [" + campaignProp.endYYYYMMDD +
                            "] dates. To suppress this validation change flag ensureWithinCampaignDates to false");
                }
            }

            ensure(campaignProp.enabled, "Campaign [" + campaignProp.campaignId + "] is disabled");
        }

        InteractionProp interactionProp = InteractionCore
            .createInteraction(client, memberId, content, interactionType.toString(), timestamp, userId, campaignId, login);

        //interaction score is anyway approximate. no need to update score in a transaction
        InteractionScore.incrementBy1(client, login, memberId);

        return interactionProp;
    }

    public static void createPlaceHolderForCalling(String client, List<UserMemberProp> userMemberProps, long campaignId,
        boolean ensureWithinCampaignDates, Date timestamp, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        CampaignProp campaignProp = Campaign.safeGet(client, campaignId).toProp();
        if (ensureWithinCampaignDates) {
            long yyyymmdd = DateUtils.toYYYYMMDD(timestamp);
            if (yyyymmdd < campaignProp.startYYYYMMDD || yyyymmdd > campaignProp.endYYYYMMDD) {
                throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_INCORRECT)
                    .object("Timestamp does not fall with campaign start [" + campaignProp.startYYYYMMDD +
                        "] and end [" + campaignProp.endYYYYMMDD +
                        "] dates. To suppress this validation change flag ensureWithinCampaignDates to false");
            }
        }

        ensure(campaignProp.enabled, "Campaign [" + campaignProp.campaignId + "] is disabled");

        long groupId = campaignProp.groupId;
        User.ensureGroupLevelPrivilege(client, groupId, login, User.GroupLevelPrivilege.ASSIGN_CALLER);

        Map<Long, String> userIdVsEmail = UserHelper.getAllUserIdVsEmail(client, login);

        //ensure valid user ids
        for (UserMemberProp prop : userMemberProps) {
            ensure(userIdVsEmail.containsKey(prop.userId), "Invalid user id [" + prop.userId + "]");
            ensure(prop.memberId != 0, "Found memberId 0");
        }

        List<InteractionEntity> entities = new ArrayList<>(userMemberProps.size());

        List<Long> interactionIds = Sequence.getNext(client, Sequence.SequenceType.INTERACTION, userMemberProps.size());

        for (int i = 0; i < userMemberProps.size(); i++) {
            InteractionEntity entity = new InteractionEntity();
            entity.interactionId = interactionIds.get(i);

            UserMemberProp userMemberProp = userMemberProps.get(i);

            entity.campaignId = campaignId;
            entity.campaignProgramId = campaignProp.programId;
            entity.interactionType = InteractionType.PHONE.toString();
            entity.isTask = true;
            entity.memberId = userMemberProp.memberId;
            entity.userId = userMemberProp.userId;
            entity.progress = Progress.NOT_STARTED;
            entity.ms = timestamp.getTime();
            entity.response = null;

            entities.add(entity);
        }

        ofy(client).save().entities(entities).now();
    }

    public static void deleteInteraction(String client, long interactionId, String login) {
        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        ensure(interactionId != 0, "interaction id is 0");

        InteractionEntity entity = InteractionCore.safeGet(client, interactionId);

        if (!User.isSuperUser(login)) {
            long userId = User.safeGet(client, login).toProp(client).userId;
            // need previlage for deleting someone else's interaction
            if (entity.userId != userId) {
                User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);
            }
        }

        InteractionCore.deleteInteraction(client, interactionId);
    }

    public static InteractionProp updateInteraction(String client, long interactionId, Long newMemberId,
        InteractionType newInteractionType, String newUser, Progress newProgress, Response newResponse, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        InteractionEntity entity = InteractionCore.safeGet(client, interactionId);

        if (newMemberId != null)
            MemberLoader.safeGet(client, newMemberId, login);

        long userId = User.safeGet(client, login).toProp(client).userId;

        // need privilege to update someone else's interaction
        if (entity.userId != userId)
            User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);

        Long newUserId = null;
        if (newUser != null) {
            newUserId = User.safeGet(client, newUser).toProp(client).userId;
            ensureNotNull(newUserId);

            if (newUserId != entity.userId)
                User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);
        }

        return InteractionCore.updateInteraction(client, interactionId, newMemberId,
            newInteractionType != null ? newInteractionType.toString() : null, newUserId, newProgress, newResponse);
    }

    public static InteractionQueryResult query(String client, InteractionQueryCondition qc, String login) {
        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        return InteractionCore.query(client, qc);
    }

    public static List<InteractionProp> queryProps(String client, InteractionQueryCondition qc, String login) {
        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        return InteractionCore.queryProps(client, qc);
    }

    public static int count(String client, InteractionQueryCondition qc) {
        return InteractionCore.queryKeys(client, qc).size();
    }

    public static InteractionProp createSubInteraction(String client, long interactionId, String content, Date timestamp,
        String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        ensure(interactionId != 0, "interactionId is 0");

        InteractionEntity interactionEntity = InteractionCore.safeGet(client, interactionId);

        long userId = User.safeGet(client, login).toProp(client).userId;

        if (interactionEntity.userId != userId) {
            User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);
        }

        if (null == timestamp)
            timestamp = new Date();

        InteractionCore.createSubInteraction(client, interactionId, content, timestamp);

        //interaction score is anyway approximate. no need to update score in a transaction
        InteractionScore.incrementBy1(client, login, interactionEntity.memberId);

        if (interactionEntity.progress == Progress.NOT_STARTED) {
            interactionEntity.progress = Progress.IN_PROGRESS;
            ofy(client).save().entity(interactionEntity).now();
        }

        return interactionEntity.toProp(client);
    }

    public static void deleteSubInteraction(String client, long interactionId, long subinteractionId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        ensure(interactionId != 0, "interactionId is 0");

        InteractionEntity entity = InteractionCore.safeGet(client, interactionId);

        long userId = User.safeGet(client, login).toProp(client).userId;

        if (entity.userId != userId)
            User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);

        InteractionCore.deleteSubInteraction(client, interactionId, subinteractionId);
    }

    public static void updateSubInteraction(String client, long interactionId, long subInteractionId, String content, Date timestamp,
        String login) {

        Client.ensureValid(client);

        InteractionEntity interactionEntity = InteractionCore.safeGet(client, interactionId);

        long userId = User.safeGet(client, login).toProp(client).userId;

        if (interactionEntity.userId != userId)
            User.ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_INTERACTION);

        InteractionCore.updateSubInteraction(client, interactionId, subInteractionId, content, timestamp);
    }

    public static void populateDependents(String client, List<InteractionProp> interactionProps) {

        List<Long> memberIds = new ArrayList<>();
        for (InteractionProp interactionProp : interactionProps) {
            memberIds.add(interactionProp.memberId);
        }

        Map<Long, MemberEntity> memberEntityMap = MemberLoader.get(client, memberIds, User.SUPER_USER);
        for (InteractionProp interactionProp : interactionProps) {
            MemberEntity memberEntity = memberEntityMap.get(interactionProp.memberId);
            if (memberEntity == null) {
                Logger logger = Logger.getLogger("Interaction");
                logger.warning("memberId " + interactionProp.memberId + " not found... deleting the interaction "
                    + interactionProp.interactionId);
                Interaction.deleteInteraction(client, interactionProp.interactionId, User.SUPER_USER);
                continue;
            }

            MemberProp memberProp = memberEntity.toProp();

            interactionProp.memberFirstName = memberProp.contact.firstName;
            interactionProp.memberLastName = memberProp.contact.lastName;

            UserProp userProp = User.safeGetById(client, interactionProp.userId).toProp(client);
            interactionProp.userFirstName = userProp.firstName;
            interactionProp.userLastName = userProp.lastName;

            if (interactionProp.campaignId != null) {
                CampaignProp campaignProp = Campaign.safeGet(client, interactionProp.campaignId).toProp();
                interactionProp.campaignName = campaignProp.shortName;
            }
        }
    }

    public static void setProgramId(InteractionEntity entity, long programId) {
        entity.campaignProgramId = programId;
    }

    public enum InteractionType {
        PHONE, EMAIL, FB, GOOGLEPLUS, SMS, FACETOFACE, WEBSITE, OTHER
    }
}
