package crmdna.mail2;

import com.googlecode.objectify.cmd.Query;
import com.microtripit.mandrillapp.lutung.model.MandrillApiError;
import crmdna.client.Client;
import crmdna.client.ClientProp;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.group.Group;
import crmdna.list.ListProp;
import crmdna.member.MemberQueryCondition;
import crmdna.program.Program;
import crmdna.sequence.Sequence;
import crmdna.sequence.Sequence.SequenceType;
import crmdna.user.User;
import crmdna.user.User.GroupLevelPrivilege;

import java.io.IOException;
import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.*;
import static crmdna.common.OfyService.ofy;

public class MailSchedule {

    private static final int DEFAULT_MAX_DELAY_HOURS = 48;
    private static final Logger LOG = Logger.getLogger(MailSchedule.class.getName());

    public enum FailureCode {
        PAST_MAX_TIME, NO_RECIPIENTS, MAIL_SEND_EXCEPTION
    }

    public static MailScheduleProp create(String client, MailScheduleInput msi, String login) {

        return createOrValidate(client, msi, true, login);
    }

    public static void throwIfInvalid(String client, MailScheduleInput msi, String login) {

        createOrValidate(client, msi, false, login);
    }

    private static MailScheduleProp createOrValidate(String client, MailScheduleInput msi, boolean create, String login) {

        Client.ensureValid(client);

        ensureNotNull(msi, "MailScheduleInput is null");
        ensure(msi.mailContentId != 0, "mailContentId is 0");
        MailContent.safeGet(client, msi.mailContentId);

        ensureNotNull(msi.scheduledTime, "scheduledTime not specified");
        ensure(msi.scheduledTime.getTime() > System.currentTimeMillis(), "scheduledTime [" + msi.scheduledTime
                + "] is in the past");

        if (msi.maxTime == null) {
            msi.maxTime = new Date(msi.scheduledTime.getTime() + DEFAULT_MAX_DELAY_HOURS * 3600 * 1000);
        } else {
            ensure(msi.maxTime.getTime() > msi.scheduledTime.getTime(),
                    "Max time [" + msi.maxTime + "] is not after scheduledTime [" + msi.scheduledTime + "]");
        }

        ensure(msi.listId != 0, "listId is 0");
        ListProp listProp = crmdna.list.List.safeGet(client, msi.listId).toProp();

        User.ensureGroupLevelPrivilege(client, listProp.groupId, login, GroupLevelPrivilege.SEND_EMAIL);

        Group.safeGet(client, listProp.groupId);

        MemberQueryCondition mqc = new MemberQueryCondition(client, 10000);
        mqc.listIds.add(msi.listId);
        if (msi.programId != null) {
            Program.safeGet(client, msi.programId);
            mqc.programIds.add(msi.programId);
        }

        //verify there is at least 1 email recipient
        MailMap mailMap = MailMapFactory.getFromMemberQueryCondition(mqc, listProp.groupId,
                msi.defaultFirstName, msi.defaultLastName, login);
        ensure(!mailMap.isEmpty(), "No email recipients");

        //throw error if all recipients have received this email
        ensureMailNotAlreadySentToAll(client, mailMap.getEmails(), msi.mailContentId);

        ensureNotNullNotEmpty(msi.senderEmail, "senderEmail not specified");
        Group.safeGetSenderNameFromEmail(client, listProp.groupId, msi.senderEmail);

        MailContent.ensureReservedMailContentExists(client, 0,
                MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_SUCCESS);
        MailContent.ensureReservedMailContentExists(client, 0,
                MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE);

        ensureNotDuplicate(client, msi.listId, msi.programId, msi.mailContentId);
        if (msi.programId != null) {
            ensureNoPendingEmailForProgramWithinDuration(client, msi.programId, msi.scheduledTime, 24);
        }

        if (! create) {
            return null;
        }

        //create entity in datastore
        MailScheduleEntity mailScheduleEntity = new MailScheduleEntity();
        mailScheduleEntity.mailScheduleId = Sequence.getNext(client, SequenceType.MAIL_SCHEDULE);
        mailScheduleEntity.defaultFirstName = msi.defaultFirstName;
        mailScheduleEntity.defaultLastName = msi.defaultLastName;
        mailScheduleEntity.groupId = listProp.groupId;
        mailScheduleEntity.mailContentId = msi.mailContentId;
        mailScheduleEntity.userEmail = login;
        mailScheduleEntity.listId = msi.listId;
        mailScheduleEntity.programId = msi.programId;
        mailScheduleEntity.scheduledTimeMs = msi.scheduledTime.getTime();
        mailScheduleEntity.maxTimeMs = msi.maxTime.getTime();
        mailScheduleEntity.senderEmail = msi.senderEmail;
        mailScheduleEntity.enabled = msi.enabled;

        ofy(client).save().entity(mailScheduleEntity).now();

        return mailScheduleEntity.toProp();
    }

    static void ensureMailNotAlreadySentToAll(String client, Set<String> emails, long mailContentId) {

        Set<String> alreadySentEmails = Mail.getAlreadySentEmails(client, mailContentId);

        int newReceipients = 0;
        for (String email : emails) {
            if (! alreadySentEmails.contains(email))
                newReceipients++;
        }

        ensure(newReceipients > 0, "Mail already sent to all specified recipients");
    }

    static void ensureNotDuplicate(String client, long listId, Long programId, long mailContentId) {
        Client.ensureValid(client);

        Query<MailScheduleEntity> q = ofy(client).load().type(MailScheduleEntity.class).filter("listId", listId)
                .filter("mailContentId", mailContentId).filter("processed", false);

        if (programId != null) {
            q = q.filter("programId", programId);
        }

        List<MailScheduleEntity> list = q.list();

        System.out.println("list.isEmpty(): " + list.isEmpty());
        if (!list.isEmpty()) {
            throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_ALREADY_EXISTS)
                    .message("Cannot schedule mailContentId ["
                            + mailContentId + "]. There is already a scheduled email with id [" + list.get(0).mailScheduleId + "] that has the same parameters");
        }
    }

    static void ensureNoPendingEmailForProgramWithinDuration(String client, long programId, Date scheduledTime, int durationInHours) {
        Client.ensureValid(client);

        List<MailScheduleEntity> list = ofy(client).load().type(MailScheduleEntity.class)
                .filter("programId", programId)
                .filter("processed", false).list();

        long durationInMs = durationInHours * 3600 * 1000;
        for (MailScheduleEntity entity : list) {
            long diffInMs = Math.abs(entity.scheduledTimeMs - scheduledTime.getTime());
            ensure(diffInMs > durationInMs, "There is already a schedule email " +
                    "[" + entity.mailScheduleId + "] for participants of program [" + programId +
                    "] within [" + durationInHours + "] hours of attempted scheduled time [" + scheduledTime + "]");
        }
    }

    public static MailScheduleEntity safeGet(String client, long mailScheduleId) {
        Client.ensureValid(client);

        ensure(mailScheduleId != 0, "mailScheduleId is 0 (should be positive)");

        MailScheduleEntity mse = ofy(client).load().type(MailScheduleEntity.class)
                .id(mailScheduleId).now();

        if (mse == null) {
            throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND)
                    .message("Mail schedule id [" + mailScheduleId +
                            "] not found for client [" + client + "]");
        }

        return mse;
    }

    public static MailScheduleProp update(String client, long mailScheduleId, String newDefaultFirstName,
                                            String newDefaultLastName, Boolean newEnabled,
                                            Date newScheduledTime, String login) {
        Client.ensureValid(client);

        MailScheduleEntity entity = safeGet(client, mailScheduleId);

        ensure(!entity.processed, "Cannot update. Mail schedule id [" + mailScheduleId + "] has already been processed.");

        User.ensureGroupLevelPrivilege(client, entity.groupId, login,
                GroupLevelPrivilege.SEND_EMAIL);

        if (newDefaultFirstName != null) {
            ensure(newDefaultFirstName.length() > 2, "Specified default first name [" + newDefaultFirstName + "] is less than 3 char");
            entity.defaultFirstName = newDefaultFirstName;
        }

        if (newDefaultLastName != null) {
            ensure(newDefaultLastName.length() > 2, "Specified default last name [" + newDefaultLastName + "] is less than 3 char");
            entity.defaultLastName = newDefaultLastName;
        }

        if (newEnabled != null) {
            entity.enabled = newEnabled;
        }

        if (newScheduledTime != null) {
            ensure(newScheduledTime.getTime() > System.currentTimeMillis(), "scheduledTime [" + newScheduledTime
                    + "] is in the past");
            entity.scheduledTimeMs = newScheduledTime.getTime();
            entity.maxTimeMs = entity.scheduledTimeMs + DEFAULT_MAX_DELAY_HOURS * 3600 * 1000;
        }

        ofy(client).save().entity(entity).now();
        return entity.toProp();
    }

    public static void delete(String client, long mailScheduleId, String login) {
        Client.ensureValid(client);

        MailScheduleEntity entity = safeGet(client, mailScheduleId);

        ensure(!entity.processed, "Mail schedule id [" + mailScheduleId + "] has already been processed");

        User.ensureGroupLevelPrivilege(client, entity.groupId, login,
                GroupLevelPrivilege.SEND_EMAIL);

        ofy(client).delete().entity(entity).now();
    }

    public static List<MailScheduleProp> query(String client, MailScheduleQueryCondition qc, String login) {
        Client.ensureValid(client);

        User.ensureValidUser(client, login);

        Query<MailScheduleEntity> q = ofy(client).load().type(MailScheduleEntity.class);

        if (qc.mailContentIds != null && !qc.mailContentIds.isEmpty()) {
            q = q.filter("mailContentId in", qc.mailContentIds);
        }
        if (qc.listId != null) {
            q = q.filter("listId", qc.listId);
        }
        if (qc.programId != null) {
            q = q.filter("programId", qc.programId);
        }
        if (qc.userEmail != null) {
            q = q.filter("userEmail", qc.userEmail);
        }
        if (qc.groupId != null) {
            q = q.filter("groupId", qc.groupId);
        }
        if (qc.senderEmail != null) {
            q = q.filter("senderEmail", qc.senderEmail);
        }
        if (qc.processed != null) {
            q = q.filter("processed", qc.processed);
        }
        if (qc.sendSuccess != null) {
            q = q.filter("sendSuccess", qc.sendSuccess);
        }
        if (qc.enabled != null) {
            q = q.filter("enabled", qc.enabled);
        }
        if (qc.scheduledTimeStart != null) {
            q = q.filter("scheduledTimeMs >=", qc.scheduledTimeStart.getTime());
        }
        if (qc.scheduledTimeEnd != null) {
            q = q.filter("scheduledTimeMs <=", qc.scheduledTimeEnd.getTime());
        }
        q = q.order("scheduledTimeMs");

        int resultSize = q.count();
        final int MAX = 5000;
        if (resultSize > MAX) {
            throw new APIException().status(APIResponse.Status.ERROR_OVERFLOW)
                    .message("Query would fetch [" + resultSize + "] entities. Max allowed is [" + MAX + "]");
        }

        List<MailScheduleEntity> entities = q.list();
        List<MailScheduleProp> props = new ArrayList();

        for (MailScheduleEntity entity : entities) {
            props.add(entity.toProp());
        }

        return props;
    }

    public static Map<String, MailScheduleProp> processFirstScheduledEmailForAllClients(Date currentTime) throws IOException, MandrillApiError {
        List<ClientProp> clientProps = Client.getAll();

        //sort clients
        TreeSet<String> clients = new TreeSet<>();
        for (ClientProp prop : clientProps) {
            clients.add(prop.name);
        }

        Map<String, MailScheduleProp> map = new TreeMap<>();
        //send first mail for each client
        for (String client : clients) {
            MailScheduleEntity mse = processFirstScheduledEmail(client, currentTime);
            if (mse != null) {
                map.put(client, mse.toProp());
                LOG.info("Processed scheduled email [" + mse.mailScheduleId + "] for client [" + client + "]");
            } else {
                LOG.info("No pending email for client [" + client + "]");
            }
        }

        return map;
    }

    //returns the MailScheduleEntity that was processed if any, else null
    public static MailScheduleEntity processFirstScheduledEmail(String client, Date currentTime) throws IOException, MandrillApiError {

        Client.ensureValid(client);
        ensureNotNull(currentTime, "currentTime not specified");

        List<MailScheduleEntity> entities = ofy(client).load().type(MailScheduleEntity.class).filter("processed", false)
                .filter("enabled", true)
                .filter("scheduledTimeMs <=", currentTime.getTime()).order("scheduledTimeMs").limit(1).list();

        if (entities.isEmpty()) {
            return null;
        }

        ensureEqual(1, entities.size(),
                "MailScheduleEntity query fetched [" + entities.size() + "] entities but expected 1");

        MailScheduleEntity mse = entities.get(0);

        try {
            if (mse.maxTimeMs != null && mse.maxTimeMs < currentTime.getTime()) {
                mse.sendSuccess = false;
                mse.processedAfterMaxTime = true;
                mse.failureCode = FailureCode.PAST_MAX_TIME;
                mse.failureMessage = "Could not process within maximum delay";
                return mse;
            }

            MailSendInput msi = new MailSendInput();
            msi.createMember = false;
            msi.groupId = mse.groupId;
            msi.isTransactionEmail = false;
            msi.mailContentId = mse.mailContentId;
            msi.senderEmail = mse.senderEmail;
            msi.suppressIfAlreadySent = true;
            msi.mailScheduleId = mse.mailScheduleId;

            MemberQueryCondition mqc = new MemberQueryCondition();
            mqc.client = client;
            mqc.listIds.add(mse.listId);
            if (mse.programId != null) {
                mqc.programIds.add(mse.programId);
            }

            try {
                MailMap mailMap = MailMapFactory.getFromMemberQueryCondition(mqc, mse.groupId, mse.defaultFirstName,
                        mse.defaultLastName, mse.userEmail);

                if (mailMap.isEmpty()) {
                    mse.numRecipients = 0;
                    mse.sendSuccess = false;
                    mse.failureCode = FailureCode.NO_RECIPIENTS;
                    mse.failureMessage = "No recipients";
                    return mse;
                }

                List<SentMailEntity> sentMailEntities = Mail.send(client, msi, mailMap, mse.userEmail);
                mse.numRecipients = sentMailEntities.size();
                mse.sendSuccess = true;

            } catch (Exception ex) {
                mse.sendSuccess = false;
                mse.failureCode = FailureCode.MAIL_SEND_EXCEPTION;
                mse.failureMessage = ex.getMessage();
                mse.stackTraceElementProps = Utils.getStackTrace(ex);
            }

        } finally {
            mse.processed = true;
            mse.processTimeMs = System.currentTimeMillis();
            ofy(client).save().entity(mse).now();
            sendSuccessOrFailureEmail(client, mse);
        }

        return mse;
    }

    private static void sendSuccessOrFailureEmail(String client, MailScheduleEntity mse) throws IOException, MandrillApiError {
        Client.ensureValid(client);

        ensureNotNull(mse.sendSuccess, "mse.sendSuccess is null");

        //populate mailMap
        MailMap mailMap = new MailMap();
        mailMap.add(mse.userEmail);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.MAIL_CONTENT_ID, mse.mailContentId);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.SCHEDULED_MAIL_ID, mse.mailScheduleId);

        MailContentProp mailContentProp = MailContent.safeGet(client, mse.mailContentId).toProp();
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.MAIL_DISPLAY_NAME, mailContentProp.displayName);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.MAIL_SUBJECT, mailContentProp.subject);

        ListProp listProp = crmdna.list.List.safeGet(client, mse.listId).toProp();
        String groupName = Group.safeGet(client, listProp.groupId).toProp().displayName;
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.LIST, listProp.displayName);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.GROUP, groupName);

        if (mse.programId != null) {
            String programName = Program.safeGet(client, mse.programId).toProp(client).getNameWOVenue();
            mailMap.setGlobalMergeVar(MailMap.MergeVarID.PROGRAM_NAME, programName);
        } else {
            mailMap.setGlobalMergeVar(MailMap.MergeVarID.PROGRAM_NAME, "Not applicable");
        }

        mailMap.setGlobalMergeVar(MailMap.MergeVarID.SCHEDULED_TIME, new Date(mse.scheduledTimeMs));

        MailSendInput msi = new MailSendInput();
        msi.isTransactionEmail = true;
        msi.senderEmail = Client.getEmailConfig(client, User.SUPER_USER).contactEmail;
        msi.groupId = mse.groupId;
        msi.suppressIfAlreadySent = false;

        if (mse.sendSuccess) {

            mailMap.setGlobalMergeVar(MailMap.MergeVarID.NUM_RECIEPIENTS, mse.numRecipients);

            MailContentEntity reservedMailContentEntity = MailContent.getByName(client,
                    MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_SUCCESS.toString(), 0);
            ensureNotNull(reservedMailContentEntity, "Mail content missing for RESERVED_SCHEDULED_EMAIL_SUCCESS");

            msi.mailContentId = reservedMailContentEntity.mailContentId;
            msi.tags = Utils.getSet(Mail.MailTag.SCHEDULED_EMAIL_SUCCESS.toString());

            Mail.send(client, msi, mailMap, mse.userEmail);

        } else {

            mailMap.setGlobalMergeVar(MailMap.MergeVarID.FAILURE_REASON, mse.failureMessage);

            MailContentEntity reservedMailContentEntity = MailContent.getByName(client,
                    MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE.toString(), 0);
            ensureNotNull(reservedMailContentEntity, "Mail content missing for RESERVED_SCHEDULED_EMAIL_FAILURE");

            msi.mailContentId = reservedMailContentEntity.mailContentId;
            msi.tags = Utils.getSet(Mail.MailTag.SCHEDULED_EMAIL_FAILURE.toString());

            Mail.send(client, msi, mailMap, mse.userEmail);
        }
    }
}
