package crmdna.mail2;

import com.googlecode.objectify.cmd.Query;
import crmdna.client.Client;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.group.Group;
import crmdna.group.GroupEntity;
import crmdna.list.ListProp;
import crmdna.program.Program;
import crmdna.programtype.ProgramType;
import crmdna.sequence.Sequence;
import crmdna.user.User;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import static crmdna.common.AssertUtils.ensure;
import static crmdna.common.AssertUtils.ensureNotNull;
import static crmdna.common.AssertUtils.ensureNotNullNotEmpty;
import static crmdna.common.OfyService.ofy;

/**
 * Created by Sathya on 28/11/2015.
 */
public class MailSequence {
    public static MailSequenceProp create(String client, MailSequenceProp input, String login) {

        Client.ensureValid(client);

        ensureNotNull(input);
        ensure(input.groupId != 0, "groupId cannot be 0.");
        Group.safeGet(client, input.groupId);

        ensure(input.mailContentIds != null, "No mailContentId specified");
        ensure(input.mailContentIds.size() != 0, "No mailContentId specified");
        ensure(input.dayOffsets != null, "No dayOffset specified");
        ensure(input.mailContentIds.size() == input.dayOffsets.size(),
                "[" + input.mailContentIds.size() + "] mailContent(s) specified but only ["
                        + input.dayOffsets + "] day offset(s) specified");

        ensureValidMailContentIds(client, input.mailContentIds, input.groupId);

        ensureValidDayOffsets(input.dayOffsets);

        ensureNotNullNotEmpty(input.displayName, "displayName not specified");
        ensure(input.programTypeId != 0, "programTypeId cannot be 0");
        ProgramType.safeGet(client, input.programTypeId);

        User.ensureGroupLevelPrivilege(client, input.groupId, login, User.GroupLevelPrivilege.UPDATE_MAIL_SEQUENCE);

        MailSequenceEntity entity = new MailSequenceEntity();
        entity.mailSequenceId = Sequence.getNext(client, Sequence.SequenceType.MAIL_SEQUENCE);
        entity.dayOffsets = input.dayOffsets;
        entity.displayName = input.displayName;
        entity.name = Utils.removeSpaceUnderscoreBracketAndHyphen(input.displayName.toLowerCase());
        entity.groupId = input.groupId;
        entity.mailContentIds = input.mailContentIds;
        entity.programTypeId = input.programTypeId;

        ofy(client).save().entity(entity).now();

        return entity.toProp();
    }

    public static List<MailSequenceProp> query(String client, MailSequenceQueryCondition qc, String login) {
        Client.ensureValid(client);

        User.ensureValidUser(client, login);

        Query<MailSequenceEntity> q = ofy(client).load().type(MailSequenceEntity.class);

        if (qc.groupId != null) {
            q = q.filter("groupId", qc.groupId);
        }
        if (qc.programTypeId != null) {
            q = q.filter("programTypeId", qc.programTypeId);
        }

        List<MailSequenceEntity> entities = q.list();

        List<MailSequenceProp> props = new ArrayList<>();
        for (MailSequenceEntity entity : entities) {
            props.add(entity.toProp());
        }

        return props;
    }

    static void ensureValidMailContentIds(String client, List<Long> mailContentIds, long groupId) {
        Client.safeGet(client);

        for (long mailContentId : mailContentIds) {
            MailContentProp prop = MailContent.safeGet(client, mailContentId).toProp();
            ensure(prop.groupId == groupId, "Group mismatch for mailContentId [" + mailContentId +
                "]. Expected groupId to be [" + groupId + "] but found [" + prop.groupId + "]");

            ensure(! prop.name.contains("reserved"), "Reserved email cannot be scheduled. mailContentId ["
                + prop.mailContentId + "] is reserved");
        }
    }

    static void ensureValidDayOffsets(List<Integer> dayOffsets) {

        ensure(dayOffsets.get(0) == 0, "First element of dayOffsets should be 0. Found ["
                + dayOffsets.get(0) + "]");

        //ensure strictly increasing order
        int prev = dayOffsets.get(0);

        for (int i = 1; i < dayOffsets.size(); i++) {
            ensure(dayOffsets.get(i) > prev, "dayOffsets should be in strictly increasing order. Error at index ["
                    + i + "] when validating dayOffsets");
            prev = dayOffsets.get(i);
        }

        final int MAX_NURTURE_CAMPAIGN_DURATION_IN_DAYS = 120;
        int first = dayOffsets.get(0);
        int last = dayOffsets.get(dayOffsets.size() - 1);
        ensure(last - first < MAX_NURTURE_CAMPAIGN_DURATION_IN_DAYS,
                "Last email should be within [" + MAX_NURTURE_CAMPAIGN_DURATION_IN_DAYS + "] days from first email");
    }

    public static void delete(String client, long mailSequenceId, String login) {

        Client.safeGet(client);
        MailSequenceEntity entity = MailSequence.safeGet(client, mailSequenceId);

        User.safeGet(client, login);
        User.ensureGroupLevelPrivilege(client, entity.groupId, login,
                User.GroupLevelPrivilege.UPDATE_MAIL_SEQUENCE);

        ofy(client).delete().entity(entity).now();

        return;
    }

    public static List<MailScheduleProp> trigger(String client, long mailSequenceId, long programId,
                                                 long listId, String defaultFirstName, String defaultLastName,
                                                 Date firstMailScheduleTime, String senderEmail, String login) {

        //triggers for program participants if they are present in list

        Client.safeGet(client);
        MailSequenceProp mailSequenceProp = safeGet(client, mailSequenceId).toProp();
        Program.safeGet(client, programId);
        ListProp listProp = crmdna.list.List.safeGet(client, listId).toProp();

        ensure(listProp.groupId == mailSequenceProp.groupId, "Group id mismatch. List has group id [" + listProp.groupId +
                "] but mail sequence has group id [" + mailSequenceProp.groupId + "]");

        User.ensureGroupLevelPrivilege(client, listProp.groupId, login, User.GroupLevelPrivilege.SEND_EMAIL);

        ensure(firstMailScheduleTime != null, "fistMailScheduleTime not specified");
        ensure(firstMailScheduleTime.after(new Date()), "firstMailScheduleTime should be in the future");

        //all or none
        //if atleast one error none of the mailSchedules should be created
        //to achieve this:
        //1) create mailScheduleInputs and validate them
        //2) if all are valid create entities in datastore

        List<MailScheduleInput> mailScheduleInputs = new ArrayList<>(mailSequenceProp.mailContentIds.size());
        for (int i = 0; i < mailSequenceProp.mailContentIds.size(); i++) {
            MailScheduleInput msi = new MailScheduleInput();
            msi.defaultFirstName = defaultFirstName;
            msi.defaultLastName = defaultLastName;
            msi.enabled = true;
            msi.listId = listId;
            msi.mailContentId = mailSequenceProp.mailContentIds.get(i);
            msi.programId = programId;

            if (i == 0) {
                msi.scheduledTime = firstMailScheduleTime;
            } else {
                long prevScheduledTime = mailScheduleInputs.get(i -1).scheduledTime.getTime();
                final long MILLIS_IN_A_DAY = 86400 * 1000;
                msi.scheduledTime = new Date(prevScheduledTime + mailSequenceProp.dayOffsets.get(i) * MILLIS_IN_A_DAY);
            }

            msi.senderEmail = senderEmail;

            MailSchedule.throwIfInvalid(client, msi, login);
            mailScheduleInputs.add(msi);
        }

        //all mailScheduleInputs are valid. create entities in datastore
        List<MailScheduleProp> mailScheduleProps = new ArrayList<>();
        for (MailScheduleInput msi : mailScheduleInputs) {
            MailScheduleProp prop = MailSchedule.create(client, msi, login);
            mailScheduleProps.add(prop);
        }

        return mailScheduleProps;
    }

    public static MailSequenceEntity safeGet(String client, long mailSequenceId) {

        Client.safeGet(client);

        MailSequenceEntity entity = ofy(client).load().type(MailSequenceEntity.class).id(mailSequenceId).now();
        if (null == entity)
            throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND).message(
                    "Mail sequence id [" + mailSequenceId + "] does not exist");

        return entity;
    }
}
