package crmdna.payment;

import com.googlecode.objectify.cmd.Query;
import crmdna.client.Client;
import crmdna.common.DateUtils;
import crmdna.common.EmailConfig;
import crmdna.common.NumberUtils;
import crmdna.common.Utils;
import crmdna.common.Utils.Currency;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.group.Group;
import crmdna.mail2.*;
import crmdna.mail2.MailContent.ReservedMailContentName;
import crmdna.user.User;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Logger;

import static crmdna.common.AssertUtils.ensureNotNull;
import static crmdna.common.OfyService.ofy;
import static crmdna.mail2.MailContent.ensureReservedMailContentExists;

public class Receipt {

    enum Purpose {
        PROGRAM_REGISTRATION,
        VOLUNTEER_CONTRIBUTION,
        ADHOC_DONATION
    }

    public static class ReceiptQueryCondition {
        public Long registrationId;
        public Long groupId;
        public Integer startYYYYMMDD;
        public Integer endYYYYMMDD;
        public String firstName;
        public String lastName;
        public String email;
        public Purpose purpose;
        public String adhocReference;
        public Currency ccy;
        public Double amount;
    }

    public static ReceiptProp generateForRegistration(String client, long groupId,
        long registrationId, boolean sendEmail) {

        ensureReservedMailContentExists(client, groupId, ReservedMailContentName.RESERVED_RECEIPT);

        ReceiptProp receiptProp = getByRegistrationId(client, registrationId);
        if (receiptProp == null) {
            ReceiptEntity receiptEntity = new ReceiptEntity(client, groupId);
            receiptEntity.registrationId = registrationId;
            receiptEntity.purpose = Purpose.PROGRAM_REGISTRATION;

            ofy(client).save().entity(receiptEntity).now();

            receiptProp = receiptEntity.toProp(client);
            sendEmail = true;
        }

        if (sendEmail) {
            sendEmail(client, receiptProp);
        }

        return receiptProp;
    }

    public static ReceiptProp generateForProgram(String client, long groupId, String firstName,
        String lastName, String email, long programId, Purpose purpose, Currency ccy, double
        amount, int dateYYYYMMDD) {

        ensureReservedMailContentExists(client, groupId, ReservedMailContentName.RESERVED_RECEIPT);

        ReceiptEntity receiptEntity = new ReceiptEntity(client, groupId, dateYYYYMMDD);
        receiptEntity.firstName = firstName;
        receiptEntity.lastName = lastName;
        receiptEntity.email = email;
        receiptEntity.programId = programId;
        receiptEntity.purpose = purpose;
        receiptEntity.ccy = ccy;
        receiptEntity.amount = amount;

        ofy(client).save().entity(receiptEntity).now();

        ReceiptProp receiptProp = receiptEntity.toProp(client);
        sendEmail(client, receiptProp);

        return receiptProp;
    }

    public static ReceiptProp generateForAdhocDonation(String client, long groupId,
        String firstName, String lastName, String email, String adhocReference,
        Currency ccy, double amount, int dateYYYYMMDD, boolean shouldSendEmail) {

        ensureReservedMailContentExists(client, groupId, ReservedMailContentName.RESERVED_RECEIPT);

        ReceiptQueryCondition rqc = new ReceiptQueryCondition();
        rqc.groupId = groupId;
        rqc.startYYYYMMDD = dateYYYYMMDD;
        rqc.endYYYYMMDD = dateYYYYMMDD;
        rqc.firstName = firstName;
        rqc.lastName = lastName;
        rqc.email = email;
        rqc.purpose = Purpose.ADHOC_DONATION;
        rqc.adhocReference = adhocReference;
        rqc.ccy = ccy;
        rqc.amount = amount;

        List<ReceiptProp> receiptProps = query(client, rqc);
        if (receiptProps.size() > 0) {
            System.out.println(receiptProps.get(0).receiptId);
            return receiptProps.get(0);
//            throw new APIException()
//                .status(APIResponse.Status.ERROR_RESOURCE_ALREADY_EXISTS)
//                .message("Receipt: " + firstName + "-" + lastName + "-" + adhocReference + "-" +
//                    dateYYYYMMDD + "-" + ccy + amount + "-" + email);
        }

        ReceiptEntity receiptEntity = new ReceiptEntity(client, groupId, dateYYYYMMDD);
        receiptEntity.firstName = firstName;
        receiptEntity.lastName = lastName;
        receiptEntity.email = email;
        receiptEntity.adhocReference = adhocReference;
        receiptEntity.purpose = Purpose.ADHOC_DONATION;
        receiptEntity.ccy = ccy;
        receiptEntity.amount = amount;

        ofy(client).save().entity(receiptEntity).now();

        ReceiptProp receiptProp = receiptEntity.toProp(client);
        if (shouldSendEmail) {
            sendEmail(client, receiptProp);
        }

        return receiptProp;
    }

    public static void cancelReceipt(String client, String receiptId) {

        ReceiptEntity receiptEntity = get(client, receiptId);
        if (receiptEntity != null) {
            receiptEntity.cancelled = true;
        }
        ofy(client).save().entity(receiptEntity).now();
    }

    public static ReceiptEntity get(String client, String receiptId) {

        ensureNotNull(client, "client should not be null");
        ensureNotNull(receiptId, "receiptId should not be null");

        return ofy(client).load().type(ReceiptEntity.class).id(receiptId).now();
    }

    public static ReceiptEntity safeGet(String client, String receiptId) {

        Client.ensureValid(client);
        ensureNotNull(receiptId, "receiptId should not be null");

        ReceiptEntity entity = ofy(client).load().type(ReceiptEntity.class).id(receiptId).now();
        if (null == entity) {
            throw new APIException().status(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND)
                .message("Cannot find receipt for id [" + receiptId + "] for client [" + client + "]");
        }

        return entity;
    }

    public static List<ReceiptProp> query(String client, ReceiptQueryCondition qc) {
        Query<ReceiptEntity> query = ofy(client).load().type(ReceiptEntity.class);

        if (qc.registrationId != null) {
            query = query.filter("registrationId", qc.registrationId);
        }

        if (qc.groupId != null) {
            query = query.filter("groupId", qc.groupId);
        }

        if (qc.startYYYYMMDD != null) {
            query = query.filter("ms >=", DateUtils.toDate(qc.startYYYYMMDD).getTime());
        }

        if (qc.endYYYYMMDD != null) {
            query = query.filter("ms <=", DateUtils.toDate(qc.endYYYYMMDD).getTime());
        }

        if (qc.firstName != null) {
            query = query.filter("firstName", qc.firstName);
        }

        if (qc.lastName != null) {
            query = query.filter("lastName", qc.lastName);
        }

        if (qc.email != null) {
            query = query.filter("email", qc.email);
        }

        if (qc.purpose != null) {
            query = query.filter("purpose", qc.purpose);
        }

        if (qc.ccy != null) {
            query = query.filter("ccy", qc.ccy);
        }

        if (qc.amount != null) {
            query = query.filter("amount", qc.amount);
        }

        if (qc.adhocReference != null) {
            query = query.filter("adhocReference", qc.adhocReference);
        }

        List<ReceiptProp> props = new ArrayList<>();
        List<ReceiptEntity> entities = query.list();
        for (ReceiptEntity entity : entities) {
            props.add(entity.toProp(client));
        }

        return props;
    }

    public static ReceiptProp getByRegistrationId(String client, long registrationId) {
        List<ReceiptEntity> entities = ofy(client).load().type(ReceiptEntity.class)
            .filter("registrationId", registrationId)
            .list();

        if (entities.size() == 0) {
            return null;
        }

        return entities.get(0).toProp(client);
    }

    static String getReceiptId(long groupId, long id) {
        return "E"+ String.format("%03d", groupId) + "-" + String.format("%04d", id);
    }

    public static void sendEmail(String client, ReceiptProp receiptProp) {

        Utils.Pair<String> ccy = Utils.getCurrencyName(receiptProp.ccy);

        MailMap mailMap = new MailMap();
        mailMap.add(receiptProp.email, receiptProp.firstName, receiptProp.lastName);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.DATE, DateUtils.toLongDateString(new Date(receiptProp.ms)));
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.TRANSACTION_ID, receiptProp.receiptId);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.PURPOSE, receiptProp.purpose);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.AMOUNT,
            receiptProp.ccy.toString() + " " + String.format("%.2f", receiptProp.amount));

        Utils.Pair<String> amountInWords = NumberUtils.toWords(receiptProp.amount);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.AMOUNT_IN_WORDS,
            ccy.first.toUpperCase() + " " + amountInWords.first.toUpperCase() +
                ((amountInWords.second != null) ?
                    " AND " + ccy.second.toUpperCase() + " " + amountInWords.second.toUpperCase() : ""));

        MailContentEntity mailContentEntity = MailContent.getByName(client, ReservedMailContentName.RESERVED_RECEIPT.toString(), receiptProp.groupId);
        ensureNotNull(mailContentEntity);
        MailContentProp mailContentProp = mailContentEntity.toProp();

        try {
            MailSendInput msi = new MailSendInput();
            msi.createMember = false;
            msi.groupId = receiptProp.groupId;
            msi.isTransactionEmail = true;
            msi.mailContentId = mailContentProp.mailContentId;
            msi.senderEmail = Group.getSenderEmail(client, receiptProp.groupId, User.SUPER_USER);
            msi.overrideSubject = "Receipt for " + receiptProp.purpose;
            msi.suppressIfAlreadySent = false;

            Mail.send(client, msi, mailMap, User.SUPER_USER);
        } catch (Exception ex) {
            Logger LOGGER = Logger.getLogger(Receipt.class.getName());
            LOGGER.severe("Mail Send Error: " + ex.toString());
        }
    }

    public static String getBody(String client, long groupId, String receiptId) {

        MailContentEntity mailContentEntity =
            MailContent.getByName(client,
                MailContent.ReservedMailContentName.RESERVED_RECEIPT.toString(), groupId);
        ensureNotNull(mailContentEntity);
        MailContentProp mailContentProp = mailContentEntity.toProp();

        ReceiptProp receiptProp = Receipt.get(client, receiptId).toProp(client);
        Utils.Pair<String> ccy = Utils.getCurrencyName(receiptProp.ccy);
        Utils.Pair<String> amountInWords = NumberUtils.toWords(receiptProp.amount);
        return mailContentProp.body
            .replace("{{FIRST_NAME}}", receiptProp.firstName).replace("{{LAST_NAME}}", receiptProp.lastName)
            .replace("{{DATE}}", DateUtils.toLongDateString(new Date(receiptProp.ms)))
            .replace("{{TRANSACTION_ID}}", receiptProp.receiptId)
            .replace("{{PURPOSE}}", receiptProp.purpose)
            .replace("{{AMOUNT}}", receiptProp.ccy.toString() + " " +
                String.format("%.2f", receiptProp.amount))
            .replace("{{AMOUNT_IN_WORDS}}", ccy.first.toUpperCase() + " " +
                amountInWords.first.toUpperCase() +
                ((amountInWords.second != null) ?
                    " AND " + ccy.second.toUpperCase() + " " + amountInWords.second.toUpperCase() :
                    ""));
    }
}
