package crmdna.registration;

import com.google.api.client.repackaged.com.google.common.base.Strings;
import com.googlecode.objectify.cmd.Query;
import com.microtripit.mandrillapp.lutung.model.MandrillApiError;
import crmdna.client.Client;
import crmdna.client.isha.IshaConfig;
import crmdna.client.isha.IshaRegistration;
import crmdna.client.isha.ieo.IshaSMM;
import crmdna.common.AssertUtils;
import crmdna.common.DateUtils;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse.Status;
import crmdna.common.api.RequestInfo;
import crmdna.common.contact.Contact;
import crmdna.common.contact.ContactProp;
import crmdna.group.Group;
import crmdna.group.PaypalApiCredentialsProp;
import crmdna.mail2.*;
import crmdna.member.Member;
import crmdna.member.MemberEntity;
import crmdna.member.MemberLoader;
import crmdna.member.MemberProp;
import crmdna.payment.IPaymentResponse;
import crmdna.payment.Payment;
import crmdna.payment.Payment.PaymentType;
import crmdna.payment.PaymentResponseFactory;
import crmdna.payment.TokenProp;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.sequence.Sequence;
import crmdna.sequence.Sequence.SequenceType;
import crmdna.user.User;

import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.logging.Logger;

import static crmdna.common.OfyService.ofy;

public class Registration {

    private static Logger logger = Logger.getLogger(Registration.class.getName());

    public static RegistrationProp register(String client, RegistrationInput regInput) {

        Client.ensureValid(client);

        if (regInput.contact == null) {
            regInput.contact = new ContactProp();

            regInput.contact.firstName = regInput.firstName;
            regInput.contact.lastName = regInput.lastName;
            regInput.contact.nickName = regInput.nickName;
            regInput.contact.email = regInput.email;
            regInput.contact.mobilePhone = Utils.removeSpaceUnderscoreBracketAndHyphen(regInput.mobilePhone);
            regInput.contact.homePhone = Utils.removeSpaceUnderscoreBracketAndHyphen(regInput.homePhone);
            regInput.contact.officePhone = Utils.removeSpaceUnderscoreBracketAndHyphen(regInput.officePhone);
            regInput.contact.gender = Contact.getGender(regInput.gender);
            regInput.contact.asOfyyyymmdd = DateUtils.toYYYYMMDD(new Date());

            regInput.contact.homeAddress.country = regInput.country;
            regInput.contact.homeAddress.pincode = regInput.postalCode;
        }

        if (regInput.batchNo == null) {
            regInput.batchNo = 1;
        }

        if (regInput.successCallbackUrl != null) {
            Utils.ensureValidUrl(regInput.successCallbackUrl);
        }

        if (regInput.errorCallbackUrl != null) {
            Utils.ensureValidUrl(regInput.errorCallbackUrl);
        } else {
            regInput.errorCallbackUrl = regInput.successCallbackUrl;
        }

        ProgramProp programProp = Program.safeGet(client, regInput.programId).toProp(client);
        if ((regInput.batchNo < 1) || (regInput.batchNo > programProp.numBatches))
            Utils.throwIncorrectSpecException("Batch number is invalid");

        // cannot register 30 days after the end of program
        // 30 days is just so that there is enough buffer
        int currentYYYYMMDD = DateUtils.toYYYYMMDD(new Date()); // current date
        Date current = new Date();
        Date end = DateUtils.toDate(programProp.endYYYYMMDD);

        if ((DateUtils.getNumDays(end, current) > 30) && !regInput.ignoreDateRestriction)
            throw new APIException().status(Status.ERROR_OPERATION_NOT_ALLOWED).message(
                    "Registration not allowed as current date [" + currentYYYYMMDD + "] "
                        + "is more than 30 days after the program end date [" + programProp.endYYYYMMDD + "]");

        Contact.ensureFirstNameAndValidEmailSpecified(regInput.contact);
        regInput.contact.email = regInput.contact.email.toLowerCase();

        Long registrationId = getMatchingRegistrationId(client, regInput.programId, regInput.contact);

        RegistrationEntity registrationEntity;
        if (null == registrationId) {
            registrationEntity = new RegistrationEntity();
            registrationEntity.registrationId = Sequence.getNext(client, SequenceType.REGISTRANT);
        } else {
            registrationEntity = safeGet(client, registrationId);
            RegistrationProp registrationProp = registrationEntity.toProp();

            if (registrationProp.getStatus() == RegistrationStatus.REGISTRATION_COMPLETE) {
                registrationProp.alreadyRegistered = true;
                return registrationProp;
            }

            if (registrationProp.getStatus() == RegistrationStatus.PAYMENT_PENDING) {
                String errMessage =
                        "Registration for email [" + regInput.contact.email + "] and Name [" + regInput.contact.getName()
                                + "] is already with status [PAYMENT_PENDING] and registration Id ["
                                + registrationProp.registrationId
                                + "]. Now a new registration for the same name and "
                                + "email is being attempted. This could lead to duplicate registration and "
                                + "duplicate payments.";
                logger.warning(errMessage);
            }

            // clear off the paypal error/payment fields
            registrationEntity.amount = null;
            registrationEntity.ccy = null;
            registrationEntity.isPaymentPending = false;
            registrationEntity.L_ERRORCODE0 = null;
            registrationEntity.L_LONGMESSAGE0 = null;
            registrationEntity.L_SEVERITYCODE0 = null;
            registrationEntity.L_SHORTMESSAGE0 = null;
            registrationEntity.redirectUrl = null;
        }

        Long memberId =
                Member.getMatchingMemberId(client, regInput.contact.email, regInput.contact.firstName, regInput.contact.lastName);
        if (memberId == null) {
            MemberProp memberProp = Member
                .create(client, programProp.groupProp.groupId, regInput.contact, true, User.SUPER_USER);
            memberId = memberProp.memberId;
        }

        populateContactDetailsAndQSTags(registrationEntity, regInput.contact);
        registrationEntity.memberId = memberId;
        registrationEntity.programId = regInput.programId;
        registrationEntity.marketingChannel = regInput.marketingChannel;
        registrationEntity.successCallbackUrl = regInput.successCallbackUrl;
        registrationEntity.errorCallbackUrl = regInput.errorCallbackUrl;

        logger.info("registrationId = [" + registrationEntity.registrationId + "], memberId = ["
                + registrationEntity.memberId + "], programId = [" + registrationEntity.programId + "]");

        if (regInput.fee == null) {
            regInput.fee = programProp.fee;
        }

        registrationEntity.amount = regInput.fee.toString();
        registrationEntity.ccy = programProp.ccy.toString();

        if ((programProp.fee == 0) || (regInput.paymentReference != null)) {
            registrationEntity.transactionId = regInput.paymentReference;
            registrationEntity.recordStateChange(RegistrationStatus.REGISTRATION_COMPLETE);
            ofy(client).save().entity(registrationEntity);

            RegistrationProp registrationProp = registrationEntity.toProp();
            Registration.handlePostRegistration(client, registrationProp, regInput.sendEmail);
            return registrationProp;
        }

        registrationEntity.recordStateChange(RegistrationStatus.REQUEST_RECEIVED);
        ofy(client).save().entity(registrationEntity).now();

        // payment is involved - ensure group is set to receive payment
        PaypalApiCredentialsProp prop = Group.getPaypalApiCredentials(client, programProp.groupProp.groupId, User.SUPER_USER);

        if ((null == prop) || prop.disable)
            throw new APIException().status(Status.ERROR_OPERATION_NOT_ALLOWED).message(
                    "Paypal info is either not available or payments are disabled for group ["
                            + programProp.groupProp.displayName + "]");

        String paymentName = programProp.programTypeProp.displayName + "-" + programProp.description
            + "-" + DateUtils.getDurationAsString(programProp.startYYYYMMDD, programProp.endYYYYMMDD);

        // do a paypal set express checkout and get the payment url
        try {

            registrationEntity.paymentUrl = Payment.setExpressCheckoutAndGetPaymentUrl(
                client, PaymentType.PROGRAM_REGISTRATION, regInput.contact.email, paymentName, registrationEntity.registrationId, prop,
                regInput.fee, programProp.ccy.toString(), regInput.rootUrl, regInput.successCallbackUrl, regInput.errorCallbackUrl);

            registrationEntity.recordStateChange(RegistrationStatus.PAYMENT_AUTHORIZATION_PENDING);
            ofy(client).save().entity(registrationEntity).now();
            return registrationEntity.toProp();

        } catch (APIException ex) {
            registrationEntity.recordStateChange(RegistrationStatus.ERROR_CREATING_PAYMENT_URL);
            ofy(client).save().entity(registrationEntity).now();
            throw ex;
        }
    }

    public static RegistrationEntity safeGet(String client, long registrationId) {
        Client.ensureValid(client);

        RegistrationEntity entity =
            ofy(client).load().type(RegistrationEntity.class).id(registrationId).now();

        if (null == entity)
            throw new APIException().status(Status.ERROR_RESOURCE_NOT_FOUND).message(
                "Cannot find Registration details for registration id [" + registrationId + "]");

        return entity;
    }

    public static Map<Long, RegistrationEntity> safeGet(String client, Collection<Long> registrationIds) {
        Client.ensureValid(client);

        return ofy(client).load().type(RegistrationEntity.class).ids(registrationIds);
    }

    static void populateContactDetailsAndQSTags(RegistrationEntity registrationEntity,
                                                ContactProp contact) {
        if (contact.email != null) {
            Utils.ensureValidEmail(contact.email);
            registrationEntity.email = contact.email.toLowerCase();
        }

        if (contact.firstName != null)
            registrationEntity.firstName = contact.firstName;

        if (contact.lastName != null)
            registrationEntity.lastName = contact.lastName;

        if (contact.nickName != null)
            registrationEntity.nickName = contact.nickName;

        if (contact.homeAddress.address != null)
            registrationEntity.homeAddress = contact.homeAddress.address;

        if (contact.homeAddress.country != null)
            registrationEntity.country = contact.homeAddress.country;

        if (contact.officeAddress.address != null)
            registrationEntity.officeAddress = contact.officeAddress.address;

        if (contact.homePhone != null) {
            Utils.ensureValidPhoneNumber(contact.homePhone);
            registrationEntity.homePhone = contact.homePhone;
        }

        if (contact.officePhone != null) {
            Utils.ensureValidPhoneNumber(contact.officePhone);
            registrationEntity.officePhone = contact.officePhone;
        }

        if (contact.mobilePhone != null) {
            Utils.ensureValidPhoneNumber(contact.mobilePhone);
            registrationEntity.mobilePhone = contact.mobilePhone;
        }

        registrationEntity.qsTags = Utils
            .getQSTags(contact.firstName, contact.lastName, contact.nickName, contact.email,
                contact.mobilePhone, contact.homePhone, contact.officePhone);
    }

    static Long getMatchingRegistrationId(String client, long programId, ContactProp contact) {

        Contact.ensureFirstNameAndValidEmailSpecified(contact);
        contact.email = contact.email.toLowerCase();

        RegistrationQueryCondition queryCondition = new RegistrationQueryCondition();
        queryCondition.programId = programId;
        queryCondition.email = contact.email;

        List<RegistrationProp> props = query(client, queryCondition, User.SUPER_USER);

        for (RegistrationProp prop : props) {
            if (Utils.closeEnough(prop.getName().toLowerCase(), contact.getName().toLowerCase())) {
                return prop.registrationId;
            }
        }

        // no matching registration id. return null
        return null;
    }

    public static boolean isAlreadyRegistered(String client, long programId, MemberProp memberProp) {

        RegistrationQueryCondition queryCondition = new RegistrationQueryCondition();
        queryCondition.programId = programId;
        queryCondition.email = memberProp.contact.email;

        List<RegistrationProp> props = query(client, queryCondition, User.SUPER_USER);

        for (RegistrationProp prop : props) {
            if (prop.memberId == memberProp.memberId) {
                return true;
            }
        }

        return false;
    }

    public static List<RegistrationProp> query(String client, RegistrationQueryCondition qc,
                                               String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        Query<RegistrationEntity> q = ofy(client).load().type(RegistrationEntity.class);

        if (null != qc.programId) {
            q = q.filter("programId", qc.programId);
        }

        if (null != qc.status) {
            q = q.filter("status", qc.status);
        }

        if (null != qc.email) {
            q = q.filter("email", qc.email);
        }

        if (Utils.isNotNullNotEmpty(qc.memberIds)) {
            q = q.filter("memberId in", qc.memberIds);
        }

        if (null != qc.searchStr) {
            if (qc.searchStr.length() < 3)
                Utils.throwIncorrectSpecException("Search string when specified should be atleast 3 char");
            qc.searchStr = qc.searchStr.toLowerCase();
            Set<String> searchTags = Utils.getQSTags(qc.searchStr);
            for (String tag : searchTags) {
                q = q.filter("qsTags", tag);
            }
        }

        List<RegistrationEntity> entities = q.list();

        List<RegistrationProp> props = new ArrayList<>();
        for (RegistrationEntity entity : entities)
            props.add(entity.toProp());

        if (qc.sortByFirstName)
            sortByFirstName(props);

        return props;
    }

    public static RegistrationProp queryByTransactionId(String client, String transactionId) {
        Client.ensureValid(client);

        Query<RegistrationEntity> q =
                ofy(client).load().type(RegistrationEntity.class).filter("transactionId",
                    transactionId);

        List<RegistrationEntity> entities = q.list();

        if (entities.size() == 0)
            return null;

        if (entities.size() > 1) {
            // should never happen
            logger.severe("Found [" + entities.size() + "] registration entities " + "with transaction id [" + transactionId + "]");
        }

        return entities.get(0).toProp();
    }

    public static RegistrationProp getByTransactionId(String client, String transactionId) {
        Client.ensureValid(client);

        TransactionEntity transactionEntity =
                ofy(client).load().type(TransactionEntity.class).id(transactionId).now();

        if (transactionEntity == null)
            return null;

        long registrationId = transactionEntity.registrationId;

        return safeGet(client, registrationId).toProp();
    }

    public static RegistrationStatusProp queryRegistrationStatus(String client, long programId,
                                                                 String email, String firstName, String lastName) {

        Client.ensureValid(client);

        ContactProp contactDetailProp = new ContactProp();
        contactDetailProp.firstName = firstName;
        contactDetailProp.lastName = lastName;
        contactDetailProp.email = email;

        Contact.ensureFirstNameAndValidEmailSpecified(contactDetailProp);

        Long registrationId = getMatchingRegistrationId(client, programId, contactDetailProp);

        if (registrationId == null) {
            RegistrationStatusProp rsp = new RegistrationStatusProp();
            rsp.alreadyRegistered = false;
            rsp.hasAttemptedRegistrationBefore = false;
            return rsp;
        }

        RegistrationStatusProp rsp = new RegistrationStatusProp();
        rsp.hasAttemptedRegistrationBefore = true;

        RegistrationProp rp = Registration.safeGet(client, registrationId).toProp();
        rsp.lastRegistrationStatus = rp.getStatus();
        rsp.registrationStatusTimestamp = rp.getStatusDate();

        if (rsp.lastRegistrationStatus == RegistrationStatus.REGISTRATION_COMPLETE)
            rsp.alreadyRegistered = true;

        return rsp;
    }

    public static RegistrationProp queryByIdOrEmail(String client, String registrationIdOrEmail, Long programId, String login) {
        RegistrationProp registrationProp;
        if (Utils.canParseAsLong(registrationIdOrEmail)) {
            long registrationId = Long.parseLong(registrationIdOrEmail);
            registrationProp = Registration.safeGet(client, registrationId).toProp();
        } else {
            AssertUtils.ensureNotNull(programId, "Program ID should be specified");

            RegistrationQueryCondition rqc = new RegistrationQueryCondition();
            rqc.programId = programId;
            rqc.email = registrationIdOrEmail;
            List<RegistrationProp> registrationProps = Registration.query(client, rqc, login);

            AssertUtils.ensure(registrationProps.size() == 1, "Error querying for registrations (got " + registrationProps.size() + " props)");
            registrationProp = registrationProps.get(0);
        }

        return registrationProp;
    }

    public static RegistrationSummaryProp getSummary(String client, long programId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        ProgramProp programProp = Program.safeGet(client, programId).toProp(client);

        RegistrationSummaryProp rsp = new RegistrationSummaryProp();

        RegistrationQueryCondition qc = new RegistrationQueryCondition();
        qc.programId = programId;

        List<RegistrationProp> registrationProps = query(client, qc, login);

        for (RegistrationProp registrationProp : registrationProps) {
            RegistrationStatus status = registrationProp.getStatus();

            if (status == RegistrationStatus.REGISTRATION_COMPLETE) {
                rsp.numCompleted++;
            } else {
                if ((status != RegistrationStatus.CANCELLED) && (status != RegistrationStatus.INVALID)) {
                    rsp.numStartedButNotCompleted++;
                }
            }

            if (!rsp.regStatusVsNum.containsKey(status))
                rsp.regStatusVsNum.put(status, 1);
            else {
                int current = rsp.regStatusVsNum.get(status);
                rsp.regStatusVsNum.put(status, current + 1);
            }

            int yyyymmdd = DateUtils.toYYYYMMDD(registrationProp.getStatusDate());

            if (status == RegistrationStatus.REGISTRATION_COMPLETE) {

                if (!rsp.yyyymmddVsNumCompleted.containsKey(yyyymmdd))
                    rsp.yyyymmddVsNumCompleted.put(yyyymmdd, 1);
                else {
                    int current = rsp.yyyymmddVsNumCompleted.get(yyyymmdd);
                    rsp.yyyymmddVsNumCompleted.put(yyyymmdd, current + 1);
                }

            } else {
                if (!rsp.yyyymmddVsNumStartedButNotCompleted.containsKey(yyyymmdd))
                    rsp.yyyymmddVsNumStartedButNotCompleted.put(yyyymmdd, 1);
                else {
                    int current = rsp.yyyymmddVsNumStartedButNotCompleted.get(yyyymmdd);
                    rsp.yyyymmddVsNumStartedButNotCompleted.put(yyyymmdd, current + 1);
                }
            }
        }

        if (client.equals(IshaConfig.getClient())) {
            rsp.stats = IshaSMM.update(programProp, login);
        }

        return rsp;
    }

    public static void transfer(String client, long registrationId, long programId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        Program.safeGet(client, programId);
        RegistrationEntity entity = safeGet(client, registrationId);

        long oldProgramId = entity.programId;
        entity.programId = programId;
        ofy(client).save().entity(entity);

        logger.info("Registration ID [" + registrationId + "] transferred from Program ID " + oldProgramId + "to Program ID " + programId);
    }

    public static void overrideAsCompleted(String client, long registrationId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        RegistrationEntity entity = safeGet(client, registrationId);

        TokenProp tokenProp = new TokenProp();
        tokenProp.client = client;
        tokenProp.paymentType = PaymentType.PROGRAM_REGISTRATION;
        tokenProp.successCallback = null;
        tokenProp.errorCallback = null;
        tokenProp.uniqueId = registrationId;

        IPaymentResponse paymentResponse = PaymentResponseFactory.getImpl(tokenProp);
        Map<String, String> response = new HashMap<>();
        response.put("TRANSACTIONID", "REFER2PAYPALTXNID");
        response.put("AMT", entity.amount);
        response.put("CURRENCYCODE", entity.ccy);
        response.put("PAYMENTSTATUS", "Success");
        paymentResponse.handleDoExpressCheckoutResponse(response);

        logger.info("Registration ID [" + registrationId + "] overriden as COMPLETE");
    }

    public static void invalidate(String client, long registrationId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        RegistrationEntity entity = safeGet(client, registrationId);

        entity.recordStateChange(RegistrationStatus.INVALID);
        ofy(client).save().entity(entity);

        logger.info("Registration ID [" + registrationId + "] INVALIDATED");
    }

    public static RegistrationProp cancel(String client, long registrationId, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);
        RegistrationEntity entity = safeGet(client, registrationId);

        entity.recordStateChange(RegistrationStatus.CANCELLED);
        ofy(client).save().entity(entity);

        logger.info("Registration ID [" + registrationId + "] CANCELLED");

        return entity.toProp();
    }

    public static void updateCheckinStatus(String client, List<RegistrationProp> registrationProps) {

        List<Long> memberIds = new ArrayList<>();
        for (RegistrationProp registrationProp : registrationProps) {
            memberIds.add(registrationProp.memberId);
        }

        Map<Long, MemberEntity> memberEntityMap = MemberLoader.get(client, memberIds, User.SUPER_USER);
        for (RegistrationProp registrationProp : registrationProps) {
            MemberEntity memberEntity = memberEntityMap.get(registrationProp.memberId);
            if (memberEntity == null) {
                logger.warning("Member not found for RID " + registrationProp.registrationId + ", MID " + registrationProp.memberId);
            } else {
                MemberProp memberProp = memberEntity.toProp();
                registrationProp.checkedIn = memberProp.programIds.contains(registrationProp.programId);
            }
        }
    }

    public static void handlePaymentAuthorization(String client, long registrationId) {
        Client.ensureValid(client);

        safeGet(client, registrationId);

        RegistrationEntity entity = safeGet(client, registrationId);
        entity.recordStateChange(RegistrationStatus.PAYMENT_AUTHORIZED);
        ofy(client).save().entity(entity).now();
    }

    public static void handlePostRegistration(String client, RegistrationProp registrationProp, boolean sendEmail) {

        Member.updateRegistration(client, registrationProp);

        try {
            if (client.equals(IshaConfig.getClient())) {
                IshaRegistration.handlePostRegistration(registrationProp);
            }

            if (sendEmail) {
                Registration.sendConfirmationEmail(client, Utils.getList(registrationProp), true);
            }

        } catch (Exception ex) {
            Utils.sendAlertEmailToDevTeam(ex, new RequestInfo().client(null));
        }
    }

    public static MailContentProp getRegistrationMailContent(String client, ProgramProp programProp) {

        String mailContentName = MailContent.ReservedMailContentName.RESERVED_REGISTRATION_CONFIRMATION.toString();

        // check if mailContent exists at Program level
        MailContentEntity mailContentEntity =
            MailContent.getByName(client, mailContentName + "_PID." + programProp.programId, programProp.groupProp.groupId);
        if (mailContentEntity != null) {
            return mailContentEntity.toProp();
        }

        // check if mailContent exists at ProgramType level
        mailContentEntity =
            MailContent.getByName(client, mailContentName + "_" + programProp.programTypeProp.name.toUpperCase(),
                programProp.groupProp.groupId);
        if (mailContentEntity != null) {
            return mailContentEntity.toProp();
        }

        // check if mailContent exists at Group level
        mailContentEntity = MailContent.getByName(client, mailContentName, programProp.groupProp.groupId);
        if (mailContentEntity != null) {
            return mailContentEntity.toProp();
        }

        // check if mailContent exists at Client level
        mailContentEntity = MailContent.getByName(client, mailContentName, 0);
        if (mailContentEntity != null) {
            return mailContentEntity.toProp();
        }

        return null;
    }

    public static void sendConfirmationEmail(String client, List<RegistrationProp> registrationProps, boolean suppressIfAlreadySent)
        throws APIException {

        MailMap mailMap = new MailMap();
        ProgramProp programProp = Program.safeGet(client, registrationProps.get(0).programId).toProp(client);

        MailContentProp mailContentProp = getRegistrationMailContent(client, programProp);
        if (mailContentProp == null) {
            throw new APIException(
                "RESERVED_REGISTRATION_CONFIRMATION MailContent " + "(PID " + programProp.programId + ") is not present")
                .status(Status.ERROR_PRECONDITION_FAILED);
        }

        SimpleDateFormat ftSrc = new SimpleDateFormat("yyyyMMdd");
        SimpleDateFormat ftDst = new SimpleDateFormat("MMM d");

        String subject = "Registration - " + programProp.programTypeProp.displayName;
        if (programProp.description != null) subject += " " + programProp.description;

        try {
            String startDate = ftDst.format(ftSrc.parse(String.valueOf(programProp.startYYYYMMDD)));
            String endDate = ftDst.format(ftSrc.parse(String.valueOf(programProp.endYYYYMMDD)));
            String dates = (startDate.equalsIgnoreCase(endDate) ? startDate : startDate + " - " + endDate);

            subject +=  ": " + dates;
            mailMap.setGlobalMergeVar(MailMap.MergeVarID.DATES, dates);
        } catch (ParseException ex) {
            logger.warning(ex.toString());
        }

        for (RegistrationProp registrationProp : registrationProps) {

            logger.info("Sending confirmation email: "
                + "(PID " + programProp.programId + ") "
                + "(MCID " + mailContentProp.mailContentId + ") "
                + "(RID " + registrationProp.registrationId + ")");

            mailMap.add(registrationProp.email, registrationProp.firstName, registrationProp.lastName);

            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.REGISTRATION_ID, registrationProp.registrationId + "");
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.EMAIL, registrationProp.email);
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.MOBILE_PHONE, registrationProp.mobilePhone);
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.HOME_PHONE, registrationProp.homePhone);
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.OFFICE_PHONE, registrationProp.officePhone);
            if (registrationProp.transactionId != null) {
                mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.TRANSACTION_ID,
                    registrationProp.transactionId.equals("REFER2PAYPALTXNID") ? "" : registrationProp.transactionId);
            }
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.INVOICE_NUMBER,
                Invoice.getInvoiceNo(programProp.programTypeProp.displayName, programProp.groupProp.displayName, registrationProp.registrationId));
            mailMap.setMergeVar(registrationProp.email, MailMap.MergeVarID.AMOUNT, programProp.ccy + " " + registrationProp.amount);

            if (client.equals(IshaConfig.getClient())) {
                IshaRegistration.update(mailMap, registrationProp);
            }
        }

        if (programProp.batch1SessionTimings != null) {
            List<ProgramSession> programSessions = new ArrayList<>();
            for (String sessionTimings : programProp.batch1SessionTimings) {

                ProgramSession programSession = new ProgramSession();
                String session = sessionTimings.replaceAll("Z", "");
                programSession.mandatory = session.indexOf('M') >= 0;

                session = session.replaceAll("M", "");

                String[] timings = session.split(">");

                ftSrc = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm");
                SimpleDateFormat ftDay = new SimpleDateFormat("EEE d MMM");
                SimpleDateFormat ftTime = new SimpleDateFormat("h:mm a");

                try {
                    Date dtStart = ftSrc.parse(String.valueOf(timings[0]));
                    programSession.day = ftDay.format(dtStart);
                    programSession.start_time = ftTime.format(dtStart);
                    programSession.end_time = ftTime.format(ftSrc.parse(String.valueOf(timings[1])));
                } catch (ParseException ex) {
                    logger.warning(ex.toString());
                }

                programSessions.add(programSession);
            }
            mailMap.setGlobalMergeVar(MailMap.MergeVarID.SESSIONS, programSessions);
        }

        mailMap.setGlobalMergeVar(MailMap.MergeVarID.VENUE, programProp.venueProp.displayName);
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.ADDRESS, programProp.venueProp.address);
        if (!Strings.isNullOrEmpty(programProp.venueProp.mapLink)) {
            mailMap.setGlobalMergeVar(MailMap.MergeVarID.MAP_LINK, programProp.venueProp.mapLink);
        }
        
        mailMap.setGlobalMergeVar(MailMap.MergeVarID.PROGRAM_NAME, programProp.programTypeProp.displayName);

        MailSendInput msi = new MailSendInput();
        msi.createMember = false;
        msi.groupId = programProp.groupProp.groupId;
        msi.isTransactionEmail = true;
        msi.mailContentId = mailContentProp.mailContentId;
        msi.senderEmail = Group.getSenderEmailByProgramType(
            client, programProp.groupProp.groupId, programProp.programTypeProp.programTypeId, User.SUPER_USER);
        msi.suppressIfAlreadySent = suppressIfAlreadySent;
        msi.overrideSubject = subject;

        try {
            Mail.send(client, msi, mailMap, User.SUPER_USER);
        } catch (MandrillApiError | IOException ex) {
            logger.warning(ex.toString());
        }
    }

    static void sortByFirstName(List<RegistrationProp> props) {
        Collections.sort(props, new Comparator<RegistrationProp>() {

            @Override
            public int compare(RegistrationProp o1, RegistrationProp o2) {
                if ((o1.firstName == null) || (o2.firstName == null))
                    return 0;

                return o1.firstName.compareTo(o2.firstName);
            }
        });
    }

    public enum RegistrationStatus {
        REQUEST_RECEIVED, PAYMENT_AUTHORIZATION_PENDING, ERROR_CREATING_PAYMENT_URL, PAYMENT_NOT_AUTHORIZED, PAYMENT_AUTHORIZED, PAYMENT_PENDING, REGISTRATION_COMPLETE, PAYPAL_ERROR, INVALID, CANCELLED
    }


    public static class ProgramSession {
        String day;
        String start_time;
        String end_time;
        boolean mandatory;
    }
}
