package crmdna.user;

import com.googlecode.objectify.ObjectifyFilter;
import com.googlecode.objectify.Work;
import crmdna.client.Client;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse.Status;
import crmdna.group.Group;
import crmdna.member.MemberLoader;
import crmdna.sequence.Sequence;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static crmdna.common.AssertUtils.ensure;
import static crmdna.common.AssertUtils.ensureNotNull;
import static crmdna.common.OfyService.ofy;

public class User {

    public static final String SUPER_USER = "sathyanarayanant@gmail.com";

    static final String DELIMITER = "||";

    public static UserProp create(String client, String email, long groupId, String login) {
        Client.ensureValid(client);

        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);
        Group.safeGet(client, groupId);

        Utils.ensureValidEmail(email);
        email = email.toLowerCase();

        UserEntity entity = ofy(client).load().type(UserEntity.class).id(email).now();

        if (null != entity)
            throw new APIException().status(Status.ERROR_RESOURCE_ALREADY_EXISTS).message(
                "There is already an user with email [" + email + "]");

        UserEntity userEntity = new UserEntity();
        userEntity.email = email;
        userEntity.userId = Sequence.getNext(client, Sequence.SequenceType.USER);
        userEntity.groupId = groupId;

        ofy(client).save().entity(userEntity).now();

        return userEntity.toProp(client);
    }

    public static UserProp updateEmail(String client, String existingEmail, String newEmail,
                                       String login) {

        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        Utils.ensureValidEmail(newEmail);

        existingEmail = existingEmail.toLowerCase();
        newEmail = newEmail.toLowerCase();

        UserEntity existingEntity = User.safeGet(client, existingEmail);

        if (existingEmail.equals(newEmail))
            return existingEntity.toProp(client);

        // newEmail should not already exist
        if (null != get(client, newEmail))
            throw new APIException().status(Status.ERROR_RESOURCE_ALREADY_EXISTS).message(
                "There is already an user with email [" + newEmail + "]");

        UserEntity newEntity = new UserEntity();
        newEntity.email = newEmail;
        newEntity.userId = existingEntity.userId;
        newEntity.groupId = existingEntity.groupId;
        newEntity.privileges = existingEntity.privileges;

        ofy(client).save().entity(newEntity);
        ofy(client).delete().entity(existingEntity);
        ObjectifyFilter.complete();

        return newEntity.toProp(client);
    }

    public static UserProp updateGroup(final String client, final String email, final long newGroupId, String login) {

        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        Group.safeGet(client, newGroupId);

        safeGet(client, email.toLowerCase());

        return ofy(client).transact(new Work<UserProp>() {
            @Override
            public UserProp run() {
                UserEntity userEntity =
                    ofy(client).load().type(UserEntity.class).id(email.toLowerCase()).now();
                if (null == userEntity)
                    throw new APIException().status(Status.ERROR_RESOURCE_NOT_FOUND).message(
                        "There is no user with email [" + email + "]");

                userEntity.groupId = newGroupId;
                ofy(client).save().entity(userEntity).now();
                return userEntity.toProp(client);
            }
        });
    }

    public static UserProp linkMember(final String client, final String email, final long newMemberId, String login) {

        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        MemberLoader.safeGet(client, newMemberId, login);

        // email should be present
        safeGet(client, email.toLowerCase());

        return ofy(client).transact(new Work<UserProp>() {
            @Override
            public UserProp run() {
                UserEntity userEntity =
                    ofy(client).load().type(UserEntity.class).id(email.toLowerCase()).now();
                if (null == userEntity)
                    throw new APIException().status(Status.ERROR_RESOURCE_NOT_FOUND).message(
                        "There is no user with email [" + email + "]");

                userEntity.memberId = newMemberId;
                ofy(client).save().entity(userEntity).now();
                return userEntity.toProp(client);
            }
        });
    }

    public static UserEntity get(String client, String email) {

        Client.ensureValid(client);
        return ofy(client).load().type(UserEntity.class).id(email.toLowerCase()).now();
    }

    public static UserEntity safeGet(String client, String email) {

        Client.ensureValid(client);
        UserEntity userEntity = get(client, email);

        if (null == userEntity)
            throw new APIException("User [" + email + "] is not valid").status(Status.ERROR_INVALID_USER);

        return userEntity;
    }

    public static List<UserProp> getAll(String client, String login) {

        Client.ensureValid(client);
        User.ensureValidUser(client, login);

        Client.ensureValid(client);

        List<UserEntity> entities = ofy(client).load().type(UserEntity.class).order("__key__").list();

        List<UserProp> props = new ArrayList<>();
        for (UserEntity entity : entities) {
            props.add(entity.toProp(client));
        }

        return props;
    }

    public static UserEntity safeGetById(String client, long userId) {
        Client.ensureValid(client);

        List<UserEntity> entities = ofy(client).load().type(UserEntity.class).filter("userId", userId).list();
        if (entities.isEmpty()) {
            throw new APIException().message("Cannot find user with id [" + userId + "]")
                    .status(Status.ERROR_RESOURCE_NOT_FOUND);
        }

        ensure(entities.size() == 1,
            "Found [" + entities.size() + "] user entities with id [" + userId + "]");
        return entities.get(0);
    }

    public static List<UserProp> getAllForGroup(String client, long groupId, String login) {
        Client.ensureValid(client);
        List<UserProp> all = getAll(client, login);

        Group.GroupProp groupProp = Group.safeGet(client, groupId).toProp();

        List<UserProp> userProps = new ArrayList<>();
        for (UserProp userProp : all) {
            if (userProp.groupId == groupId) {
                userProps.add(userProp);
            } else if (userProp.groupLevelPrivileges.get(groupProp.displayName) != null) {
                userProps.add(userProp);
            }
        }

        return userProps;
    }

    public static UserProp addClientLevelPrivilege(String client, String email, ClientLevelPrivilege privilege, String login) {
        return addOrDeleteClientLevelPrivilege(client, email, privilege, true, login);
    }

    public static UserProp deleteClientLevelPrivilege(String client, String email, ClientLevelPrivilege privilege, String login) {

        return addOrDeleteClientLevelPrivilege(client, email, privilege, false, login);
    }

    private static UserProp addOrDeleteClientLevelPrivilege(String client, String email,
                                                            ClientLevelPrivilege privilege, boolean add, String login) {
        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        return addOrDeletePrivilege(client, email, "CLIENT", client, privilege.toString(), add);
    }

    public static UserProp addGroupLevelPrivilege(String client, long groupId, String email, GroupLevelPrivilege privilege, String login) {

        return addOrDeleteGroupLevelPrivilege(client, groupId, email, privilege, true, login);
    }

    public static UserProp deleteGroupLevelPrivilege(String client, long groupId, String email,
                                                     GroupLevelPrivilege privilege, String login) {

        return addOrDeleteGroupLevelPrivilege(client, groupId, email, privilege, false, login);
    }

    private static UserProp addOrDeleteGroupLevelPrivilege(String client, long groupId, String email,
                                                           GroupLevelPrivilege privilege, boolean add, String login) {
        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        return addOrDeletePrivilege(client, email, "GROUP", groupId + "", privilege.toString(), add);
    }

    public static UserProp clonePrivileges(final String client, final String sourceEmail, final String targetEmail, String login) {

        Client.ensureValid(client);
        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        if (null == sourceEmail)
            Utils.throwIncorrectSpecException("sourceUser cannot be null");

        if (null == targetEmail)
            Utils.throwIncorrectSpecException("targetUser cannot be null");

        safeGet(client, sourceEmail.toLowerCase());
        safeGet(client, targetEmail.toLowerCase());

        return ofy(client).transact(new Work<UserProp>() {

            @Override
            public UserProp run() {
                UserEntity sourceEntity =
                    ofy(client).load().type(UserEntity.class).id(sourceEmail.toLowerCase()).now();
                if (null == sourceEntity)
                    throw new APIException().status(Status.ERROR_RESOURCE_NOT_FOUND).message(
                        "Source email [" + sourceEmail + "] does not exist");

                UserEntity targetEntity =
                    ofy(client).load().type(UserEntity.class).id(targetEmail.toLowerCase()).now();
                if (null == targetEntity)
                    throw new APIException().status(Status.ERROR_RESOURCE_NOT_FOUND).message(
                        "Target email [" + targetEmail + "] does not exist");

                targetEntity.privileges = sourceEntity.privileges;
                ofy(client).save().entity(targetEntity).now();
                return targetEntity.toProp(client);
            }
        });
    }

    public static void ensureValidUser(String client, String login) {

        Client.ensureValid(client);
        ensureNotNull(client, "client is null");
        ensure(!client.isEmpty(), "client is empty");

        if (null == login)
            throw new APIException().status(Status.ERROR_LOGIN_REQUIRED).message("Email cannot be null or empty");

        if (User.isSuperUser(login))
            return;

        login = login.toLowerCase();
        UserEntity userEntity = ofy(client).load().type(UserEntity.class).id(login).now();

        if (null == userEntity)
            throw new APIException("Invalid user [" + login + "] for client [" + client + "]").status(Status.ERROR_INVALID_USER);
    }

    public static void ensureClientLevelPrivilege(String client, String login,
                                                  ClientLevelPrivilege privilege) {

        Client.ensureValid(client);

        if (hasPrivilege(client, login, "CLIENT", client, privilege.toString()))
            return;

        // throw exception
        String message =
                "User [" + login + "] does not have client level privilege [" + privilege
                        + "] for client [" + client + "]";
        throw new APIException(message).status(Status.ERROR_INSUFFICIENT_PERMISSION);
    }

    public static boolean hasPrivilege(String client, String login, ClientLevelPrivilege privilege) {

        Client.ensureValid(client);
        return hasPrivilege(client, login, "CLIENT", client, privilege.toString());
    }

    public static void ensureGroupLevelPrivilege(String client, long groupId, String login,
                                                 GroupLevelPrivilege privilege) {

        Client.ensureValid(client);

        if (hasPrivilege(client, login, "GROUP", groupId + "", privilege.toString()))
            return;

        // throw exception
        String message =
                "User [" + login + "] does not have group level privilege [" + privilege + "] for client ["
                        + client + "], group [" + groupId + "]";
        throw new APIException(message).status(Status.ERROR_INSUFFICIENT_PERMISSION);
    }

    public static UserProp addApp(String client, String email, App app, String login) {
        Client.ensureValid(client);

        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        return addOrDeletePrivilege(client, email, "APP", app.toString(), app.toString(), true);
    }

    public static UserProp removeApp(String client, String email, App app, String login) {
        Client.ensureValid(client);

        ensureClientLevelPrivilege(client, login, ClientLevelPrivilege.UPDATE_USER);

        return addOrDeletePrivilege(client, email, "APP", app.toString(), app.toString(), false);
    }

    static UserProp addOrDeletePrivilege(final String client, final String email,
        final String resourceType, final String resourceId, final String privilege, final boolean add) {

        safeGet(client, email.toLowerCase()); // ensures user is valid

        return ofy(client).transact(new Work<UserProp>() {

            @Override
            public UserProp run() {
                UserEntity userEntity =
                    ofy(client).load().type(UserEntity.class).id(email.toLowerCase()).now();

                if (null == userEntity)
                    throw new APIException().status(Status.ERROR_INVALID_USER).message(
                        "Invalid user [" + email + "]");

                String rawprivilege = getRawPrivilege(resourceType, resourceId, privilege);

                boolean change;
                if (add)
                    change = userEntity.privileges.add(rawprivilege);
                else
                    change = userEntity.privileges.remove(rawprivilege);

                if (change)
                    ofy(client).save().entity(userEntity).now();

                return userEntity.toProp(client);
            }
        });
    }

    static boolean hasPrivilege(String client, String login, String resourceType, String resourceId, String privilege) {

        ensureNotNull(client, "client is null");
        ensure(!client.isEmpty(), "client is empty");

        if (null == login)
            throw new APIException().status(Status.ERROR_LOGIN_REQUIRED).message(
                "Email cannot be null or empty");

        if (isSuperUser(login))
            return true;

        login = login.toLowerCase();
        UserEntity userEntity = ofy(client).load().type(UserEntity.class).id(login).now();

        if (null == userEntity)
            throw new APIException("Invalid user [" + login + "] for client [" + client + "]")
                .status(Status.ERROR_INVALID_USER);

        return userEntity.privileges.contains(getRawPrivilege(resourceType, resourceId, privilege));
    }

    static String getRawPrivilege(String resourceType, String resourceId, String action) {
        return resourceType + DELIMITER + resourceId + DELIMITER + action;
    }

    public static boolean isSuperUser(String login) {
        if (login == null) {
            return false;
        }

        String[] superUsers = {"sathyanarayanant@gmail.com", "thulasidhar@gmail.com"};
        return Arrays.asList(superUsers).contains(login.toLowerCase());
    }

    public enum ResourceType {
        MEMBER,
        GROUP,
        USER,
        INTERACTION,
        PRACTICE,
        PROGRAM_TYPE,
        VENUE,
        TEACHER,
        PROGRAM,
        CUSTOM_CONFIG,
        PAYPAL_CREDENTIALS,
        FROM_EMAIL,
        EMAIL_KEY,
        CONFIG,
        INVENTORY_ITEM_TYPE,
        INVENTORY_ITEM,
        DEPARTMENT,
        LIST,
        MAIL_CONTENT,
        MEAL
    }

    public enum GroupLevelPrivilege {
        SEND_EMAIL,
        VIEW_LIST,
        UPDATE_LIST,
        UPDATE_PROGRAM,
        CHECK_IN,
        UPDATE_EMAIL_CONFIG,
        UPDATE_PAYMENT_CONFIG,
        UPDATE_INVENTORY_QUANTITY,
        UPDATE_INVENTORY_ITEM,
        UPDATE_MAIL_CONTENT,
        UPDATE_PARTICIPANT,
        UPDATE_CAMPAIGN,
        UPDATE_MAIL_SEQUENCE,
        ASSIGN_CALLER
    }

    public enum ClientLevelPrivilege {
        UPDATE_GROUP,
        ENABLE_DISABLE_ACCOUNT,
        UPDATE_USER,
        UPDATE_CLIENT_CONTACT_EMAIL,
        PURGE_MEMBER_DATA,
        VIEW_API_KEY,
        UPDATE_CUSTOM_CONFIG,
        UPDATE_DEPARTMENT,
        UPDATE_INTERACTION,
        UPDATE_INVENTORY_ITEM_TYPE,
        UPDATE_MEAL_COUNT,
        UPDATE_MAIL_CONTENT,
        VERIFY_EMAIL,
        UPDATE_PAYMENT,
        UPDATE_PRACTICE,
        UPDATE_PROGRAM_TYPE,
        UPDATE_SESSION_PASS,
        UPDATE_TEACHER,
        UPDATE_VENUE,
        VIEW_FINANCIALS,
        SUBSCRIBE_GROUP, //required to undo unsubscribe. should be given only to select people
        IMPERSONATE_USER
    }

    public enum App {
        SEARCH, CHECK_IN, INVENTORY
    }
}
