package crmdna.interaction;

import com.google.appengine.tools.development.testing.LocalDatastoreServiceTestConfig;
import com.google.appengine.tools.development.testing.LocalServiceTestHelper;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.google.gson.Gson;
import com.googlecode.objectify.ObjectifyFilter;
import crmdna.calling.Campaign;
import crmdna.calling.CampaignProp;
import crmdna.client.Client;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.contact.ContactProp;
import crmdna.group.Group;
import crmdna.member.Member;
import crmdna.member.MemberProp;
import crmdna.practice.Practice;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.programtype.ProgramType;
import crmdna.programtype.ProgramTypeProp;
import crmdna.teacher.Teacher;
import crmdna.user.User;
import crmdna.user.UserProp;
import crmdna.venue.Venue;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

/**
 * Created by sathya on 17/12/15.
 */
public class CallingTest {

    private final LocalServiceTestHelper datastoreHelper = new LocalServiceTestHelper(
            new LocalDatastoreServiceTestConfig().setApplyAllHighRepJobPolicy());

    String client;
    String invalidClient = "invalid";
    Group.GroupProp sgp, kl;
    UserProp validUser, userWithAssignCallerPermissionInSgp;

    ProgramProp ishaKriya20150823, ishaKriya20150830;

    MemberProp kumaraguru, murugavel, thulasiram, sharmila;

    @Before
    public void setUp() {
        datastoreHelper.setUp();
        ObjectifyFilter.complete();

        client = "isha";
        Client.create(client);

        sgp = crmdna.group.Group.create(client, "Singapore", User.SUPER_USER);
        assertEquals(1, sgp.groupId);

        kl = Group.create(client, "KL", User.SUPER_USER);
        assertEquals(2, kl.groupId);

        validUser = User.create(client, "validuser@dummy.com", sgp.groupId, User.SUPER_USER);
        assertEquals(1, User.get(client, validUser.email).toProp(client).userId);

        userWithAssignCallerPermissionInSgp = User.create(client, "withpermission@dummy.com", sgp.groupId, User.SUPER_USER);
        assertEquals(2, User.get(client, userWithAssignCallerPermissionInSgp.email).toProp(client).userId);

        User.addGroupLevelPrivilege(client, sgp.groupId, userWithAssignCallerPermissionInSgp.email,
                User.GroupLevelPrivilege.ASSIGN_CALLER,
                User.SUPER_USER);

        Venue.VenueProp woodlandsCC =
                Venue.create(client, "Woodlands CC", "Woodlands CC", "Woodlands CC", "", sgp.groupId, User.SUPER_USER);
        Teacher.TeacherProp thulasi = Teacher.create(client, "", "", "thulasidhar@gmail.com", sgp.groupId, User.SUPER_USER);

        Practice.PracticeProp ishaKriya = Practice.create(client, "Isha Kriya", User.SUPER_USER);

        ProgramTypeProp ishaKriyaTeacherLed = ProgramType.create(client, "Isha Kriya (Teacher led)",
                Utils.getSet(ishaKriya.practiceId), false, User.SUPER_USER);

        ishaKriya20150823 = Program.create(client, sgp.groupId, ishaKriyaTeacherLed.programTypeId,
                woodlandsCC.venueId, thulasi.teacherId, 20150823, 20150823, 1, null, 0, null, User.SUPER_USER);

        ishaKriya20150830 = Program.create(client, sgp.groupId, ishaKriyaTeacherLed.programTypeId,
                woodlandsCC.venueId, thulasi.teacherId, 20150830, 20150830, 1, null, 0, null, User.SUPER_USER);

        ContactProp c = new ContactProp();
        c.email = "kumarakuru@dummy.com";
        c.asOfyyyymmdd = 20151217;
        kumaraguru = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.subscribeGroup(client, kumaraguru.memberId, sgp.groupId, User.SUPER_USER);

        c.email = "murugavel@dummy.com";
        murugavel = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.subscribeGroup(client, murugavel.memberId, sgp.groupId, User.SUPER_USER);

        c.email = "thulasiram@dummy.com";
        thulasiram = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.subscribeGroup(client, thulasiram.memberId, sgp.groupId, User.SUPER_USER);

        c.email = "sharmila@dummy.com";
        sharmila = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.subscribeGroup(client, sharmila.memberId, sgp.groupId, User.SUPER_USER);
    }

    @After
    public void tearDown() {
        ObjectifyFilter.complete();
        datastoreHelper.tearDown();
    }

    @Test
    public void testMultipleCallsNotAssignedForSameMember() {

        CampaignProp campaignProp = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
                20150801, 20150823, User.SUPER_USER);

        //getAlreadyAssignedMembers returns empty set with no assignment
        Set<Long> memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp.campaignId);
        assertTrue(memberIds.isEmpty());

        //assign some calls
        CallAssignmentResult result = Calling.assign(client, campaignProp.campaignId,
                Sets.newHashSet(kumaraguru.memberId, murugavel.memberId),
                Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);

        assertTrue(result.allMembersAssigned());
        assertTrue(result.userIdVsMemberIds.containsKey(validUser.userId));
        assertTrue(result.userIdVsMemberIds.get(validUser.userId).containsAll(Sets.newHashSet(kumaraguru.memberId, murugavel.memberId)));

        //now getAlreadyAssignedMembers should get the assignments
        memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp.campaignId);
        assertTrue(memberIds.containsAll(Sets.newHashSet(kumaraguru.memberId, murugavel.memberId)));

        //try to assign again - no new calls should get assigned
        result = Calling.assign(client, campaignProp.campaignId,
                Sets.newHashSet(kumaraguru.memberId, murugavel.memberId),
                Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);
        assertTrue(result.userIdVsMemberIds.isEmpty());
    }

    @Test
    public void testMultipleCallsNotAssignedForSameProgram() {

        CampaignProp campaignProp1 = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
            20150801, 20150823, User.SUPER_USER);
        CampaignProp campaignProp2 = Campaign.create(client, ishaKriya20150823.programId, "Default2", "",
            20150801, 20150823, User.SUPER_USER);

        //getAlreadyAssignedMembers returns empty set with no assignment
        Set<Long> memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp1.campaignId);
        assertTrue(memberIds.isEmpty());

        //assign some calls
        CallAssignmentResult result = Calling.assign(client, campaignProp1.campaignId,
            Sets.newHashSet(kumaraguru.memberId, murugavel.memberId),
            Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);

        assertTrue(result.allMembersAssigned());
        assertTrue(result.userIdVsMemberIds.containsKey(validUser.userId));
        assertTrue(result.userIdVsMemberIds.get(validUser.userId).containsAll(Sets.newHashSet(kumaraguru.memberId, murugavel.memberId)));

        //assign some calls
        CallAssignmentResult result2 = Calling.assign(client, campaignProp2.campaignId,
            Sets.newHashSet(thulasiram.memberId, murugavel.memberId),
            Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);

        assertTrue(result2.allMembersAssigned());
        assertTrue(result2.userIdVsMemberIds.containsKey(validUser.userId));
        assertTrue(result2.userIdVsMemberIds.get(validUser.userId).containsAll(Sets.newHashSet(thulasiram.memberId)));

        //now getAlreadyAssignedMembers should get the assignments
        memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp1.campaignId);
        assertTrue(memberIds.containsAll(Sets.newHashSet(kumaraguru.memberId, murugavel.memberId, thulasiram.memberId)));

        memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp2.campaignId);
        assertTrue(memberIds.containsAll(Sets.newHashSet(thulasiram.memberId, murugavel.memberId, kumaraguru.memberId)));

        //try to assign again - no new calls should get assigned
        result = Calling.assign(client, campaignProp1.campaignId,
            Sets.newHashSet(kumaraguru.memberId, murugavel.memberId, thulasiram.memberId),
            Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);
        assertTrue(result.userIdVsMemberIds.isEmpty());
    }

    @Test
    public void cannotAssignWithoutPrivilege() {

        CampaignProp campaignProp = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
                20150801, 20150823, User.SUPER_USER);

        try {
            CallAssignmentResult result = Calling.assign(client, campaignProp.campaignId,
                    Sets.newHashSet(kumaraguru.memberId, murugavel.memberId),
                    Sets.newHashSet(validUser.userId), null, false, validUser.email);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_INSUFFICIENT_PERMISSION, ex.statusCode);
        }
    }

    @Test
    public void testNoCallsAssignedWhenPreviewing() {

        CampaignProp campaignProp = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
                20150801, 20150823, User.SUPER_USER);

        //preview
        CallAssignmentResult result = Calling.preview(client, campaignProp.campaignId,
                Sets.newHashSet(kumaraguru.memberId, murugavel.memberId),
                Sets.newHashSet(validUser.userId), null, false, userWithAssignCallerPermissionInSgp.email);

        //now getAlreadyAssignedMembers should return empty set
        Set<Long> memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp.campaignId);
        assertTrue(memberIds.isEmpty());
    }

    @Test
    public void testEqualAssignmentMultipleUsers() {

        CampaignProp campaignProp = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
                20150801, 20150823, User.SUPER_USER);

        //getAlreadyAssignedMembers returns empty set with no assignment
        Set<Long> memberIds = Calling.getAlreadyAssignedMemberIds(client, campaignProp.campaignId);
        assertTrue(memberIds.isEmpty());

        //assign some calls
        memberIds = Sets.newHashSet(kumaraguru.memberId, murugavel.memberId);
        Set<Long> userIds = Sets.newHashSet(validUser.userId, userWithAssignCallerPermissionInSgp.userId);
        Map<Long, Integer> userIdVsMaxMembers =
                CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());
        userIdVsMaxMembers.put(validUser.userId, 1);
        userIdVsMaxMembers.put(userWithAssignCallerPermissionInSgp.userId, 1);
        CallAssignmentResult result = Calling.assign(client, campaignProp.campaignId,
                memberIds, userIds,
                userIdVsMaxMembers, false, userWithAssignCallerPermissionInSgp.email);

        assertTrue(result.allMembersAssigned());
        assertEquals(2, result.userIdVsMemberIds.size());
        assertEquals(1, result.userIdVsMemberIds.get(validUser.userId).size());
        assertEquals(1, result.userIdVsMemberIds.get(userWithAssignCallerPermissionInSgp.userId).size());
    }

    @Test
    public void testOnlySubscribedMembersAssigned() {

        CampaignProp campaignProp = Campaign.create(client, ishaKriya20150823.programId, "Default", "",
                20150801, 20150823, User.SUPER_USER);


        //unsubscribe kumaraguru and murugavel from isha singapore
        Member.unsubscribeGroup(client, kumaraguru.memberId, sgp.groupId, User.SUPER_USER);
        Member.unsubscribeGroup(client, murugavel.memberId, sgp.groupId, User.SUPER_USER);

        //try to call kumuraguru and murugavel. no calls should get assigned as they are not subscribed
        Set<Long> memberIds = Sets.newHashSet(kumaraguru.memberId, murugavel.memberId);
        Set<Long> userIds = Sets.newHashSet(validUser.userId, userWithAssignCallerPermissionInSgp.userId);
        Map<Long, Integer> userIdVsMaxMembers =
                CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());

        CallAssignmentResult result = Calling.assign(client, campaignProp.campaignId,
                memberIds, userIds,
                userIdVsMaxMembers, false, userWithAssignCallerPermissionInSgp.email);
        assertTrue(result.userIdVsMemberIds.isEmpty());

        //subscribe them to group and assign now. calls should get assigned
        Member.subscribeGroup(client, kumaraguru.memberId, sgp.groupId, User.SUPER_USER);
        Member.subscribeGroup(client, murugavel.memberId, sgp.groupId, User.SUPER_USER);
        memberIds = Sets.newHashSet(kumaraguru.memberId, murugavel.memberId);
        userIds = Sets.newHashSet(validUser.userId, userWithAssignCallerPermissionInSgp.userId);
        userIdVsMaxMembers =
                CallAssignmentHelper.getUserIdVsMaxMembersForEqualSplit(userIds, memberIds.size());
        result = Calling.assign(client, campaignProp.campaignId,
                memberIds, userIds,
                userIdVsMaxMembers, false, userWithAssignCallerPermissionInSgp.email);
        assertEquals(2, result.userIdVsMemberIds.size());
        assertEquals(1, result.userIdVsMemberIds.get(validUser.userId).size());
        assertEquals(1, result.userIdVsMemberIds.get(userWithAssignCallerPermissionInSgp.userId).size());
    }
}
