package crmdna.mail2;

import com.google.appengine.tools.development.testing.LocalDatastoreServiceTestConfig;
import com.google.appengine.tools.development.testing.LocalServiceTestHelper;
import com.googlecode.objectify.ObjectifyFilter;
import com.microtripit.mandrillapp.lutung.model.MandrillApiError;
import crmdna.client.Client;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.contact.ContactProp;
import crmdna.group.Group;
import crmdna.group.Group.GroupProp;
import crmdna.list.List;
import crmdna.list.ListProp;
import crmdna.member.Member;
import crmdna.member.MemberEntity;
import crmdna.member.MemberLoader;
import crmdna.member.MemberProp;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.programtype.ProgramType;
import crmdna.programtype.ProgramTypeProp;
import crmdna.teacher.Teacher;
import crmdna.user.User;
import crmdna.user.User.GroupLevelPrivilege;
import crmdna.user.UserProp;
import crmdna.venue.Venue;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.io.IOException;
import java.util.Date;
import java.util.Map;

import static org.junit.Assert.*;

public class MailScheduleTest {
    private final LocalServiceTestHelper datastoreHelper = new LocalServiceTestHelper(
            new LocalDatastoreServiceTestConfig().setApplyAllHighRepJobPolicy());

    private final String client = "isha";
    private GroupProp sgp;
    private GroupProp kl;

    private UserProp validUser;
    private UserProp userWithSgpEmailPrivilage;

    private MailContentProp newsletter;
    private MailContentProp welcome;
    private ListProp newsletterListSgp;
    private ListProp ishakriyaListSgp;
    private ListProp newsletterListKl;

    private ProgramProp ishakriya20151004;

    private MemberProp sharmila, patma;

    @Before
    public void setUp() {
        datastoreHelper.setUp();
        ObjectifyFilter.complete();

        Client.create(client);
        sgp = Group.create(client, "Singapore", User.SUPER_USER);
        assertEquals(1, sgp.groupId);

        kl = Group.create(client, "KL", User.SUPER_USER);
        assertEquals(2, kl.groupId);

        Group.addOrDeleteAllowedEmailSender(client, sgp.groupId,
                "singapore@ishayoga.org", "Isha Singapore", true,
                User.SUPER_USER);

        userWithSgpEmailPrivilage = User.create(client, "userwithsgpemail@invalid.com", sgp.groupId,
                User.SUPER_USER);
        User.addGroupLevelPrivilege(client, sgp.groupId, userWithSgpEmailPrivilage.email,
                GroupLevelPrivilege.SEND_EMAIL, User.SUPER_USER);

        validUser = User.create(client, "validuser@invalid.com", sgp.groupId,
                User.SUPER_USER);

        newsletter = MailContent.create(client, "Newsletter", sgp.groupId, "Mail content subject",
                "Mail content body", User.SUPER_USER);
        welcome = MailContent.create(client, "Welcome", sgp.groupId, "Welcome", "Welcome", User.SUPER_USER);

        newsletterListSgp = List.createPublic(client, sgp.groupId, "Newsletter sgp list", User.SUPER_USER);
        ishakriyaListSgp = List.createPublic(client, sgp.groupId, "Isha kriya sgp list", User.SUPER_USER);
        newsletterListKl = List.createPublic(client, kl.groupId, "Newsletter kl list", User.SUPER_USER);

        ProgramTypeProp ishakriya = ProgramType.create(client, "Isha Kriya", null, false, User.SUPER_USER);
        Teacher.TeacherProp shilpi = Teacher.create(client, "Shilpi", "Malhotra", "s1@gmail.com"
                , sgp.groupId, User.SUPER_USER);
        Venue.VenueProp woodlands = Venue.create(client, "Woodlands", "w", "w", "", sgp.groupId, User.SUPER_USER);
        ishakriya20151004 = Program.create(client, sgp.groupId, ishakriya.programTypeId,
                woodlands.venueId, shilpi.teacherId, 20151004, 20151004, 1, null, 0.0, null, User.SUPER_USER);

        ContactProp c = new ContactProp();
        c.asOfyyyymmdd = 20151118;
        c.firstName = "Sharmila";
        c.email = "sharmila@sharmila.com";
        sharmila = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.addOrDeleteList(client, sharmila.memberId, newsletterListSgp.listId, true, User.SUPER_USER);
        Member.addOrDeleteList(client, sharmila.memberId, ishakriyaListSgp.listId, true, User.SUPER_USER);
        Member.addOrDeleteProgram(client, sharmila.memberId, ishakriya20151004.programId, true, User.SUPER_USER);
        Member.subscribeGroup(client, sharmila.memberId, sgp.groupId, User.SUPER_USER);

        c = new ContactProp();
        c.asOfyyyymmdd = 20151118;
        c.firstName = "Patma";
        c.email = "patma@patma.com";
        patma = Member.create(client, kl.groupId, c, false, User.SUPER_USER);
        Member.addOrDeleteList(client, patma.memberId, newsletterListKl.listId, true, User.SUPER_USER);

        MailContent.create(client, MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_SUCCESS.toString(), 0,
                "Scheduled mail sent successfully", "Hello, Mail sent successfully", User.SUPER_USER);

        MailContent.create(client, MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE.toString(), 0,
                "Scheduled mail failure", "Hello, Mail send failure", User.SUPER_USER);

        Client.addOrDeleteAllowedEmailSender(client, "singapore@ishayoga.org", "IshaCRM Bot", true, User.SUPER_USER);
        Client.setContactNameAndEmail(client, "singapore@ishayoga.org", "Isha", User.SUPER_USER);
    }

    @After
    public void tearDown() {
        ObjectifyFilter.complete();
        datastoreHelper.tearDown();
    }

    @Test(expected = APIException.class)
    public void cannotScheduleEmailForInvalidClient() {

        MailScheduleInput input = getTestMailScheduleInput();

        MailSchedule.create("invalid", input, User.SUPER_USER);
    }

    private MailScheduleInput getTestMailScheduleInput() {
        MailScheduleInput input = new MailScheduleInput();
        input.defaultFirstName = "Isha";
        input.defaultLastName = "Isha";
        input.enabled = true;
        input.listId = ishakriyaListSgp.listId;
        input.mailContentId = newsletter.mailContentId; //don't change this as some test cases depend on this
        Date nowPlusOneDay = new Date(new Date().getTime() + 86400 * 1000);
        input.scheduledTime = nowPlusOneDay;
        input.senderEmail = "singapore@ishayoga.org";

        return input;
    }


    @Test
    public void userWOPermissionCannotScheduleEmail() {
        try {
            MailScheduleInput msi = getTestMailScheduleInput();

            MailSchedule.create(client, msi, validUser.email);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_INSUFFICIENT_PERMISSION, ex.statusCode);
        }
    }

    @Test
    public void cannotScheduleForTimeInPast() {
        try {
            Date before24Hours = new Date(new Date().getTime() - 86400 * 1000);

            MailScheduleInput msi = getTestMailScheduleInput();
            msi.scheduledTime = before24Hours;

            MailSchedule.create(client, msi, User.SUPER_USER);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }
    }


    @Test
    public void cannotCreateScheduledEmailForInvalidList() {
        try {
            MailScheduleInput msi = getTestMailScheduleInput();
            msi.listId = 93848; //invalid list id

            MailSchedule.create(client, msi, User.SUPER_USER);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND, ex.statusCode);
        }
    }

    @Test
    public void cannotCreateScheduleIfListHasNoMembers() {
        ListProp listProp = List.createPublic(client, sgp.groupId, "List", User.SUPER_USER);
        assertNotNull(listProp);

        MailScheduleInput msi = getTestMailScheduleInput();
        msi.listId = listProp.listId;

        try {
            MailSchedule.create(client, msi, User.SUPER_USER);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }
    }

    @Test
    public void cannotCreateScheduleIfMailAlreadySentToAll() throws IOException, MandrillApiError {
        ListProp listProp = List.createPublic(client, sgp.groupId, "List", User.SUPER_USER);
        assertNotNull(listProp);

        MailScheduleInput msi = getTestMailScheduleInput();
        msi.listId = listProp.listId;

        //add member to list
        Member.addOrDeleteList(client, sharmila.memberId, listProp.listId, true, User.SUPER_USER);

        MailTest.suppressEmailInTestEnv();
        java.util.List<SentMailEntity> entities = Mail.sendToList(client, listProp.listId, newsletter.mailContentId,
                "singapore@ishayoga.org", null, User.SUPER_USER, "Isha", "Isha");
        assertEquals(1, entities.size());

        //try to schedule the same email. should get an error
        try {
            MailSchedule.create(client, msi, User.SUPER_USER);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //add another member to list
        boolean change = Member.addOrDeleteList(client, patma.memberId, listProp.listId, true, User.SUPER_USER);
        Member.subscribeGroup(client, patma.memberId, listProp.groupId, User.SUPER_USER);
        assertTrue(change);
        Map<Long, MemberEntity> map = MemberLoader.get(client, Utils.getSet(patma.memberId), User.SUPER_USER);

        //should be able to schedule the same email now
        MailSchedule.create(client, msi, User.SUPER_USER); //no exception
    }

    @Test
    public void userWithPermissionCanScheduleEmail() {
        MailScheduleInput msi = new MailScheduleInput();
        msi.listId = ishakriyaListSgp.listId;
        Date after24Hours = new Date(new Date().getTime() + 86400 * 1000);
        msi.scheduledTime = after24Hours;
        msi.senderEmail = "singapore@ishayoga.org";
        msi.mailContentId = newsletterListSgp.listId;
        msi.defaultFirstName = "Isha";
        msi.defaultLastName = "Isha";
        msi.enabled = true;
        Date after72Hours = new Date(new Date().getTime() + 86400 * 1000 * 3);
        msi.maxTime = after72Hours;

        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);

        mailScheduleProp = MailSchedule.safeGet(client, mailScheduleProp.mailScheduleId).toProp();

        //check all fields populated correctly
        assertEquals(1, mailScheduleProp.mailScheduleId);
        assertEquals(newsletter.mailContentId, mailScheduleProp.mailContentId);
        assertEquals(after24Hours, mailScheduleProp.scheduledTime);
        assertEquals(after72Hours, mailScheduleProp.maxTime);
        assertEquals(ishakriyaListSgp.listId, mailScheduleProp.listId);
        assertEquals(userWithSgpEmailPrivilage.email, mailScheduleProp.userEmail);
        assertEquals(newsletterListSgp.groupId, mailScheduleProp.groupId);
        assertEquals(false, mailScheduleProp.processed);
        assertNull(mailScheduleProp.processTime);
        assertNull(mailScheduleProp.sendSuccess);
        assertNull(mailScheduleProp.failureMessage);
    }

    @Test
    public void maxTimeDefaultsTo48Hours() {
        MailScheduleInput msi = new MailScheduleInput();
        msi.listId = ishakriyaListSgp.listId;
        Date after24Hours = new Date(new Date().getTime() + 86400 * 1000);
        msi.scheduledTime = after24Hours;
        msi.senderEmail = "singapore@ishayoga.org";
        msi.mailContentId = newsletterListSgp.listId;
        msi.defaultFirstName = "Isha";
        msi.defaultLastName = "Isha";
        msi.enabled = true;
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);

        mailScheduleProp = MailSchedule.safeGet(client, mailScheduleProp.mailScheduleId).toProp();

        //check all fields populated correctly
        assertEquals(1, mailScheduleProp.mailScheduleId);
        long diff = mailScheduleProp.maxTime.getTime() - mailScheduleProp.scheduledTime.getTime();
        assertEquals(48 * 3600 * 1000, diff);
    }

    @Test
    public void mailScheduleIdAssignedInSequence() {

        MailScheduleInput msi = getTestMailScheduleInput();
        MailScheduleProp msp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, msp.mailScheduleId);

        Date after1Day = new Date(new Date().getTime() + 86400 * 1000);
        msi.scheduledTime = after1Day;
        msi.listId = newsletterListSgp.listId;
        msp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(2, msp.mailScheduleId);
    }

    @Test
    public void userWithPermissionCanDeleteScheduledEmail() {
        MailScheduleInput msi = getTestMailScheduleInput();
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, mailScheduleProp.mailScheduleId);

        MailSchedule.delete(client, mailScheduleProp.mailScheduleId, userWithSgpEmailPrivilage.email);

        try {
            MailSchedule.safeGet(client, mailScheduleProp.mailScheduleId);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND, ex.statusCode);
        }
    }

    @Test
    public void cannotScheduleMultipleEmailForProgramParticipantsWithin24Hours() {
        MailScheduleInput msi = getTestMailScheduleInput();
        msi.programId = ishakriya20151004.programId;
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, mailScheduleProp.mailScheduleId);

        //another one after 3 hours
        msi.scheduledTime = new Date(msi.scheduledTime.getTime() + 3 * 3600 * 1000);
        msi.mailContentId = welcome.mailContentId;

        try {
            MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //another one before 3 hours
        msi.scheduledTime = new Date(msi.scheduledTime.getTime() - 3 * 3600 * 1000);
        msi.mailContentId = welcome.mailContentId;

        try {
            MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }
    }

    @Test
    public void canScheduleMultipleEmailForProgramParticipantsAfter24Hours() {
        MailScheduleInput msi = getTestMailScheduleInput();
        msi.programId = ishakriya20151004.programId;
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, mailScheduleProp.mailScheduleId);

        msi.scheduledTime = new Date(msi.scheduledTime.getTime() + 36 * 3600 * 1000);
        msi.mailContentId = welcome.mailContentId;

        MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        //no exception
    }


    @Test
    public void cannotScheduleSameMailAgainForListTwice() {
        MailScheduleInput msi = getTestMailScheduleInput();
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, mailScheduleProp.mailScheduleId);

        final int THREE_DAYS_IN_MS = 3 * 24 * 86400 * 1000;
        msi.scheduledTime = new Date(msi.scheduledTime.getTime() + THREE_DAYS_IN_MS);
        msi.maxTime = new Date(msi.scheduledTime.getTime() + 48 * 3600 * 1000); //maximum 48 hour delay

        try {
            MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_ALREADY_EXISTS, ex.statusCode);
        }

        //should be able to schedule another mail content
        msi.mailContentId = welcome.mailContentId;
        MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        //no exception
    }

    @Test
    public void queryTest() {
        //query should return empty list when no entities
        MailScheduleQueryCondition qc = new MailScheduleQueryCondition();
        java.util.List<MailScheduleProp> mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertTrue(mailScheduleProps.isEmpty());

        MailScheduleInput msi = getTestMailScheduleInput();
        MailScheduleProp mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, mailScheduleProp.mailScheduleId);

        msi.mailContentId = welcome.mailContentId;
        msi.scheduledTime = new Date(msi.scheduledTime.getTime() + 86400 * 1000);
        mailScheduleProp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(2, mailScheduleProp.mailScheduleId);

        mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertEquals(2, mailScheduleProps.size());
        //should be sorted by time
        assertEquals(1, mailScheduleProps.get(0).mailScheduleId);
        assertEquals(2, mailScheduleProps.get(1).mailScheduleId);

        //query by mailContentId
        //only welcome
        qc.mailContentIds.add(welcome.mailContentId);
        mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertEquals(1, mailScheduleProps.size());
        assertEquals(2, mailScheduleProps.get(0).mailScheduleId);
        //both welcome and newsletter
        qc.mailContentIds.add(newsletter.mailContentId);
        mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertEquals(2, mailScheduleProps.size());
        assertEquals(1, mailScheduleProps.get(0).mailScheduleId);
        assertEquals(2, mailScheduleProps.get(1).mailScheduleId);

        //query for a group
        qc.groupId = kl.groupId;
        mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertTrue(mailScheduleProps.isEmpty());

        //query for user and group
        qc.groupId = sgp.groupId;
        qc.userEmail = userWithSgpEmailPrivilage.email;
        qc.mailContentIds.clear();
        mailScheduleProps = MailSchedule.query(client, qc, validUser.email);
        assertEquals(2, mailScheduleProps.size());
        //should be sorted by time
        assertEquals(1, mailScheduleProps.get(0).mailScheduleId);
        assertEquals(2, mailScheduleProps.get(1).mailScheduleId);
    }

    @Test
    public void processScheduledEmailTest() throws InterruptedException, IOException, MandrillApiError {
        MailScheduleInput msi = getTestMailScheduleInput();
        Date after10Ms = new Date(new Date().getTime() + 10);
        msi.scheduledTime = after10Ms;
        MailScheduleProp msp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);

        Thread.sleep((long) 200);
        long time = System.currentTimeMillis();
        MailTest.suppressEmailInTestEnv();
        MailScheduleEntity mse = MailSchedule.processFirstScheduledEmail(client, new Date());
        assertEquals(msp.mailScheduleId, mse.mailScheduleId);
        assertEquals(true, mse.processed);
        assertEquals(true, mse.sendSuccess);

        msp = MailSchedule.safeGet(client, mse.mailScheduleId).toProp();
        assertTrue(msp.processed);
        assertTrue(Math.abs(msp.processTime.getTime() - time) < 50);
        assertTrue(msp.sendSuccess);
        assertNull(msp.failureCode);
        assertNull(msp.failureMessage);

        SentMailQueryCondition smqc = new SentMailQueryCondition();
        smqc.mailScheduleId = msp.mailScheduleId;
        java.util.List<SentMailEntity> sentMailEntities = Mail.queryEntitiesSortedByTimeDesc(client, smqc, User.SUPER_USER);
        assertEquals(1, sentMailEntities.size());

        //success notification should be sent to user
        smqc = new SentMailQueryCondition();
        smqc.tags = Utils.getSet(Mail.MailTag.SCHEDULED_EMAIL_SUCCESS.toString());
        sentMailEntities = Mail.queryEntitiesSortedByTimeDesc(client, smqc, User.SUPER_USER);
        assertEquals(1, sentMailEntities.size());
        long successMailContentId = MailContent.getByName(client,
                MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_SUCCESS.toString(), 0).mailContentId;
        assertEquals(successMailContentId, sentMailEntities.get(0).mailContentId);

        //attempt to send one more time
        mse = MailSchedule.processFirstScheduledEmail(client, new Date());
        assertNull(mse); //no schedule email was processed
    }

    @Test
    public void notificationSentToUserAfterError() throws InterruptedException, IOException, MandrillApiError {
        MailScheduleInput msi = getTestMailScheduleInput();
        Date after100Ms = new Date(new Date().getTime() + 100);
        msi.scheduledTime = after100Ms;
        MailScheduleProp msp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);

        //disable the list
        List.disable(client, ishakriyaListSgp.listId, User.SUPER_USER);

        Thread.sleep((long) 200);
        long time = System.currentTimeMillis();
        MailTest.suppressEmailInTestEnv();
        MailScheduleEntity mse = MailSchedule.processFirstScheduledEmail(client, new Date());
        assertEquals(msp.mailScheduleId, mse.mailScheduleId);

        msp = MailSchedule.safeGet(client, mse.mailScheduleId).toProp();
        assertTrue(msp.processed);
        assertTrue(Math.abs(msp.processTime.getTime() - time) < 50);
        assertFalse(msp.sendSuccess);
        assertEquals(MailSchedule.FailureCode.MAIL_SEND_EXCEPTION, msp.failureCode);
        System.out.println("msp.failureMessage: " + msp.failureMessage);

        SentMailQueryCondition smqc = new SentMailQueryCondition();
        smqc.mailScheduleId = msp.mailScheduleId;
        java.util.List<SentMailEntity> sentMailEntities = Mail.queryEntitiesSortedByTimeDesc(client, smqc, User.SUPER_USER);
        assertTrue(sentMailEntities.isEmpty());

        //failure notification should be sent to user
        smqc = new SentMailQueryCondition();
        smqc.tags = Utils.getSet(Mail.MailTag.SCHEDULED_EMAIL_FAILURE.toString());
        sentMailEntities = Mail.queryEntitiesSortedByTimeDesc(client, smqc, User.SUPER_USER);
        assertEquals(1, sentMailEntities.size());
        long successMailContentId = MailContent.getByName(client,
                MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE.toString(), 0).mailContentId;
        assertEquals(successMailContentId, sentMailEntities.get(0).mailContentId);
    }

    @Test
    public void notSentWhenPastMaxTime() throws InterruptedException, IOException, MandrillApiError {
        MailScheduleInput msi = getTestMailScheduleInput();
        Date after10Ms = new Date(new Date().getTime() + 10);
        msi.scheduledTime = after10Ms;
        Date after20Ms = new Date(after10Ms.getTime() + 10);
        msi.maxTime = after20Ms;
        MailScheduleProp msp = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);

        Thread.sleep((long) 200);
        long time = System.currentTimeMillis();
        MailTest.suppressEmailInTestEnv();
        MailScheduleEntity mse = MailSchedule.processFirstScheduledEmail(client, new Date());
        assertEquals(msp.mailScheduleId, mse.mailScheduleId);
        assertEquals(true, mse.processed);
        assertEquals(false, mse.sendSuccess);
        assertEquals(0, mse.numRecipients);

        //error notification should be sent to user
        SentMailQueryCondition smqc = new SentMailQueryCondition();
        smqc.tags = Utils.getSet(Mail.MailTag.SCHEDULED_EMAIL_FAILURE.toString());
        java.util.List<SentMailEntity> sentMailEntities = Mail.queryEntitiesSortedByTimeDesc(client, smqc, User.SUPER_USER);
        assertEquals(1, sentMailEntities.size());
        long successMailContentId = MailContent.getByName(client,
                MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE.toString(), 0).mailContentId;
        assertEquals(successMailContentId, sentMailEntities.get(0).mailContentId);
    }

    @Test
    public void multipleClients() {
        //todo: sathya
    }
}
