package crmdna.mail2;

import com.google.appengine.tools.development.testing.LocalDatastoreServiceTestConfig;
import com.google.appengine.tools.development.testing.LocalServiceTestHelper;
import com.google.common.collect.Lists;
import com.googlecode.objectify.ObjectifyFilter;
import crmdna.client.Client;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.contact.ContactProp;
import crmdna.group.Group;
import crmdna.group.Group.GroupProp;
import crmdna.list.List;
import crmdna.list.ListProp;
import crmdna.member.Member;
import crmdna.member.MemberProp;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.programtype.ProgramType;
import crmdna.programtype.ProgramTypeProp;
import crmdna.teacher.Teacher;
import crmdna.user.User;
import crmdna.user.User.GroupLevelPrivilege;
import crmdna.user.UserProp;
import crmdna.venue.Venue;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;


import java.util.Date;

import static org.junit.Assert.*;

public class MailSequenceTest {
    private final LocalServiceTestHelper datastoreHelper = new LocalServiceTestHelper(
            new LocalDatastoreServiceTestConfig().setApplyAllHighRepJobPolicy());

    private final String client = "isha";
    private GroupProp sgp;
    private GroupProp kl;

    private UserProp validUser;
    private UserProp userWithSgpEmailPrivilage;

    private MailContentProp newsletter;
    private MailContentProp welcome;
    private ListProp newsletterListSgp;
    private ListProp ishakriyaListSgp;
    private ListProp newsletterListKl;

    private ProgramProp ishakriya20151004;

    private MemberProp sharmila, patma;

    @Before
    public void setUp() {
        datastoreHelper.setUp();
        ObjectifyFilter.complete();

        Client.create(client);
        sgp = Group.create(client, "Singapore", User.SUPER_USER);
        assertEquals(1, sgp.groupId);

        kl = Group.create(client, "KL", User.SUPER_USER);
        assertEquals(2, kl.groupId);

        Group.addOrDeleteAllowedEmailSender(client, sgp.groupId,
                "singapore@ishayoga.org", "Isha Singapore", true,
                User.SUPER_USER);

        userWithSgpEmailPrivilage = User.create(client, "userwithsgpemail@invalid.com", sgp.groupId,
                User.SUPER_USER);
        User.addGroupLevelPrivilege(client, sgp.groupId, userWithSgpEmailPrivilage.email,
                GroupLevelPrivilege.SEND_EMAIL, User.SUPER_USER);

        validUser = User.create(client, "validuser@invalid.com", sgp.groupId,
                User.SUPER_USER);

        newsletter = MailContent.create(client, "Newsletter", sgp.groupId, "Mail content subject",
                "Mail content body", User.SUPER_USER);
        welcome = MailContent.create(client, "Welcome", sgp.groupId, "Welcome", "Welcome", User.SUPER_USER);

        newsletterListSgp = List.createPublic(client, sgp.groupId, "Newsletter sgp list", User.SUPER_USER);
        ishakriyaListSgp = List.createPublic(client, sgp.groupId, "Isha kriya sgp list", User.SUPER_USER);
        newsletterListKl = List.createPublic(client, kl.groupId, "Newsletter kl list", User.SUPER_USER);

        ProgramTypeProp ishakriya = ProgramType.create(client, "Isha Kriya", null, false, User.SUPER_USER);
        Teacher.TeacherProp shilpi = Teacher.create(client, "Shilpi", "Malhotra", "s1@gmail.com"
                , sgp.groupId, User.SUPER_USER);
        Venue.VenueProp woodlands = Venue.create(client, "Woodlands", "w", "w", "", sgp.groupId, User.SUPER_USER);
        ishakriya20151004 = Program.create(client, sgp.groupId, ishakriya.programTypeId,
                woodlands.venueId, shilpi.teacherId, 20151004, 20151004, 1, null, 0.0, null, User.SUPER_USER);

        ContactProp c = new ContactProp();
        c.asOfyyyymmdd = 20151118;
        c.firstName = "Sharmila";
        c.email = "sharmila@sharmila.com";
        sharmila = Member.create(client, sgp.groupId, c, false, User.SUPER_USER);
        Member.addOrDeleteList(client, sharmila.memberId, newsletterListSgp.listId, true, User.SUPER_USER);
        Member.addOrDeleteList(client, sharmila.memberId, ishakriyaListSgp.listId, true, User.SUPER_USER);
        Member.addOrDeleteProgram(client, sharmila.memberId, ishakriya20151004.programId, true, User.SUPER_USER);
        Member.subscribeGroup(client, sharmila.memberId, sgp.groupId, User.SUPER_USER);

        c = new ContactProp();
        c.asOfyyyymmdd = 20151118;
        c.firstName = "Patma";
        c.email = "patma@patma.com";
        patma = Member.create(client, kl.groupId, c, false, User.SUPER_USER);
        Member.addOrDeleteList(client, patma.memberId, newsletterListKl.listId, true, User.SUPER_USER);

        MailContent.create(client, MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_SUCCESS.toString(), 0,
                "Scheduled mail sent successfully", "Hello, Mail sent successfully", User.SUPER_USER);

        MailContent.create(client, MailContent.ReservedMailContentName.RESERVED_SCHEDULED_EMAIL_FAILURE.toString(), 0,
                "Scheduled mail failure", "Hello, Mail send failure", User.SUPER_USER);

        Client.addOrDeleteAllowedEmailSender(client, "singapore@ishayoga.org", "IshaCRM Bot", true, User.SUPER_USER);
        Client.setContactNameAndEmail(client, "singapore@ishayoga.org", "Isha", User.SUPER_USER);
    }

    @After
    public void tearDown() {
        ObjectifyFilter.complete();
        datastoreHelper.tearDown();
    }

    @Test
    public void createTest() {
        //happy path when all inputs are valid
        MailSequenceProp input = new MailSequenceProp();
        input.groupId = sgp.groupId;
        input.mailContentIds = Lists.newArrayList(welcome.mailContentId, newsletter.mailContentId);
        input.dayOffsets = Lists.newArrayList(0, 2);
        input.displayName = "Isha Kriya Nurture";
        input.programTypeId = ishakriya20151004.programTypeProp.programTypeId;

        MailSequence.create(client, input, User.SUPER_USER);
        MailSequenceEntity prop = MailSequence.safeGet(client, 1);
        assertEquals(1, prop.mailSequenceId);
        assertEquals(2, prop.dayOffsets.size());
        assertEquals(0, prop.dayOffsets.get(0).intValue());
        assertEquals("Isha Kriya Nurture", prop.displayName);
        assertEquals("ishakriyanurture", prop.name);
        assertEquals(sgp.groupId, prop.groupId);
        assertEquals(2, prop.mailContentIds.size());
        assertEquals(welcome.mailContentId, prop.mailContentIds.get(0).longValue());
        assertEquals(newsletter.mailContentId, prop.mailContentIds.get(1).longValue());
        assertEquals(input.programTypeId, prop.programTypeId);
    }

    @Test
    public void queryTest() {
        //empty when no records
        MailSequenceQueryCondition qc = new MailSequenceQueryCondition();

        java.util.List<MailSequenceProp> props = MailSequence.query(client, qc, validUser.email);
        assertTrue(props.isEmpty());

        qc.groupId = sgp.groupId;
        props = MailSequence.query(client, qc, validUser.email);
        assertTrue(props.isEmpty());

        MailSequenceProp input = new MailSequenceProp();
        input.groupId = sgp.groupId;
        input.mailContentIds = Lists.newArrayList(welcome.mailContentId, newsletter.mailContentId);
        input.dayOffsets = Lists.newArrayList(0, 2);
        input.displayName = "Isha Kriya Nurture";
        input.programTypeId = ishakriya20151004.programTypeProp.programTypeId;
        MailSequence.create(client, input, User.SUPER_USER);

        qc.groupId = kl.groupId;
        qc.programTypeId = ishakriya20151004.programTypeProp.programTypeId;
        props = MailSequence.query(client, qc, validUser.email);
        assertEquals(0, props.size());

        qc.groupId = sgp.groupId;
        props = MailSequence.query(client, qc, validUser.email);
        assertEquals(1, props.size());
    }

    @Test
    public void triggerHappyPathTest() {
        MailSequenceProp input = new MailSequenceProp();
        input.groupId = sgp.groupId;
        input.mailContentIds = Lists.newArrayList(welcome.mailContentId, newsletter.mailContentId);
        input.dayOffsets = Lists.newArrayList(0, 2);
        input.displayName = "Isha Kriya Nurture";
        input.programTypeId = ishakriya20151004.programTypeProp.programTypeId;

        MailSequenceProp prop = MailSequence.create(client, input, User.SUPER_USER);

        Date firstMailScheduleTime = new Date(new Date().getTime() + 3600 * 1000);
        MailSequence.trigger(client, prop.mailSequenceId, ishakriya20151004.programId, ishakriyaListSgp.listId,
                "Isha", "Isha", firstMailScheduleTime, "singapore@ishayoga.org", userWithSgpEmailPrivilage.email);

        MailScheduleQueryCondition msqc = new MailScheduleQueryCondition();
        msqc.programId = ishakriya20151004.programId;

        java.util.List<MailScheduleProp> mailScheduleProps = MailSchedule.query(client, msqc, validUser.email);
        assertEquals(2, mailScheduleProps.size());
        assertEquals(firstMailScheduleTime, mailScheduleProps.get(0).scheduledTime);
        assertEquals(welcome.mailContentId, mailScheduleProps.get(0).mailContentId);
        final long MILLIS_IN_A_DAY = 86400 * 1000;
        Date secondEmailScheduleTime = new Date(firstMailScheduleTime.getTime() + 2 * MILLIS_IN_A_DAY);
        assertEquals(secondEmailScheduleTime, mailScheduleProps.get(1).scheduledTime);
        assertEquals(newsletter.mailContentId, mailScheduleProps.get(1).mailContentId);
    }

    @Test
    public void cannotTriggerIfOneMailAlreadyScheduled() {
        MailScheduleInput msi = new MailScheduleInput();
        msi.senderEmail = "singapore@ishayoga.org";
        Date oneHourFromNow = new Date(new Date().getTime() + 3600 * 1000);
        msi.scheduledTime = oneHourFromNow;
        msi.programId = ishakriya20151004.programId;
        msi.defaultFirstName = "Isha";
        msi.defaultLastName = "Isha";
        msi.listId = ishakriyaListSgp.listId;
        msi.mailContentId = newsletter.mailContentId;

        MailScheduleProp prop = MailSchedule.create(client, msi, userWithSgpEmailPrivilage.email);
        assertEquals(1, prop.mailScheduleId);

        //now trigger a nurture campaign that includes the newsletter email
        MailSequenceProp input = new MailSequenceProp();
        input.groupId = sgp.groupId;
        input.mailContentIds = Lists.newArrayList(welcome.mailContentId, newsletter.mailContentId);
        input.dayOffsets = Lists.newArrayList(0, 2);
        input.displayName = "Isha Kriya Nurture";
        input.programTypeId = ishakriya20151004.programTypeProp.programTypeId;

        MailSequenceProp mailSequenceProp = MailSequence.create(client, input, User.SUPER_USER);

        Date fiveDaysFromNow = new Date(new Date().getTime() + 5 * 86400 * 1000);
        Date firstMailScheduleTime = fiveDaysFromNow;

        try {
            MailSequence.trigger(client, mailSequenceProp.mailSequenceId,
                    ishakriya20151004.programId, ishakriyaListSgp.listId,
                    "Isha", "Isha", firstMailScheduleTime,
                    "singapore@ishayoga.org", userWithSgpEmailPrivilage.email);
            assertTrue(false);
        } catch (APIException ex) {
            //System.out.println("message: " + ex.userFriendlyMessage);
            assertEquals(APIResponse.Status.ERROR_RESOURCE_ALREADY_EXISTS, ex.statusCode);
        }

        //no new mailschedule should be created
        MailScheduleQueryCondition msqc = new MailScheduleQueryCondition();
        msqc.programId = ishakriya20151004.programId;
        java.util.List<MailScheduleProp> mailScheduleProps = MailSchedule.query(client, msqc, validUser.email);
        assertEquals(1, mailScheduleProps.size());
        assertEquals(newsletter.mailContentId, mailScheduleProps.get(0).mailContentId);
    }

    @Test
    public void ensureValidDayOffsetsTest() {

        //not starting with zero
        java.util.List<Integer> dayOffsets = Lists.newArrayList(1, 2, 3);

        try {
            MailSequence.ensureValidDayOffsets(dayOffsets);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //not in increasing order
        dayOffsets = Lists.newArrayList(0, 3, 3);
        try {
            MailSequence.ensureValidDayOffsets(dayOffsets);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //more than 120 days diff between first and last
        dayOffsets = Lists.newArrayList(0, 3, 200);
        try {
            MailSequence.ensureValidDayOffsets(dayOffsets);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //all good
        dayOffsets = Lists.newArrayList(0, 3, 5, 7, 9, 11);
        MailSequence.ensureValidDayOffsets(dayOffsets);

        dayOffsets = Lists.newArrayList(0, 7, 15, 21);
        MailSequence.ensureValidDayOffsets(dayOffsets);
    }

    @Test
    public void ensureValidMailContentIdsTest() {

        java.util.List<Long> mailContentIds = Lists.newArrayList((long) 100, (long) 101, (long) 102);

        //all mailContentIds should be valid
        try {
            MailSequence.ensureValidMailContentIds(client, mailContentIds, sgp.groupId);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_NOT_FOUND, ex.statusCode);
        }

        //group should be valid
        mailContentIds = Lists.newArrayList(welcome.mailContentId, newsletter.mailContentId);
        try {
            MailSequence.ensureValidMailContentIds(client, mailContentIds, kl.groupId);
            assertTrue(false);
        } catch (APIException ex) {
            assertEquals(APIResponse.Status.ERROR_RESOURCE_INCORRECT, ex.statusCode);
        }

        //all good
        MailSequence.ensureValidMailContentIds(client, mailContentIds, sgp.groupId);
    }
}
