package crmdna.payment;

import com.google.appengine.tools.development.testing.LocalDatastoreServiceTestConfig;
import com.google.appengine.tools.development.testing.LocalServiceTestHelper;
import com.googlecode.objectify.ObjectifyFilter;
import crmdna.client.Client;
import crmdna.common.DateUtils;
import crmdna.common.Utils;
import crmdna.common.api.APIException;
import crmdna.common.api.APIResponse;
import crmdna.common.contact.Contact;
import crmdna.common.contact.ContactProp;
import crmdna.group.Group;
import crmdna.mail2.MailContent;
import crmdna.mail2.MailTest;
import crmdna.payment.Receipt.Purpose;
import crmdna.practice.Practice;
import crmdna.program.Program;
import crmdna.program.ProgramProp;
import crmdna.programtype.ProgramType;
import crmdna.programtype.ProgramTypeProp;
import crmdna.registration.Registration;
import crmdna.registration.RegistrationInput;
import crmdna.registration.RegistrationProp;
import crmdna.teacher.Teacher;
import crmdna.user.User;
import crmdna.venue.Venue;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.util.*;

import static org.junit.Assert.*;

public class ReceiptTest {

    private final LocalServiceTestHelper datastoreHelper = new LocalServiceTestHelper(
        new LocalDatastoreServiceTestConfig().setApplyAllHighRepJobPolicy());

    private final String client = "isha";
    Group.GroupProp sgp;
    Practice.PracticeProp shambhavi;
    ProgramTypeProp innerEngineering7Day;
    ProgramProp program;
    Venue.VenueProp giis;
    Teacher.TeacherProp teacher;
    RegistrationProp registration;

    @Before
    public void setUp() {
        datastoreHelper.setUp();
        ObjectifyFilter.complete();

        Client.create(client);
        sgp = Group.create(client, "Singapore", User.SUPER_USER);
        shambhavi = Practice.create(client, "Shambhavi", User.SUPER_USER);

        Group.setContactInfo(client, sgp.groupId, "contact@group.com", "name", User.SUPER_USER);

        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(System.currentTimeMillis() + (86400 * 1000)); // tomorrow

        int startDate = calendar.get(Calendar.DAY_OF_MONTH) +
            (calendar.get(Calendar.MONTH) + 1) * 100 +
            calendar.get(Calendar.YEAR) * 10000;

        calendar.setTimeInMillis(System.currentTimeMillis() + (86400 * 1000 * 2)); // day after tomorrow
        int endDate = calendar.get(Calendar.DAY_OF_MONTH) +
            (calendar.get(Calendar.MONTH) + 1) * 100 +
            calendar.get(Calendar.YEAR) * 10000;

        Set<Long> practiceIds = new HashSet<>();
        practiceIds.add(shambhavi.practiceId);
        innerEngineering7Day = ProgramType.create(client, "Inner Engineering 7 day", practiceIds, true, User.SUPER_USER);
        giis = Venue.create(client, "GIIS", "GIIS", "GIIS", "", sgp.groupId, User.SUPER_USER);
        teacher = Teacher.create(client, "", "", "teacher@if.org", sgp.groupId, User.SUPER_USER);
        program = Program.create(client, sgp.groupId, innerEngineering7Day.programTypeId,
            giis.venueId, teacher.teacherId, startDate, endDate, 1, null, 200.0, Utils.Currency.SGD,
            User.SUPER_USER);

        ContactProp contact = new ContactProp();
        contact.firstName = "firstName";
        contact.lastName = "lastName";
        contact.nickName = "nickName";
        contact.gender = Contact.Gender.MALE;
        contact.email = "thulasidhar@gmail.com";
        contact.homePhone = "+124567951";
        contact.mobilePhone = "+124567951";
        contact.officePhone = "+124567951";
        contact.homeAddress.country = "Singapore";
        contact.homeAddress.pincode = "113456";
        contact.asOfyyyymmdd = DateUtils.toYYYYMMDD(new Date());

        MailContent.create(client, "RESERVED_RECEIPT", sgp.groupId, "RECEIPT", "body", User.SUPER_USER);
        MailContent.create(client, "RESERVED_REGISTRATION_CONFIRMATION", 0, "CONFIRMATION", "body", User.SUPER_USER);

        RegistrationInput regInput = new RegistrationInput(contact, program.programId, program.fee, "offline");
        regInput.successCallbackUrl = "http://google.com";
        regInput.errorCallbackUrl = "http://google.com";
        regInput.rootUrl = "http://google.com";
        registration = Registration.register(client, regInput);
        assertEquals(registration.status, Registration.RegistrationStatus.REGISTRATION_COMPLETE);

        Group.setContactInfo(client, sgp.groupId, "singapore@ishayoga.org", "Isha Singapore", User.SUPER_USER);
    }

    @After
    public void tearDown() {
        ObjectifyFilter.complete();
        datastoreHelper.tearDown();
    }

    @Test
    public void generateForRegIdTest() {

        try {
            MailTest.suppressEmailInTestEnv();
            Receipt.generateForRegistration(client, sgp.groupId, registration.registrationId, false);
            ReceiptProp receiptProp = Receipt.getByRegistrationId(client, registration.registrationId);

            assertNotNull(receiptProp);
            assertTrue(receiptProp.ms <= (new Date()).getTime());
            assertEquals(sgp.groupId, receiptProp.groupId, sgp.groupId);
            assertEquals(registration.registrationId, receiptProp.registrationId.longValue());
            assertEquals(registration.firstName, receiptProp.firstName);
            assertEquals(registration.lastName, receiptProp.lastName);
            assertEquals(registration.email, receiptProp.email);
            assertEquals(program.getDetailedName(), receiptProp.purpose);
            assertEquals(Utils.Currency.SGD, receiptProp.ccy);
            assertEquals(program.fee, receiptProp.amount, 0);
        } catch (APIException ex) {
            System.out.println(ex.toString());
            fail();
        }
    }

    @Test
    public void generateForProgramContribution() {

        String firstName = "FirstName";
        String lastName = "LastName";
        String email = "thulasidhar@gmail.com";
        Purpose purpose = Purpose.VOLUNTEER_CONTRIBUTION;
        double amount = 200.0;

        try {
            MailTest.suppressEmailInTestEnv();
            Receipt.generateForProgram(client, sgp.groupId, firstName, lastName, email,
                program.programId, purpose, Utils.Currency.SGD, amount, 20151030);

            Receipt.ReceiptQueryCondition qc = new Receipt.ReceiptQueryCondition();
            qc.groupId = sgp.groupId;
            List<ReceiptProp> receiptProps = Receipt.query(client, qc);
            assertTrue(receiptProps.size() == 1);

            ReceiptProp receiptProp = receiptProps.get(0);
            assertTrue(Math.abs(receiptProp.ms - DateUtils.toDate(20151030).getTime()) < 1000);

            assertEquals(sgp.groupId, receiptProp.groupId);
            assertNull(receiptProp.registrationId);
            assertEquals(firstName, receiptProp.firstName);
            assertEquals(lastName, receiptProp.lastName);
            assertEquals(email, receiptProp.email);
            assertEquals(purpose.toString() + " - " + program.getDetailedName(), receiptProp.purpose);
            assertEquals(Utils.Currency.SGD, receiptProp.ccy);
            assertEquals(amount, receiptProp.amount, 0);
        } catch (APIException ex) {
            System.out.println(ex.toString());
            fail();
        }
    }

    @Test
    public void generateForAdhocDonation() {

        String firstName = "FirstName";
        String lastName = "LastName";
        String email = "thulasidhar@gmail.com";
        double amount = 150.50;
        String adhocReference = "Adhoc Donation";

        try {
            MailTest.suppressEmailInTestEnv();
            Receipt.generateForAdhocDonation(client, sgp.groupId, firstName, lastName, email,
                adhocReference, Utils.Currency.SGD, amount, 20151030, true);

            Receipt.ReceiptQueryCondition qc = new Receipt.ReceiptQueryCondition();
            qc.groupId = sgp.groupId;
            List<ReceiptProp> receiptProps = Receipt.query(client, qc);
            assertTrue(receiptProps.size() == 1);

            ReceiptProp receiptProp = receiptProps.get(0);
            assertTrue(Math.abs(receiptProp.ms - DateUtils.toDate(20151030).getTime()) < 1000);

            assertEquals(sgp.groupId, receiptProp.groupId);
            assertNull(receiptProp.registrationId);
            assertEquals(firstName, receiptProp.firstName);
            assertEquals(lastName, receiptProp.lastName);
            assertEquals(email, receiptProp.email);
            assertEquals(adhocReference, receiptProp.purpose);
            assertEquals(Utils.Currency.SGD, receiptProp.ccy);
            assertEquals(amount, receiptProp.amount, 0);

        } catch (APIException ex) {
            System.out.println(ex.toString());
            fail();
        }
    }

}
